#define VERSION "0.986 - February 2015"
#define SWITCHES "[-uaAgsh -c#xm#pe#f# -od -v -n]"

/* surftri.c :  generate imbedded graphs in surface   Version 0.986 */

/*

  Based extensively on the code, data structures, and ideas taken from plantri.
  The latest edition of plantri is available from the plantri page
  http://cs.anu.edu.au/~bdm/plantri .

  Authors of plantri:
        Gunnar Brinkmann
	  University of Bielefeld
	Brendan McKay
	  Australian National University

*/

/*
---------

  Author of surftri:
        Thom Sulanke
	Indiana University
	tsulanke@indiana.edu

---------

   Output formats:

      The output for each graph consists of the number of vertices,
      then for each vertex a list of all its neighbours in clockwise
      order.  These neighbour lists are separated by a "separator",
      and terminated by a "terminator".

                             planar_code          ascii format
                              (default)            (with -a)

      file type               binary               text
      number of vertices      one byte          decimal + blank
      vertex names          bytes 1,2,3...        letters 'a','b',...
      separators              zero byte           comma ','
      terminator              zero byte           newline '\n'

      For example, the planar dual of the cube appears like this:

      planar_code:
         6 4 3 2 5 0 3 6 5 1 0 1 4 6 2 0 5 6 3 1 0 4 1 2 6 0 4 5 2 3 0
         (One number per byte, no newlines).

      ascii format:
         6 dcbe,cfea,adfb,efca,dabf,debc
         (Including the space after "6".  Followed by newline.)

      For planar_code, the standard header ">>planar_code<<"
      (without null or newline) is written at the start of the
      output unless the -h switch is given.

      Using -g or -s, the graph6 or sparse6 formats can be selected instead.
      These are ASCII formats for general undirected graphs, and do not
      encode the imbedding.  graph6 does not represent loops or edge
      multiplicities either.  They are described in plantri-guide.txt.

---------

Size limits:

The space used by surftri is O(n^2).  Apart from that, the only
practical limits on MAXN (the maximum permitted number of vertices)
is determined by the limits imposed by the output syntax.

The following table gives the largest legal MAXN value.

 Switches:    none    -d    -a    -ad     -g    -gd  -s    -sd      -u
 Output:      planar_code     ascii        graph6     sparse6      none
             primal  dual  primal dual  primal dual  primal dual
 MAXN limit:  255    129     99     51   255   129    255   129    1023

The limits for ascii code could be raised to 114 and 59 easily.

**************************************************************************/

#include <stdio.h>
#include <string.h>

#if __STDC__
#include <stdlib.h>
#endif

#define CPUTIME 1          /* Whether to measure the cpu time or not */

#if CPUTIME
#include <sys/times.h>
#include <time.h>
#if !defined(CLK_TCK) && !defined(_SC_CLK_TCK)
#include <unistd.h>
#endif
#if !defined(CLK_TCK) && defined(_SC_CLK_TCK)
#define CLK_TCK sysconf(_SC_CLK_TCK)
#endif
#if !defined(CLK_TCK) && defined(CLOCKS_PER_SEC)
#define CLK_TCK CLOCKS_PER_SEC
#endif
#if !defined(CLK_TCK)
#define CLK_TCK 60     /* If the CPU time stated by the program appears
                      to be out by a constant ratio, the most likely
                      explanation is that the code got to this point but
                      60 is the wrong guess.  Another common value is 100. */
#endif
#endif

#ifndef MAXN
#define MAXN 50            /* the maximum number of vertices; see above */
#endif
#define MAXG 20             /* the maximum value for 2-(Euler characteristic) 
			      of a surface */
#define MAXE (6*(MAXN-2+MAXG)) /* the maximum number of oriented edges (for
				  oriented) or edge pairs (for nonoriented) in
				  a triagulation */
#define MAXF (2*(MAXN-2+MAXG)) /* the maximum number of faces */

typedef struct e /* The data type used for edges */
{
  int start;         /* vertex where the edge starts */
  int end;           /* vertex where the edge ends */
  int rightface;     /* face on the right side of the edge
			note: only valid if make_dual() called */
  struct e *prev;    /* previous edge in clockwise direction */
  struct e *next;    /* next edge in clockwise direction */
  struct e *invers;  /* the edge that is inverse to this one */
  struct e *op;      /* the edge that is oposite (other side of surface) to
			this one */
  int mark,index;    /* two ints for temporary use;
			Only access mark via the MARK macros. */
  int left_facesize; /* size of the face in prev-direction of the edge.
			Only used for -p option. */
} EDGE;

typedef struct
{
    EDGE *e1,*e2,*e3;
} triangle;

#undef FALSE
#undef TRUE
#define FALSE 0
#define TRUE  1

/* Global variables */

static char *outfilename;  /* name of output file (NULL for stdout) */
static FILE *outfile;      /* output file for graphs */
static FILE *msgfile;      /* file for informational messages */
static FILE *irrfile;      /* input file for irreducible graphs */

static int maxnv;          /* order of output graphs */
static int eulerc;         /* Euler characteristic of surface */
static int ggenus;         /* genus of triangulation (number of handles or 
			      cross-caps) */
                           /* if surface and triangulation are orientable then
			      eulerc = 2 - 2 * ggenus
			      if surface and triangulation are nonorientable
			      then eulerc = 2 - ggenus    */
static int orient;         /* surface and triangulations are orientable? */

static int res,mod;        /* res/mod from command line (default 0/1) */
static int splitlevel,
           splitcount;     /* used for res/mod splitting */
#ifdef PLUGIN
static int splithint = -1; /* used by plugins to set splitting level */
#endif
static int minconnec;      /* lower bound on minimum connectivity */
static int minpolyconnec;  /* lower bound on minumum connectivity for
			      polytopes.  Defaults to same as minconnec. */
static int polyconnec;     /* connectivity of current irreducible map. */
static int xconnec;        /* Value of connectivity appropriate for -x.
                              The same as either minconnec or minpolyconnec. */
static int edgebound[2];   /* edge count min,max for polytopes */
static int maxfacesize;    /* maximum face size for polytopes */

static int polygonsize;    /* polygon size for -P.
                              -1 means -P is absent
                               0 means size is unrestricted */

static int minimumdeg;     /* lower bound on minimum degree.
                              -1 means nothing specified */
static int minpolydeg;     /* lower bound on minimum degree for polytopes.
                              -1 will cause it to be reset to the same as
                              minimumdeg, but plugins can set it otherwise. */
static int irrhole;        /* true if an irreducible triangulations with holes
			      is being expanded */

static int aswitch,        /* presence of command-line switches */
           upaswitch,
           gswitch,
           sswitch,
           hswitch,
           dswitch,
           oswitch,
           qswitch,
	   pswitch,
           uswitch,
           nswitch,
           vswitch,
           xswitch,
           yswitch;

static int zeroswitch;     /* Undocumented option -0 for debugging */
static int oneswitch;      /* Undocumented option -1, implies -0 */
#ifdef STATS2
static int twoswitch;      /*  Undocumented option -2 */
#endif
static int needgroup;      /* Is group needed at end of scansimple()
			      and similar routines? */
static int gotone_nbop;    /* Used only by got_one() */
static int gotone_nbtot;

extern int errno;
static int dosummary;      /* used by plugin */
static char *cmdname;      /* points to arg[0] */

/* The variables below are used at each level of the iteration,
   updating and restoring as we move up and down the search tree */

static int nv;             /* number of vertices; they are 0..nv-1 */
static int ne;             /* number of directed edges in triangulation on one
			      side of surface, (6*(nv-2+genus)) */
static int nirr;           /* index of usable irreducible map (has small 
			      enough nv) as read from file */
static int nvroot;         /* nv of current root (the irreducible
			      triangulation generating the current
			      triangulation */
static int neroot;         /* ne of the current face irreducible map generating
			      the current map */

/* NUMEDGES is the number of the data type "edge" used to construct
   triangulations in the oriented case.  2*NUMEDGES are required for
   nonoriented case. */
#define NUMEDGES (MAXE+(16+128)*MAXN)
static EDGE edges[2*NUMEDGES];

#define init_edge edges
#define STAR3(n) (edges + MAXE + ((n)<<3))
#define STAR4(n) (edges + MAXE + 8*MAXN + ((n)<<3))
/* maximum number of edges per STARN < 2*MAXN <= 2*64 = 2**7 */
#define STARN(n) (edges + MAXE + 16*MAXN + ((n)<<7))

#define four_op(n) (edges + MAXE + 6*(n))
#define five_op(n) (edges + MAXE + 6*MAXN + 6*(n))
#define S_op(n) (edges + MAXE + 12*MAXN + 18*(n))

#define quadr_P0(n) (edges + MAXE -8 + 4*MAXN + 4*(n))
/* The smallest n for which it is called is 3. Then it should start at entry 
   edges +4 + 4*MAXN -- right AFTER the edges for quadr_P1(n) which is
   used in the same run. */

#define quadr_P1(n) (edges + MAXE +4 + 4*(n))
/* edges + 24 + 4*(n) - 5*4 since the smallest n for which it is possibly
   called is n=5  and then it should start at edge 24 */

#define quadr_P3(n) (edges + MAXE -60 + 4*MAXN + 16*(n))
/* The smallest n for which it is called is 4.  Then it should start at entry
   edges +4 + 4*MAXN -- right AFTER the edges for quadr_P1(n) which is 
   used in the same run. */

static int degree[MAXN];   /* the degrees of the vertices 
			      valid only for < nv */
static int ncontract[MAXN]; /* number of contractible edges from vertex 
			       valid only for < nv */
static int adj[MAXN][MAXN]; /* 1 if i and j are adjacent 
			       valid for all i and j even those >= nv */
static int nnei[MAXN][MAXN]; /* number of common neighbor of i and j plus 
				number of holes of which (i,j) is border edge.
			        valid only for i and j < nv */
static EDGE *firstedge[MAXN]; /* pointer to arbitrary edge out of vertex i. */
  /* This pointer may change during the run, so all one can rely on is that
     at any point it is "some" edge out of i */

static EDGE *numbering[2*MAXE][MAXE];
  /* holds numberings produced by canon() */
static EDGE *saved_numbering[2*MAXE][MAXE]; 
  /* a copy of numbering used by scanordloops() */

/* The following unpacked adjacency matrix is used for general planar
   graphs of connectivity 1 and 2. */

static char am2[MAXN][64];
#define AMADDARC(i,j) am2[i][j] = 1;
#define AMDELARC(i,j) am2[i][j] = 0;
#define AMADDEDGE(i,j) { am2[i][j] = am2[j][i] = 1; }
#define AMDELEDGE(i,j) { am2[i][j] = am2[j][i] = 0; }
#define ISEQADJ(i,j) (am2[i][j] != 0)
#define ISNEQADJ(i,j) (am2[i][j] == 0)

#define PCODE ">>planar_code<<"
#define PCODELEN (sizeof(PCODE)-1)    /* "-1" to avoid the null */
#define G6CODE ">>graph6<<"
#define G6CODELEN (sizeof(G6CODE)-1)    /* "-1" to avoid the null */
#define S6CODE ">>sparse6<<"
#define S6CODELEN (sizeof(S6CODE)-1)    /* "-1" to avoid the null */

static EDGE *code_edge = NULL;
/* if code_edge is not NULL, it is taken as the start for coding for
   ASCII or planar_code. Otherwise firstedge[0] is the start. This
   method implies comparatively few changes due to outputting 
   triangulations of disks. 

   In case of triangulations of disks, *code_edge should be an edge
   with the "outer" face on the left for the non-mirror case and on
   the right for the mirror case to have the outer face left of 1->2.

   In case of dual output (mirror image or not), the face on the left 
   of *code_edge gets the number 1. So for duals of triangulations of 
   disks, handing in an edge with the disk on the right outputs the 
   "marked" vertex as 1.
*/

static int missing_vertex = -1;
/* The vertices are numbered 0..nv-1 if missing_vertex<0, and
   0..missing_vertex-1,missing_vertex..nv otherwise.  This is
   only used in the code for polygon triangulations. */

static int zero[MAXN];   /* permanently 0 */

/* The program is so fast that the count of output graphs can quickly
   overflow a 32-bit integer.  Therefore, we use two long values
   for each count, with a ratio of 10^9 between them.  The macro
   ADDBIG adds a small number to one of these big numbers.
   BIGTODOUBLE converts a big number to a double (approximately).
   SUMBIGS adds a second big number into a first big number.
   SUBBIGS subtracts a second big number from a first big number.
   PRINTBIG prints a big number in decimal.
   ZEROBIG sets the value of a big number to 0.
   ISZEROBIG tests if the value is 0.
   SETBIG sets a big number to a value at most 10^9-1.
   ISEQBIG tests if two big numbers are equal.
*/

typedef struct
{
  long hi,lo;
} bigint;

#define ZEROBIG(big) big.hi = big.lo = 0L
#define ISZEROBIG(big) (big.lo == 0 && big.hi == 0)
#define SETBIG(big,value) {big.hi = 0L; big.lo = (value);}
#define ADDBIG(big,extra) if ((big.lo += (extra)) >= 1000000000L) \
    { ++big.hi; big.lo -= 1000000000L;}
#define PRINTBIG(file,big) if (big.hi == 0) \
 fprintf(file,"%ld",big.lo); else fprintf(file,"%ld%09ld",big.hi,big.lo)
#define BIGTODOUBLE(big) (big.hi * 1000000000.0 + big.lo)
#define SUMBIGS(big1,big2) {if ((big1.lo += big2.lo) >= 1000000000L) \
    {big1.lo -= 1000000000L; big1.hi += big2.hi + 1L;} \
    else big1.hi += big2.hi;}
#define SUBBIGS(big1,big2) {if ((big1.lo -= big2.lo) < 0L) \
    {big1.lo += 1000000000L; big1.hi -= big2.hi + 1L;} \
    else big1.hi -= big2.hi;}
/* Note: SUBBIGS must not allow the value to go negative.
   SUMBIGS and SUBBIGS both permit big1 and big2 to be the same bigint. */
#define ISEQBIG(big1,big2) (big1.lo == big2.lo && big1.hi == big2.hi)

static bigint nout[6];      /* counts of output graphs, per connectivity */
static bigint nout_op[6];   /* counts of output graphs, per connectivity, OP */
static bigint nout_e[MAXE/2+1];  /* .. per undirected edge number */
static bigint nout_e_op[MAXE/2+1];  /* .. per undirected edge number, OP */
static bigint totalout;       /* Total count */
static bigint totalout_op;    /* Total count, OP */

static bigint irrcount[MAXN];   /* count of irreducible by nv */

static char outtypename[50];  /* How to describe output objects */
static char intypename[50];  /* How to describe input objects */

static int markvalue = 30000;
#define RESETMARKS {int mki; if ((markvalue += 2) > 30000) \
       { markvalue = 2; for (mki=0;mki<2*NUMEDGES;++mki) edges[mki].mark=0;}}
#define MARKLO(e) (e)->mark = markvalue
#define MARKHI(e) (e)->mark = markvalue+1
#define UNMARK(e) (e)->mark = markvalue-1
#define ISMARKED(e) ((e)->mark >= markvalue)
#define ISMARKEDLO(e) ((e)->mark == markvalue)
#define ISMARKEDHI(e) ((e)->mark > markvalue)

/* and the same for vertices */

static int markvalue_v = 30000;
static int marks__v[MAXN];
#define RESETMARKS_V {int mki; if ((++markvalue_v) > 30000) \
       { markvalue_v = 1; for (mki=0;mki<MAXN;++mki) marks__v[mki]=0;}}
#define UNMARK_V(x) (marks__v[x] = 0)
#define ISMARKED_V(x) (marks__v[x] == markvalue_v)
#define MARK_V(x) (marks__v[x] = markvalue_v)

#define RIGHTHOLE(e) (e->invers->prev->invers->prev->invers->prev != e)

static EDGE *inmaxface[MAXN*MAXN-3*MAXN/2];   /* Used for polytope
  generation - lists edges whose left face is maximum size */

static void (*write_graph)(FILE*,int,int);
static void (*write_dual_graph)(FILE*,int,int);
static void check_it(int code);

#define CHECKSWITCH(name) check_switch(name,ok_switches)
#define INCOMPAT(cond,x,y) if (cond) \
  {fprintf(stderr,">E %s: %s and %s are incompatible\n",cmdname,x,y); exit(1);}
#define CHECKRANGE(var,varname,lo,hi) if ((var)<(lo)||(var)>(hi)) \
  {fprintf(stderr,">E %s: the value of %s must be ",cmdname,varname); \
  if ((lo)==(hi)) \
    fprintf(stderr,"%d\n",lo); else fprintf(stderr,"%d..%d\n",lo,hi); \
    exit(1);}
#define PERROR(cond,msg) if (cond) \
   {fprintf(stderr,">E %s: %s\n",cmdname,msg); exit(1);}

#define BOOLSWITCH(name,var)  \
    else if (arg[j]==name) {CHECKSWITCH(name); var = TRUE;}
#define INTSWITCH(name,var)  \
    else if (arg[j]==name) {CHECKSWITCH(name); var = getswitchvalue(arg,&j);}

#define SECRET_SWITCHES "0"

#define MAX(x,y) ((x)<(y) ? (y) : (x))
#define MIN(x,y) ((x)>(y) ? (y) : (x))

#ifdef SPLITTEST
static bigint splitcases;
#endif
//static int debug_count;
/**************************************************************************/

/* Include optional file for special processing. */

#ifdef PLUGIN
#include PLUGIN
#endif

/*************************************************************************/
#include "extend.c"
/*************************************************************************/

static int
testcanon(EDGE *givenedge, int representation[], int colour[])

/* Tests whether starting from a given edge and constructing the code in
   "->next" direction, an automorphism or even a better representation
   can be found. Returns 0 for failure, 1 for an automorphism and 2 for
   a better representation.  This function exits as soon as a better
   representation is found. A function that computes and returns the
   complete better representation can work pretty similar.*/
{
  EDGE *temp, *run;
  EDGE *startedge[MAXN+1]; /* startedge[i] is the starting edge for
			      exploring the vertex with the number i+1 */
  int number[MAXN], i;   /* The new numbers of the vertices, starting
			    at 1 in order to have "0" as a possibility to
			    mark ends of lists and not yet given numbers */
  int last_number, actual_number, vertex;
  int rep;

  RESETMARKS;

  for (i = 0; i < nv; i++) number[i] = 0;

  number[givenedge->start] = 1;

  if (givenedge->start != givenedge->end) /* no loop start */
    {
      number[givenedge->end] = 2;
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else last_number = 1 ;

  actual_number = 1;
  temp = givenedge;

/* A representation is a clockwise ordering of all neighbours ended with a 0.
   The neighbours are numbered in the order that they are reached by the BFS
   procedure. In case a vertex is reached for the first time, not the (new)
   number of the vertex is listed, but its colour. When the list around a
   vertex is finished, it is ended with a 0. Since the colours can be
   distinguished from the vertices (requirement for the colour function), the
   adjacency list can be reconstructed: Every time a colour is listed, its
   number would be the smallest number not given yet.
   Since the edges when a vertex is reached for the first time are remembered,
   for these edges we in fact have more than just the vertex information -- for
   these edges we also have the exact information which edge occurs in the
   cyclic order. This makes the routine work also for double edges.

   Since every representation starts with the colour of vertex 2, which is
   the same colour all the time, we do not have to store that.

   In case of a loop as the starting point, the colour of 1 is omitted.
   Nevertheless also in this case it cannot be mixed up with a non-loop
   starting point, since the number of times a colour occurs is larger
   for loop starters than for non-loop starters.

   Every first entry in a new clockwise ordering (the starting point of the
   edge it was numbered from is determined by the entries before (the first
   time it occurs in the list to be exact), so this is not given either.
   The K4 could be denoted  c3 c4 0 4 3 0 2 3 0 3 2 0 if ci is the colour
   of vertex i.  Note that the colour of vertex 1 is -- by definition --
   always the smallest one */

  while (last_number < nv)
    {
      for (run = temp->next; run != temp; run = run->next)
	/* this loop marks all edges around temp->origin. */
	{ vertex = run->end;
	  if (!number[vertex])
	    { startedge[last_number] = run->invers;
	      last_number++; number[vertex] = last_number;
	      rep = colour[vertex]; }
	  else if (number[vertex] > actual_number)
	    {
	      rep = number[vertex];
	      MARKLO(run);
	    }
	  else if (ISMARKED(run->invers))
	    {
	      rep = number[vertex];
	    }
	  else
	    {
	      rep = number[vertex]+2*MAXN;
	    }
	  if (rep > (*representation)) return 0;
	  if (rep < (*representation)) return 2;
	  representation++;
	}
      /* check whether representation[] is also at the end of a cyclic list */
      if ((*representation) != 0) return 2;
      representation++;
      /* Next vertex to explore: */
      temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nv)
    /* Now we know that all numbers have been given */
    {
      for (run = temp->next; run != temp; run = run->next)
	/* this loop marks all edges around temp->origin. */
	{
	  rep = number[run->end];
	  if (rep > actual_number)
	    {
	      MARKLO(run);
	    }
	  else if (!ISMARKED(run->invers))
	    {
	      rep = rep + 2*MAXN;
	    }
	  if (rep > (*representation)) return 0;
	  if (rep < (*representation)) return 2;
	  representation++;
	}
      /* check whether representation[] is also at the end of a cyclic list */
      if ((*representation) != 0) return 2;
      representation++;
      /* Next vertex to explore: */
      temp = startedge[actual_number];  actual_number++;
    }

  return 1;
}

/*****************************************************************************/

static int
testcanon_mirror(EDGE *givenedge, int representation[], int colour[])

/* Tests whether starting from a given edge and constructing the code in
   "->prev" direction, an automorphism or even a better representation can
   be found. Comments see testcanon -- it is exactly the same except for
   the orientation */
{
  EDGE *temp, *run;
  EDGE *startedge[MAXN+1];
  int number[MAXN], i;
  int last_number, actual_number, vertex;

  int rep;

  RESETMARKS;

  for (i = 0; i < nv; i++) number[i] = 0;
  
  number[givenedge->start] = 1;
  
  if (givenedge->start != givenedge->end)
    {
      number[givenedge->end] = 2;
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else last_number = 1;
  
  actual_number = 1;
  temp = givenedge;
  
  while (last_number < nv)
    {
      for (run = temp->prev; run != temp; run = run->prev)
	{ vertex = run->end;
	  if (!number[vertex])
	    { startedge[last_number] = run->invers;
	      last_number++; number[vertex] = last_number;
	      rep = colour[vertex]; }
	  else if (number[vertex] > actual_number)
	    {
	      rep = number[vertex];
	      MARKLO(run);
	    }
	  else if (ISMARKED(run->invers))
	    {
	      rep = number[vertex];
	    }
	  else
	    {
	      rep = number[vertex]+2*MAXN;
	    }
	  if (rep > (*representation)) return 0;
	  if (rep < (*representation)) return 2;
	  representation++;
	}
      if ((*representation) != 0) return 2;
      representation++;
      temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nv)
    {
      for (run = temp->prev; run != temp; run = run->prev)
	{
	  rep = number[run->end];
	  if (rep > actual_number)
	    {
	      MARKLO(run);
	    }
	  else if (!ISMARKED(run->invers))
	    {
	      rep = rep + 2*MAXN;
	    }
	  if (rep > (*representation)) return 0;
	  if (rep < (*representation)) return 2;
	  representation++;
	}
      if ((*representation) != 0) return 2;
      representation++;
      temp = startedge[actual_number];  actual_number++;
    }

  return 1;
}

/****************************************************************************/
static void
testcanon_first_init(EDGE *givenedge, int representation[], int colour[])

/* Tests whether starting from a given edge and constructing the code in
   "->next" direction, an automorphism or even a better representation can
   be found. A better representation will be completely constructed and
   returned in "representation".  It works pretty similar to testcanon except
   for obviously necessary changes, so for extensive comments see testcanon */
{
  register EDGE *run;
  register int vertex;
  EDGE *temp;
  EDGE *startedge[MAXN+1];
  int number[MAXN], i;
  int last_number, actual_number;

  RESETMARKS;

  for (i = 0; i < nv; i++) number[i] = 0;

  number[givenedge->start] = 1;
  if (givenedge->start != givenedge->end)
    {
      number[givenedge->end] = 2;
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else last_number = 1;

  actual_number = 1;
  temp = givenedge;

  while (last_number < nv)
    {
      for (run = temp->next; run != temp; run = run->next)
	{ vertex = run->end;
	if (!number[vertex])
	  { startedge[last_number] = run->invers;
	  last_number++; number[vertex] = last_number;
	  *representation = colour[vertex]; }
	else if (number[vertex] > actual_number)
	  {
	    *representation = number[vertex];
	    MARKLO(run);
	  }
	else if (ISMARKED(run->invers))
	  {
	    *representation = number[vertex];
	  }
	else
	  {
	    *representation = number[vertex]+2*MAXN;
	  }
	representation++;
	}
      *representation = 0;
      representation++;
      temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nv)
    {
      for (run = temp->next; run != temp; run = run->next)
	{
	  vertex = run->end;
	  if (number[vertex] > actual_number)
	    {
	      *representation = number[vertex];
	      MARKLO(run);
	    }
	  else if (ISMARKED(run->invers))
	    {
	      *representation = number[vertex];
	    } 
	  else
	    {
	      *representation = number[vertex]+2*MAXN;
	    } 
	  representation++;
	}
      *representation = 0;
      representation++;
      temp = startedge[actual_number];  actual_number++;
    }

  return;
}

/****************************************************************************/

static void
testcanon_first_init_mirror(EDGE *givenedge, int representation[],
			    int colour[])

/* Tests whether starting from a given edge and constructing the code in
   "->prev" direction, an automorphism or even a better representation can
   be found. A better representation will be completely constructed and
   returned in "representation".  It works pretty similar to testcanon except
   for obviously necessary changes, so for extensive comments see testcanon */
{
  register EDGE *run;
  register int vertex;
  EDGE *temp;
  EDGE *startedge[MAXN+1];
  int number[MAXN], i;
  int last_number, actual_number;

  RESETMARKS;

  for (i = 0; i < nv; i++) number[i] = 0;

  number[givenedge->start] = 1;
  if (givenedge->start != givenedge->end)
    {
      number[givenedge->end] = 2;
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else last_number = 1;

  actual_number = 1;
  temp = givenedge;

  while (last_number < nv)
    {
      for (run = temp->prev; run != temp; run = run->prev)
	{ vertex = run->end;
	if (!number[vertex])
	  { startedge[last_number] = run->invers;
	  last_number++; number[vertex] = last_number;
	  *representation = colour[vertex]; }
	else if (number[vertex] > actual_number)
	  {
	    *representation = number[vertex];
	    MARKLO(run);
	  }
	else if (ISMARKED(run->invers))
	  {
	    *representation = number[vertex];
	  }
	else
	  {
	    *representation = number[vertex]+2*MAXN;
	  }
	representation++;
	}
      *representation = 0;
      representation++;
      temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nv)
    {
      for (run = temp->prev; run != temp; run = run->prev)
	{
	  vertex = run->end;
	  if (number[vertex] > actual_number)
	    {
	      *representation = number[vertex];
	      MARKLO(run);
	    }
	  else if (ISMARKED(run->invers))
	    {
	      *representation = number[vertex];
	    } 
	  else
	    {
	      *representation = number[vertex]+2*MAXN;
	    } 
	  representation++;
	}
      *representation = 0;
      representation++;
      temp = startedge[actual_number];  actual_number++;
    }

  return;
}

/****************************************************************************/

static int
testcanon_init(EDGE *givenedge, int representation[], int colour[])

/* Tests whether starting from a given edge and constructing the code in
   "->next" direction, an automorphism or even a better representation can
   be found. A better representation will be completely constructed and
   returned in "representation".  It works pretty similar to testcanon except
   for obviously necessary changes, so for extensive comments see testcanon */
{
  register EDGE *run;
  register int vertex;
  EDGE *temp;
  EDGE *startedge[MAXN+1];
  int number[MAXN], i;
  int better = FALSE; /* is the representation already better ? */
  int last_number, actual_number;
  int rep;

  RESETMARKS;

  for (i = 0; i < nv; i++) number[i] = 0;

  number[givenedge->start] = 1;
  if (givenedge->start != givenedge->end)
    {
      number[givenedge->end] = 2;
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else last_number = 1;

  actual_number = 1;
  temp = givenedge;

  while (last_number < nv)
    {
      for (run = temp->next; run != temp; run = run->next)
	{ vertex = run->end;
	if (!number[vertex])
	  { startedge[last_number] = run->invers;
	  last_number++; number[vertex] = last_number;
	  rep = colour[vertex]; }
	else if (number[vertex] > actual_number)
	  {
	    rep = number[vertex];
	    MARKLO(run);
	  }
	else if (ISMARKED(run->invers))
	  {
	    rep = number[vertex];
	  }
	else
	  {
	    rep = number[vertex]+2*MAXN;
	  }
	if (better) *representation = rep;
	else { if (rep > (*representation)) return 0;
	else if (rep < (*representation))
	  { better = TRUE; *representation = rep; }
	}
	representation++;
	}
      if ((*representation) != 0) { better = TRUE; *representation = 0; }
      representation++;
      temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nv)
    {
      for (run = temp->next; run != temp; run = run->next)
	{ 
	  rep = number[run->end];
	  if (rep > actual_number)
	    {
	      MARKLO(run);
	    }
	  else if (!ISMARKED(run->invers))
	    {
	      rep = rep + 2*MAXN;
	    }
	  if (better) *representation = rep;
	  else
	    {
	      if (rep > (*representation)) return 0;
	      if (rep < (*representation))
		{ better = TRUE; *representation = rep; }
	    }
	  representation++;
	}
      if ((*representation) != 0) { better = TRUE; *representation = 0; }
      representation++;
      temp = startedge[actual_number];  actual_number++;
    }

  if (better) return 2;
  return 1;
}

/****************************************************************************/

static int
testcanon_mirror_init(EDGE *givenedge, int representation[], int colour[])

/* Tests whether starting from a given edge and constructing the code in
   "->prev" direction, an automorphism or even a better representation can
   be found. A better representation will be completely constructed and
   returned in "representation".  It works pretty similar to testcanon except
   for obviously necessary changes, so for extensive comments see testcanon */
{
  EDGE *temp, *run;
  EDGE *startedge[MAXN+1];
  int number[MAXN], i;
  int better = FALSE; /* is the representation already better ? */
  int last_number, actual_number, vertex;
  int rep;

  RESETMARKS;

  for (i = 0; i < nv; i++) number[i] = 0;

  number[givenedge->start] = 1;
  if (givenedge->start != givenedge->end)
    {
      number[givenedge->end] = 2;
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else last_number = 1;

  actual_number = 1;
  temp = givenedge;

  while (last_number < nv)
    {
      for (run = temp->prev; run != temp; run = run->prev)
	{ vertex = run->end;
	  if (!number[vertex])
	    { startedge[last_number] = run->invers;
	      last_number++; number[vertex] = last_number;
	      rep = colour[vertex]; }
	  else if (number[vertex] > actual_number)
	    {
	      rep = number[vertex];
	      MARKLO(run);
	    }
	  else if (ISMARKED(run->invers))
	    {
	      rep = number[vertex];
	    }
	  else
	    {
	      rep = number[vertex]+2*MAXN;
	    }
	  if (better) *representation = rep;
	  else { if (rep > (*representation)) return 0;
	    else if (rep < (*representation))
	      { better = TRUE; *representation = rep; }
	  }
	  representation++;
	}
      if ((*representation) != 0) { better = TRUE; *representation = 0; }
      representation++;
      temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nv)
    {
      for (run = temp->prev; run != temp; run = run->prev)
	{ 
	  rep = number[run->end];
	  if (rep > actual_number)
	    {
	      MARKLO(run);
	    }
	  else if (!ISMARKED(run->invers))
	    {
	      rep = rep + 2*MAXN;
	    }
	  if (better) *representation = rep;
	  else
	    {
	      if (rep > (*representation)) return 0;
	      if (rep < (*representation))
		{ better = TRUE; *representation = rep; }
	    }
	  representation++;
	}
      if ((*representation) != 0) { better = TRUE; *representation = 0; }
      representation++;
      temp = startedge[actual_number];  actual_number++;
    }

  if (better) return 2;
  return 1;
}

/****************************************************************************/

static void
construct_numb(EDGE *givenedge, EDGE *numbering[])

/* Starts at givenedge and writes the edges in the well defined order
   into the list.  Works like testcanon. Look there for comments. */
{
  EDGE *temp, **tail, **limit, *run;
  EDGE *startedge[MAXN+1];
  int last_number, actual_number, vertex;

  RESETMARKS_V;

  tail = numbering;
  limit = numbering+ne-1;

  MARK_V(givenedge->start);
  if (givenedge->start != givenedge->end)
    {
      MARK_V(givenedge->end);
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else last_number = 1;

  actual_number = 1;
  temp = *tail = givenedge;

  while (last_number < nv)
    {
      for (run = temp->next; run != temp; run = run->next)
	/* this loop marks all edges around temp->origin. */
	{ vertex = run->end;
	if (!ISMARKED_V(vertex))
	  { startedge[last_number] = run->invers;
	  last_number++; MARK_V(vertex); }
	tail++; *tail = run;
	}
      if (tail != limit)
	{ tail++;
	*tail = temp = startedge[actual_number];  actual_number++; }
    }

  while (tail != limit)
    /* Now we know that all numbers have been given */
    {
      for (run = temp->next; run != temp; run = run->next)
	/* this loop marks all edges around temp->origin. */
	{ tail++; *tail = run; }
      if (tail != limit)
        {
	  /* Next vertex to explore: */
	  tail++;
	  *tail = temp = startedge[actual_number];  actual_number++; }
    }
}

/****************************************************************************/

static void
construct_numb_mirror(EDGE *givenedge, EDGE *numbering[])

/* Starts at givenedge and writes the edges in the well defined order
   into the list.  Works like testcanon. Look there for comments.  */
{
  EDGE *temp, **tail, **limit, *run;
  EDGE *startedge[MAXN+1];
  int last_number, actual_number, vertex;

  RESETMARKS_V;

  tail = numbering; /* The first entry of the numbering list */
  limit = numbering+ne-1;  /* Last valid entry of the numbering list */

  MARK_V(givenedge->start);
  if (givenedge->start != givenedge->end)
    {
      MARK_V(givenedge->end);
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else last_number = 1;

  actual_number = 1;
  temp = *tail = givenedge;

  while (last_number < nv)
    {
      for (run = temp->prev; run != temp; run = run->prev)
	/* this loop marks all edges around temp->origin. */
	{ vertex = run->end;
	if (!ISMARKED_V(vertex))
	  { startedge[last_number] = run->invers;
	  last_number++; MARK_V(vertex); }
	tail++; *tail = run;
	}
      if (tail != limit)
	{
	  tail++;
	  *tail = temp = startedge[actual_number];  actual_number++; }
    }

  while (tail != limit)
    /* Now we know that all numbers have been given */
    {
      for (run = temp->prev; run != temp; run = run->prev)
	/* this loop marks all edges around temp->origin. */
	{ tail++; *tail = run; }
      if (tail != limit)
        {
	  /* Next vertex to explore: */
	  tail++;
	  *tail = temp = startedge[actual_number];  actual_number++; }
    }
}

/****************************************************************************/

static int
canon(int lcolour[], EDGE *can_numberings[][MAXE],
      int *num_can_numberings, int *num_can_numberings_or_pres)

/* Checks whether the last vertex (number: nv-1) is canonical or not.
   Returns 1 if yes, 0 if not. One of the criterions a canonical vertex
   must fulfill is that its colour is minimal (among the contractible
   vertices if any).  If the triangulation is reducible then the canonical
   edge (can_numberings[0][0]) must be contractible.

   IT IS ASSUMED that the values of the colour function are positive
   and at most INT_MAX-MAXN.

   A possible starting edge for the construction of a representation is
   one with lexicographically minimal colour pair (start,INT_MAX-end).
   In can_numberings[][] the canonical numberings are stored as sequences
   of oriented edges.  For every 0 <= i,j < *num_can_numberings and every
   0 <= k < ne the edges can_numberings[i][k] and can_numberings[j][k] can
   be mapped onto each other by an automorphism. The first
   *num_can_numberings_or_pres numberings are orientation preserving while
   the rest is orientation reversing.

   In case of only 1 automorphism, in can_numberings[0][0] the "canonical"
   edge is given.  It is one edge emanating at the canonical vertex. The
   rest of the numbering is not given.

   In case of nontrivial automorphisms, can[0] starts with a list of edges
   adjacent to nv-1. In case of an orientation preserving numbering the edges
   are listed in ->next direction, otherwise in ->prev direction.

   Works OK if at least one vertex has valence >= 3. Otherwise some numberings
   are computed twice, since changing the orientation (the cyclic order around
   each vertex) doesn't change anything

*/

{
  int i, last_vertex, test;
  int minstart, maxend; /* (minstart,maxend) will be the chosen colour
			   pair of an edge */
  EDGE *startlist_last[MAXN-1], *startlist[(MAXN-1)*(MAXN-1)], *run, *end;
  int list_length_last, list_length;
  int representation[MAXE];
  EDGE *numblist[MAXE], *numblist_mirror[MAXE]; /* lists of edges where
						   starting gives a canonical
						   representation */
  int numbs = 1, numbs_mirror = 0;
  int colour[MAXN];
  int irreducible;

  irreducible = TRUE;

  /* without this, irreducible is always TRUE and 
     canonical edge is not guarenteed to be contractible.
     but if min degree is 3 then canonical edge will be from 
     degree 3 vertex and so contractible.
     why is this done? */
  
  for (i=0; i<nv-1 && irreducible; i++)
    if (ncontract[i]>0) irreducible = FALSE;
  

  for (i=0; i<nv; i++) colour[i]=lcolour[i]+MAXN;
  /* to distinguish colours from vertices */
  last_vertex = nv-1;
  if (!irreducible && ncontract[last_vertex]==0) return 0;
  minstart = colour[last_vertex];

  /* determine the smallest possible end for the vertex "last_vertex" */

  list_length_last = 1; startlist_last[0] = end = firstedge[last_vertex];
  if (irreducible || nnei[end->start][end->end] == 2)
    maxend = colour[end->end];
  else
    maxend = 0;

  for (run = end->next; run != end; run = run->next)
    if (irreducible || nnei[run->start][run->end] == 2)
      { if (colour[run->end] > maxend)
	{ startlist_last[0] = run;
	list_length_last = 1; maxend = colour[run->end];}
      else if (colour[run->end] == maxend)
	{ startlist_last[list_length_last] = run; list_length_last++; }
      }

  /* Now we know the pair that SHOULD be minimal and we can determine a list
     of all edges with this colour pair. If a new pair is found that is even
     smaller, we can return 0 at once */

  list_length = 0;

  for (i = 0; i < last_vertex; i++)
    if (irreducible || ncontract[i]>0)
      { if (colour[i] < minstart) return 0;
      if (colour[i] == minstart)
	{ run = end = firstedge[i];
	do
	  {
	    if (irreducible || nnei[run->start][run->end] == 2)
	      {
		if (colour[run->end] > maxend) return 0;
		if (colour[run->end] == maxend)
		  { startlist[list_length] = run; list_length++; }
	      }
	    run = run->next;
	  } while (run != end);
	}
      }

  /* OK -- so there is no smaller pair and now we have to determine the
     smallest representation around vertex "last_vertex": */

  testcanon_first_init(startlist_last[0], representation, colour);
  numblist[0] = startlist_last[0];
  test = testcanon_mirror_init(startlist_last[0], representation, colour);
  if (test == 1)
    { numbs_mirror = 1; numblist_mirror[0] = startlist_last[0]; }
  else if (test == 2)
    { numbs_mirror = 1; numbs = 0;
    numblist_mirror[0] = startlist_last[0]; }

  for (i = 1; i < list_length_last; i++)
    {
      test = testcanon_init(startlist_last[i], representation, colour);
      if (test == 1) { numblist[numbs] = startlist_last[i]; numbs++; }
      else if (test == 2)
	{ numbs_mirror = 0; numbs = 1; numblist[0] = startlist_last[i]; }
      test = testcanon_mirror_init(startlist_last[i],
				   representation, colour);
      if (test == 1)
	{ numblist_mirror[numbs_mirror] = startlist_last[i];
	numbs_mirror++; }
      else if (test == 2)
	{ numbs_mirror = 1; numbs = 0;
	numblist_mirror[0] = startlist_last[i]; }
    }

  /* Now we know the best representation we can obtain starting at last_vertex.
     Now we will check all the others. We can return 0 at once if we find a
     better one */

  for (i = 0; i < list_length; i++)
    { test = testcanon(startlist[i], representation, colour);
    if (test == 1) { numblist[numbs] = startlist[i]; numbs++; }
    else if (test == 2) return 0;
    test = testcanon_mirror(startlist[i], representation, colour);
    if (test == 1)
      { numblist_mirror[numbs_mirror] = startlist[i]; numbs_mirror++; }
    else if (test == 2) return 0;
    }

  *num_can_numberings_or_pres = numbs;
  *num_can_numberings = numbs+numbs_mirror;

  if (*num_can_numberings>1)
    { for (i = 0; i < numbs; i++)
      construct_numb(numblist[i], can_numberings[i]);
    for (i = 0; i < numbs_mirror; i++, numbs++)
      construct_numb_mirror(numblist_mirror[i],can_numberings[numbs]);
    }
  else
    { if (numbs) can_numberings[0][0] = numblist[0];
    else can_numberings[0][0] = numblist_mirror[0]; }

  return 1;
}

/****************************************************************************/

static int
canon_quad(int lcolour[], EDGE *can_numberings[][MAXE],
      int *num_can_numberings, int *num_can_numberings_or_pres)

/* Checks whether the last vertex (number: nv-1) is canonical or not.
   Returns 1 if yes, 0 if not. One of the criterions a canonical vertex
   must fulfill is that its colour is minimal (among the contractible
   vertices if any).  If the quadrangulation is reducible then the canonical
   edge (can_numberings[0][0]) must be contractible.

   IT IS ASSUMED that the values of the colour function are positive
   and at most INT_MAX-MAXN.

   A possible starting edge for the construction of a representation is
   one with lexicographically minimal colour pair (start,INT_MAX-end).
   In can_numberings[][] the canonical numberings are stored as sequences
   of oriented edges.  For every 0 <= i,j < *num_can_numberings and every
   0 <= k < ne the edges can_numberings[i][k] and can_numberings[j][k] can
   be mapped onto each other by an automorphism. The first
   *num_can_numberings_or_pres numberings are orientation preserving while
   the rest is orientation reversing.

   In case of only 1 automorphism, in can_numberings[0][0] the "canonical"
   edge is given.  It is one edge emanating at the canonical vertex. The
   rest of the numbering is not given.

   In case of nontrivial automorphisms, can[0] starts with a list of edges
   adjacent to nv-1. In case of an orientation preserving numbering the edges
   are listed in ->next direction, otherwise in ->prev direction.

   Works OK if at least one vertex has valence >= 3. Otherwise some numberings
   are computed twice, since changing the orientation (the cyclic order around
   each vertex) doesn't change anything

*/

{
  int i, last_vertex, test;
  int minstart, maxend; /* (minstart,maxend) will be the chosen colour
			   pair of an edge */
  EDGE *startlist_last[MAXN-1], *startlist[(MAXN-1)*(MAXN-1)], *run, *end;
  int list_length_last, list_length;
  int representation[MAXE];
  EDGE *numblist[MAXE], *numblist_mirror[MAXE]; /* lists of edges where
						   starting gives a canonical
						   representation */
  int numbs = 1, numbs_mirror = 0;
  int colour[MAXN];
  int irreducible;

  irreducible = TRUE;

  /* without this, irreducible is always TRUE and 
     canonical edge is not guarenteed to be contractible.
     but if min degree is 3 then canonical edge will be from 
     degree 3 vertex and so contractible.
     why is this done? */
  
  for (i=0; i<nv-1 && irreducible; i++)
    if (ncontract[i]>0) irreducible = FALSE;
  

  for (i=0; i<nv; i++) colour[i]=lcolour[i]+MAXN;
  /* to distinguish colours from vertices */
  last_vertex = nv-1;
  if (!irreducible && ncontract[last_vertex]==0) return 0;
  minstart = colour[last_vertex];

  /* determine the smallest possible end for the vertex "last_vertex" */

  list_length_last = 1; startlist_last[0] = end = firstedge[last_vertex];
  if (irreducible || 
      (nnei[end->start][end->invers->prev->end] == 2 &&
       !adj[end->start][end->invers->prev->end]))
    maxend = colour[end->end];
  else
    maxend = 0;

  for (run = end->next; run != end; run = run->next)
    if (irreducible || 
	(nnei[run->start][run->invers->prev->end] == 2 &&
	 !adj[end->start][end->invers->prev->end]))
      { if (colour[run->end] > maxend)
	{ startlist_last[0] = run;
	list_length_last = 1; maxend = colour[run->end];}
      else if (colour[run->end] == maxend)
	{ startlist_last[list_length_last] = run; list_length_last++; }
      }

  /* Now we know the pair that SHOULD be minimal and we can determine a list
     of all edges with this colour pair. If a new pair is found that is even
     smaller, we can return 0 at once */

  list_length = 0;

  for (i = 0; i < last_vertex; i++)
    if (irreducible || ncontract[i]>0)
      { if (colour[i] < minstart) return 0;
      if (colour[i] == minstart)
	{ run = end = firstedge[i];
	do
	  {
	    if (irreducible || 
		(nnei[run->start][run->invers->prev->end] == 2 &&
		 !adj[end->start][end->invers->prev->end]))
	      {
		if (colour[run->end] > maxend) return 0;
		if (colour[run->end] == maxend)
		  { startlist[list_length] = run; list_length++; }
	      }
	    run = run->next;
	  } while (run != end);
	}
      }

  /* OK -- so there is no smaller pair and now we have to determine the
     smallest representation around vertex "last_vertex": */

  testcanon_first_init(startlist_last[0], representation, colour);
  numblist[0] = startlist_last[0];
  test = testcanon_mirror_init(startlist_last[0], representation, colour);
  if (test == 1)
    { numbs_mirror = 1; numblist_mirror[0] = startlist_last[0]; }
  else if (test == 2)
    { numbs_mirror = 1; numbs = 0;
    numblist_mirror[0] = startlist_last[0]; }

  for (i = 1; i < list_length_last; i++)
    {
      test = testcanon_init(startlist_last[i], representation, colour);
      if (test == 1) { numblist[numbs] = startlist_last[i]; numbs++; }
      else if (test == 2)
	{ numbs_mirror = 0; numbs = 1; numblist[0] = startlist_last[i]; }
      test = testcanon_mirror_init(startlist_last[i],
				   representation, colour);
      if (test == 1)
	{ numblist_mirror[numbs_mirror] = startlist_last[i];
	numbs_mirror++; }
      else if (test == 2)
	{ numbs_mirror = 1; numbs = 0;
	numblist_mirror[0] = startlist_last[i]; }
    }

  /* Now we know the best representation we can obtain starting at last_vertex.
     Now we will check all the others. We can return 0 at once if we find a
     better one */

  for (i = 0; i < list_length; i++)
    { test = testcanon(startlist[i], representation, colour);
    if (test == 1) { numblist[numbs] = startlist[i]; numbs++; }
    else if (test == 2) return 0;
    test = testcanon_mirror(startlist[i], representation, colour);
    if (test == 1)
      { numblist_mirror[numbs_mirror] = startlist[i]; numbs_mirror++; }
    else if (test == 2) return 0;
    }

  *num_can_numberings_or_pres = numbs;
  *num_can_numberings = numbs+numbs_mirror;

  if (*num_can_numberings>1)
    { for (i = 0; i < numbs; i++)
      construct_numb(numblist[i], can_numberings[i]);
    for (i = 0; i < numbs_mirror; i++, numbs++)
      construct_numb_mirror(numblist_mirror[i],can_numberings[numbs]);
    }
  else
    { if (numbs) can_numberings[0][0] = numblist[0];
    else can_numberings[0][0] = numblist_mirror[0]; }

  return 1;
}

/****************************************************************************/

static int 
canon_edge_oriented(EDGE *edgelist_or[], int num_edges_or, int can_edges_or,
		    EDGE *edgelist_inv[], int num_edges_inv, int can_edges_inv,
		    int lcolour[], EDGE *can_numberings[][MAXE], 
		    int *num_can_numberings, int *num_can_numberings_or_pres)

/* 
   IT IS ASSUMED that the values of the colour function are positive
   and at most INT_MAX-MAXN.

   This routine checks all num_edges_or elements of edgelist_or just for one 
   orientation and all num_edges_inv elements of the list edgelist_inv just
   for the other. It returns 1 if and only if one of the first can_edges_or
   elements of the first list or first can_edges_inv elements of the second 
   give an optimal numbering among all the possibilities provided by the
   lists.

   Edges given are not in minimal form -- but it is guaranteed that all
   colours of the startpoints are the same and all colours of the endpoints
   are the same.

   In case of only the identity automorphism, the entries of can_numberings[][]
   are undefined.

   Otherwise in can_numberings[][] the canonical numberings are stored as
   sequences of oriented edges.  For every 0 <= i,j < *num_can_numberings
   and every 0 <= k < ne the edges can_numberings[i][k] and
   can_numberings[j][k] can be mapped onto each other by an automorphism.
   The first *num_can_numberings_or_pres numberings are orientation
   preserving while the rest are orientation reversing.

   In case of an orientation preserving numbering the edges are listed in 
   ->next direction, otherwise in ->prev direction.

   Works OK if at least one vertex has valence >= 3. Otherwise some numberings 
   are computed twice, since changing the orientation (the cyclic order around 
   each vertex) doesn't change anything */
{
    int i, test;
    int representation[MAXE];
    EDGE *numblist[MAXE], *numblist_mirror[MAXE]; /* lists of edges where 
                            starting gives a canonical representation */
    int numbs = 1, numbs_mirror = 0;
    int colour[MAXN];

    for (i=0; i<nv; i++) colour[i]=lcolour[i]+MAXN;
			     /* to distinguish colours from vertices */

/* First we have to determine the smallest representation possible with
   edgelist_or */

    if (can_edges_or > 0)
      { testcanon_first_init(edgelist_or[0], representation, colour);
	numblist[0] = edgelist_or[0];
	for (i=1; i<can_edges_or; i++)
	  { test = testcanon_init(edgelist_or[i], representation, colour);
	    if (test == 1) { numblist[numbs] = edgelist_or[i]; numbs++; }
	    else if (test == 2) 
	      { numbs = 1; numblist[0] = edgelist_or[i]; }
	  }
	i=0; /* the next for-loop can start at the beginning */
      }
    else
      { numbs=0; numbs_mirror=1;
        testcanon_first_init_mirror(edgelist_inv[0], representation, colour);
        numblist_mirror[0] = edgelist_inv[0];
	i=1; /* the next for-loop must start at position 1 */
      }

    for (   ; i<can_edges_inv; i++)
      { test = testcanon_mirror_init(edgelist_inv[i], representation, colour);
        if (test == 1)
          { numblist_mirror[numbs_mirror] = edgelist_inv[i]; numbs_mirror++; }
	else if (test == 2) 
	  { numbs = 0; numbs_mirror=1; numblist_mirror[0] = edgelist_inv[i]; }
      }


    /* now we know the best we can get for a "canonical edge".
       Next we will check all the others. We can return 0 at once if we find a 
       better one */

    for (i=can_edges_or ; i < num_edges_or; i++)
      { test = testcanon(edgelist_or[i], representation, colour);
        if (test == 1) { numblist[numbs] = edgelist_or[i]; numbs++; }
        else if (test == 2) return 0;
      }
    for (i=can_edges_inv ; i < num_edges_inv; i++)
      { test = testcanon_mirror(edgelist_inv[i], representation, colour);
        if (test == 1) 
          { numblist_mirror[numbs_mirror] = edgelist_inv[i]; numbs_mirror++; }
        else if (test == 2) return 0;
      }

    *num_can_numberings_or_pres = numbs;
    *num_can_numberings = numbs+numbs_mirror;

    if (*num_can_numberings > 1)
      {
        for (i = 0; i < numbs; i++) 
            construct_numb(numblist[i], can_numberings[i]); 
        for (i = 0; i < numbs_mirror; i++, numbs++) 
            construct_numb_mirror(numblist_mirror[i],can_numberings[numbs]);
      }

    return 1;
}

/**************************************************************************/

static void
initialize_min4()

/* initialize edges for mindegree=4 generation */
{
    EDGE *run,*start,*orun,*ostart;
    int i,j,n;

/* First initialise the edges for the four-operation. */

    for (n=6; n<=MAXN; n++)
      { start=four_op(n);
        ostart=start+NUMEDGES;
        for (i=0, run=start, orun=ostart; i<6; i+=2)
          { run->start=n; run->invers=run+1; run++;
	    orun->start=n; orun->invers=orun+1; orun++;
            run->end=n; run->invers=run-1; run++; 
	    orun->end=n; orun->invers=orun-1; orun++; }
        run=start; run->next=run+2; 
	orun=ostart; orun->prev=orun+2;
        run+=2; /*2*/ run->prev=start; run->next=run+2;
	orun+=2;      orun->next=ostart; orun->prev=orun+2;
        run+=2; /*4*/ run->prev=start+2; 
	orun+=2;      orun->next=ostart+2;
      }

/* Then initialise the edges for the five-operation. */

    for (n=6; n<=MAXN; n++)
      { start=five_op(n);
        ostart=start+NUMEDGES;
        for (i=0, run=start, orun=ostart; i<6; i+=2)
          { run->start=n; run->invers=run+1; run++;
	    orun->start=n; orun->invers=orun+1; orun++;
            run->end=n; run->invers=run-1; run++; 
	    orun->end=n; orun->invers=orun-1; orun++; }
        run=start; run->next=run+2; 
	orun=ostart; orun->prev=orun+2;
        run+=2; /*2*/ run->prev=start;
	orun+=2;      orun->next=ostart;
      }

/* Then initialise the edges for the S-operation. */

    for (n=6; n<=MAXN-3; n++)
      { start=S_op(n);
        ostart=start+NUMEDGES;
        for (i=0;i<18;i+=2)
          { run=start+i; run->invers=run+1;
	    orun=ostart+i; orun->invers=orun+1;
            run++; run->invers=run-1; 
	    orun++; orun->invers=orun-1; }

        run=start;    run->start=n+1; run->end=n; run->next=start+4;
                      run->prev=start+9;
        run++; /*1*/  run->start=n; run->end=n+1; run->next=start+11;
                      run->prev=start+3;
        run++; /*2*/  run->start=n+2; run->end=n; run->next=start+15;
                      run->prev=start+5;
        run++; /*3*/  run->start=n; run->end=n+2; run->next=start+1;
                      run->prev=start+13;
        run++; /*4*/  run->start=n+1; run->end=n+2; run->next=start+7;
                      run->prev=start;
        run++; /*5*/  run->start=n+2; run->end=n+1; run->next=start+2;
                      run->prev=start+17;
        run++; /*6*/  run->end=n+1; run->next=start+16; 
        run++; /*7*/  run->start=n+1; run->next=start+9; run->prev=start+4;
        run++; /*8*/  run->end=n+1; run->prev=start+10;
        run++; /*9*/  run->start=n+1; run->next=start; run->prev=start+7;
        run++; /*10*/ run->end=n; run->next=start+8; 
        run++; /*11*/ run->start=n;  run->next=start+13; run->prev=start+1;
        run++; /*12*/ run->end=n; run->prev=start+14;
        run++; /*13*/ run->start=n; run->next=start+3; run->prev=start+11;
        run++; /*14*/ run->end=n+2; run->next=start+12;
        run++; /*15*/ run->start=n+2; run->next=start+17; run->prev=start+2;
        run++; /*16*/ run->end=n+2; run->prev=start+6;
        run++; /*17*/ run->start=n+2; run->next=start+5; run->prev=start+15;

        orun=ostart;    orun->start=n+1; orun->end=n; orun->prev=ostart+4;
                      orun->next=ostart+9;
        orun++; /*1*/  orun->start=n; orun->end=n+1; orun->prev=ostart+11;
                      orun->next=ostart+3;
        orun++; /*2*/  orun->start=n+2; orun->end=n; orun->prev=ostart+15;
                      orun->next=ostart+5;
        orun++; /*3*/  orun->start=n; orun->end=n+2; orun->prev=ostart+1;
                      orun->next=ostart+13;
        orun++; /*4*/  orun->start=n+1; orun->end=n+2; orun->prev=ostart+7;
                      orun->next=ostart;
        orun++; /*5*/  orun->start=n+2; orun->end=n+1; orun->prev=ostart+2;
                      orun->next=ostart+17;
        orun++; /*6*/  orun->end=n+1; orun->prev=ostart+16; 
        orun++; /*7*/  orun->start=n+1; orun->prev=ostart+9; orun->next=ostart+4;
        orun++; /*8*/  orun->end=n+1; orun->next=ostart+10;
        orun++; /*9*/  orun->start=n+1; orun->prev=ostart; orun->next=ostart+7;
        orun++; /*10*/ orun->end=n; orun->prev=ostart+8; 
        orun++; /*11*/ orun->start=n;  orun->prev=ostart+13; orun->next=ostart+1;
        orun++; /*12*/ orun->end=n; orun->next=ostart+14;
        orun++; /*13*/ orun->start=n; orun->prev=ostart+3; orun->next=ostart+11;
        orun++; /*14*/ orun->end=n+2; orun->prev=ostart+12;
        orun++; /*15*/ orun->start=n+2; orun->prev=ostart+17; orun->next=ostart+2;
        orun++; /*16*/ orun->end=n+2; orun->next=ostart+6;
        orun++; /*17*/ orun->start=n+2; orun->prev=ostart+5; orun->next=ostart+15;
      }

  for (j = 0; j < NUMEDGES; ++j)
    {
      edges[j].op = edges + NUMEDGES + j;
      edges[j+NUMEDGES].op = edges + j;
    }

}

/****************************************************************************/

static void
initialize()

/* initialize stars */
{
  register int i,j;
  register EDGE *si;
  register EDGE *oi;

  for (i = 0; i < MAXN; ++i)
    {
      si = STAR3(i);
      oi = si + NUMEDGES;

      for (j = 0; j < 3; ++j)
        {
	  si[j].end = si[j+3].start = i;
	  si[j].invers = si + 3 + j;
	  si[j+3].invers = si + j;
	  si[j+3].next = si + 3 + (j+1) % 3;
	  si[j+3].prev = si + 3 + (j+2) % 3;
	  oi[j].end = oi[j+3].start = i;
	  oi[j].invers = oi + 3 + j;
	  oi[j+3].invers = oi + j;
	  oi[j+3].prev = oi + 3 + (j+1) % 3;
	  oi[j+3].next = oi + 3 + (j+2) % 3;
        }

      si = STAR4(i);
      oi = si + NUMEDGES;

      for (j = 0; j < 4; ++j)
        {
	  si[j].end = si[j+4].start = i;
	  si[j].invers = si + 4 + j;
	  si[j+4].invers = si + j;
	  si[j+4].next = si + 4 + (j+1) % 4;
	  si[j+4].prev = si + 4 + (j+3) % 4;
	  oi[j].end = oi[j+4].start = i;
	  oi[j].invers = oi + 4 + j;
	  oi[j+4].invers = oi + j;
	  oi[j+4].prev = oi + 4 + (j+1) % 4;
	  oi[j+4].next = oi + 4 + (j+3) % 4;
        }

      si = STARN(i);
      oi = si + NUMEDGES;

      for (j = 0; j < MAXN; ++j)
        {
	  si[j].end = si[j+MAXN].start = i;
	  si[j].invers = si + MAXN + j;
	  si[j+MAXN].invers = si + j;
	  si[j+MAXN].next = si + MAXN + (j+1) % MAXN;
	  si[j+MAXN].prev = si + MAXN + (j+MAXN-1) % MAXN;
	  oi[j].end = oi[j+MAXN].start = i;
	  oi[j].invers = oi + MAXN + j;
	  oi[j+MAXN].invers = oi + j;
	  oi[j+MAXN].prev = oi + MAXN + (j+1) % MAXN;
	  oi[j+MAXN].next = oi + MAXN + (j+MAXN-1) % MAXN;
        }
    }
  for (j = 0; j < NUMEDGES; ++j)
    {
      edges[j].op = edges + NUMEDGES + j;
      edges[j+NUMEDGES].op = edges + j;
    }
}

/**************************************************************************/

static int
get_newend(char code, int *flip_edge)

/* get next end vertex number from irreducible triangulation file */
/* if no more neibors return -1 */
/* if end of line return -2 */

{
  *flip_edge = FALSE;
  if (code == ',' || code == ' ')
    return -1;
  else if (code == '\0' || code == '\n')
    return -2;
  else if (code >= 'A' && code <= 'Z') {
    *flip_edge = TRUE;
    return (code - 'A');
  }
  else
    return (code - 'a');
}

/**************************************************************************/

static void
initialize_irr(char *inputline, int expect_ne)

/* create initial irreducible triangulation */
/* assumes representation has canonical vertex first.
   reverses order of vertices so that canonical vertex is last.
   then canon() will succeed and numbering[0][0] (or its inverse)
   will be the canonical edge. */
{

  int center;          /* vertex around which edges are being added */
  int newend;          /* end of new edge being added */
  EDGE *ed,*ex;
  EDGE *newedge;       /* edge being added */
  int i,j,k;
  int inchar;

  int irreducible;
  int flip_edge;

  if (nv < 10)
    inchar = 1;
  else
    inchar = 2;

  for (i=0;i<maxnv;i++)
    for (j=0;j<maxnv;j++)
      adj[i][j] = FALSE;
  for (i=0;i<maxnv;i++)
    for (j=0;j<maxnv;j++)
      nnei[i][j] = 0;
  for (i=0;i<maxnv;i++)
    degree[i] = 0;

  ne = 0;

  center = -1;
  newedge = init_edge;

  /* skip over blank */

  newend = get_newend(inputline[inchar++],&flip_edge);

  while (ne < expect_ne && newend != -2)
    {
      if (newend == -1)
	{
	  center++;
	  newend = get_newend(inputline[inchar++],&flip_edge);
	  degree[nv-1-center] = 0;
	  firstedge[nv-1-center] = newedge;
	  newedge->next = newedge->prev = newedge;
	}
      else
	{
	  newedge->next = firstedge[nv-1-center];
	  newedge->prev = firstedge[nv-1-center]->prev;
	  firstedge[nv-1-center]->prev = newedge;
	  newedge->prev->next = newedge;
	}
      newedge->start = nv-1-center;
      newedge->end = nv-1-newend;
      adj[nv-1-center][nv-1-newend] = TRUE;
      if (newend < center)
	{
	  ed = firstedge[nv-1-newend];
	  while (ed->end != nv-1-center)
	    ed = ed->next;
	  if (!flip_edge) {
	    ed->invers = newedge;
	    newedge->invers = ed;
	    ed->op->invers = newedge->op;
	    newedge->op->invers = ed->op;
	  }
	  else {
	    ed->invers = newedge->op;
	    newedge->invers = ed->op;
	    ed->op->invers = newedge;
	    newedge->op->invers = ed;
	  }
	}

      degree[nv-1-center]++;
      newedge++;
      ne++;

      newend = get_newend(inputline[inchar++],&flip_edge);
    }

  /*  */

  for (i=0;i<nv;i++) {
    ed = firstedge[i];
    for (j=0;j<degree[i];j++) {
      ed->op->start = ed->start;
      ed->op->end = ed->end;
      ed->op->next = ed->prev->op;
      ed->op->prev = ed->next->op;
      ed = ed->next;
    }
  }

  /* count common neighbors */

  for (i=0;i<nv-1;i++) {
    nnei[i][i] = 0;
    for (j=i+1;j<nv;j++)
      {
	nnei[i][j] = 0;
	for (k=0;k<nv;k++)
	  if (adj[i][k] && adj[j][k])
	    nnei[i][j]++;
	nnei[j][i] = nnei[i][j];
      }
  }

  /* add in hole border edges */
  
  if (irrhole) {
    for (i = 0; i < nv; ++i)
      {
	ed = ex = firstedge[i];
	do
	  {
	    /* check both sides of each edge only once */
	    if (ed->start < ed->end) {
	      if (RIGHTHOLE(ed))
		{
		  nnei[ed->start][ed->end]++;
		  nnei[ed->end][ed->start]++;
		}
	      if (RIGHTHOLE(ed->invers))
		{
		  nnei[ed->start][ed->end]++;
		  nnei[ed->end][ed->start]++;
		}
	    }
	    ed = ed->next;
	  } while (ed != ex);
      }
  }

  irreducible = TRUE;
  for (i=0;i<nv;i++)
    {
      ncontract[i] = 0;
      for (j=0;j<nv;j++)
	if (nnei[i][j]==2 && adj[i][j]) {
	  ncontract[i]++;
	  irreducible = FALSE;
	}
    }
  if (!irreducible &&  nv > 6 && !yswitch) {
    fprintf(stderr,">E %s: input triangulation is not irreducible\n",cmdname);
    fprintf(stderr,">E input line is %s\n",inputline);
    perror(">E ");
    exit(1);
    }

}

/**************************************************************************/

static void
update_ncontract_quad()

/* for quadrangulations update the arrary ncontract[].
   ncontract[v] = FALSE if for all faces (v,v1,v2,v3) the diagonal (v,v2) is 
   not contractable.  ncontract[v] = TRUE, otherwise.  
   if minimumdeg is 3 and degree of v1 or v3 is 3 then ncontract[v] = FALSE.
*/

{
  int v, v1, v2, v3;
  EDGE *e,*elast;
  
  if (minimumdeg == 2)

    /* for degenerate case, the square, make it irreducible. */
    
    if (nv == 4) 
      for (v=0; v<nv; v++)
	ncontract[v] = FALSE;
    else
      for (v=0; v<nv; v++) {
	ncontract[v] = FALSE;
	e = elast = firstedge[v];
	do {
	  v2 = e->invers->prev->end;
	  if (!adj[v][v2] && nnei[v][v2] == 2)
	    ncontract[v] = TRUE;
	  else
	    e = e->next;
	}
	while (!ncontract[v] && e != elast);
      }

  else /* minimumdeg == 3 */

    for (v=0; v<nv; v++) {
      ncontract[v] = FALSE;
      e = elast = firstedge[v];
      do {
	v1 = e->end;
	v2 = e->invers->prev->end;
	v3 = e->next->end;
	if (!adj[v][v2] && nnei[v][v2] == 2 &&
	    degree[v1] > 3 && degree[v3] > 3)
	  ncontract[v] = TRUE;
	else
	  e = e->next;
      }
      while (!ncontract[v] && e != elast);
    }
}

/**************************************************************************/

static void
initialize_irr_quad(char *inputline, int expect_ne)

/* create initial irreducible quadrangulation */
/* assumes representation has canonical vertex first.
   reverses order of vertices so that canonical vertex is last.
   then canon() will succeed and numbering[0][0] (or its inverse)
   will be the canonical edge. */
{

  int center;          /* vertex around which edges are being added */
  int newend;          /* end of new edge being added */
  EDGE *ed,*ed1;
  EDGE *newedge;       /* edge being added */
  int i,j,k;
  int inchar;

  int input_orient;
  //int irreducible;
  int flip_edge;  /* not used */

  if (nv < 10)
    inchar = 1;
  else
    inchar = 2;

  for (i=0;i<maxnv;i++)
    for (j=0;j<maxnv;j++) {
      adj[i][j] = FALSE;
      nnei[i][j] = 0;
    }
  for (i=0;i<maxnv;i++)
    degree[i] = 0;

  ne = 0;

  center = -1;
  newedge = init_edge;

  /* skip over blank */

  newend = get_newend(inputline[inchar++],&flip_edge);

  /* an edge is MARKed if both of its directed EDGEs have the same 
     orientation */

  RESETMARKS;

  while (ne < expect_ne)
    {
      if (newend == -1)
	{
	  center++;
	  newend = get_newend(inputline[inchar++],&flip_edge);
	  degree[nv-1-center] = 0;
	  firstedge[nv-1-center] = newedge;
	  newedge->next = newedge->prev = newedge;
	}
      else
	{
	  newedge->next = firstedge[nv-1-center];
	  newedge->prev = firstedge[nv-1-center]->prev;
	  firstedge[nv-1-center]->prev = newedge;
	  newedge->prev->next = newedge;
	}
      newedge->start = nv-1-center;
      newedge->end = nv-1-newend;
      adj[nv-1-center][nv-1-newend] = TRUE;
      if (newend < center)
	{
	  ed = firstedge[nv-1-newend];
	  while (ed->end != nv-1-center)
	    ed = ed->next;
	  ed->invers = newedge;
	  newedge->invers = ed;
	}

      degree[nv-1-center]++;
      newedge++;
      ne++;

      newend = get_newend(inputline[inchar++],&flip_edge);
    }

  /* set link for ->op edges */

  for (i=0;i<nv;i++) {
    ed = firstedge[i];
    MARKLO(ed);
    for (j=0;j<degree[i];j++) {
      ed->op->start = ed->start;
      ed->op->end = ed->end;
      ed->op->next = ed->prev->op;
      ed->op->prev = ed->next->op;
      ed->op->invers = ed->invers->op;
      ed = ed->next;
    }
  }
  
  /* reorient edges as needed */

  input_orient = TRUE;
  for (i=nv-1;i>=0;i--) {
    ed = firstedge[i];
    for (j=0;j<degree[i];j++) {
      ed1 = ed->next;
      if (!ISMARKED(ed1) &&
	  !ISMARKED(ed1->invers) &&
	  !ISMARKED(ed1->op) &&
	  !ISMARKED(ed1->invers->op)) {
	if (ed1->invers->next->end != ed->invers->prev->end) {
	  input_orient = FALSE;
	  ed1->invers->invers = ed1->op;
	  ed1->invers->op->invers = ed1;
	  ed1->op->invers = ed1->invers;
	  ed1->invers = ed1->invers->op;
	}
	MARKLO(ed1);
      }
      ed = ed->next;
    }
  }

  if (input_orient != orient) {
    fprintf(stderr,
	    "Expected orientability is %d.  Orientability of input is %d.\n",
	    orient,input_orient);
    exit(1);
  }

  /* count common neighbors */

  for (i=0;i<nv-1;i++) {
    nnei[i][i] = 0;
    for (j=i+1;j<nv;j++)
      {
	nnei[i][j] = 0;
	for (k=0;k<nv;k++)
	  if (adj[i][k] && adj[j][k])
	    nnei[i][j]++;
	nnei[j][i] = nnei[i][j];
      }
  }

  update_ncontract_quad();

  /*
  irreducible = TRUE;
  for (i=0;i<nv;i++)
    if (ncontract[i] > 0)
      irreducible = FALSE;
  */

  /*
  if (!irreducible &&  nv > 6) {
    fprintf(stderr,">E %s: input triangulation is not irreducible\n",cmdname);
    fprintf(stderr,">E input line is %s\n",inputline);
    perror(">E ");
    exit(1);
    }
  */

}

/**************************************************************************/

static void
find_extensions(int numb_total, int numb_pres,
                EDGE *ext[], int degext[], int *numext)

/* Find all nonequivalent places for extension.
   These are listed in ext with the degree of the future new vertex in degext.
   The number of cases is returned in numext. */
{
  register int i,j,k,ideg;
  int deg3;
  int mindeg;        /* minimum degree */
  int nmindeg;       /* number of vertices with degree of mindeg */
  register EDGE *e,*e1,*ex;
  EDGE **nb0,**nb1,**nboplim,**nblim;

  mindeg = nv;
  nmindeg = 0;
  for (i = 0; i < nv; ++i)
    if (ncontract[i]>0) {
      if (degree[i] == mindeg)
	++nmindeg;
      else if (degree[i] < mindeg)
	{
	  mindeg = degree[i];
	  nmindeg = 1;
	}
    }

  if (mindeg == 3)
    deg3 = nmindeg;
  else
    deg3 = 0;

  /* code for trivial group */

  k = 0;
  if (numb_total == 1)
    {

      /* extend with degree 3 */

      for (i = 0; i < nv; ++i)
        {
	  e = ex = firstedge[i];
	  do
            {
	      if (e->end > i)
                {
		  if (e->next->end > i) {degext[k] = 3; ext[k++] = e;}
                }
	      e = e->next;
            } while (e != ex);
        }

      /* extend with degree 4 */

      if (deg3 <= 2)
	for (i = 0; i < nv; ++i)
	  {
	    if (degree[i] == 3) continue;
	    e = ex = firstedge[i];
	    do
	      {
		if (e->next->end > i)
		  {
		    if ((degree[e->end] == 3)
			+ (degree[e->next->next->end] == 3) == deg3)
		      {degext[k] = 4; ext[k++] = e;}
		  }
		e = e->next;
	      } while (e != ex);
	  }

      /* extend with degree n=5,6,...mindeg */

      for (ideg=5;ideg<=mindeg;ideg++)
	for (i = 0; i < nv; ++i)
	  {
	    if (degree[i] < 2*ideg-4) continue;
	    e = ex = firstedge[i];
	    do
	      {
		degext[k] = ideg;
		ext[k++] = e;
		e = e->next;
	      } while (e != ex);
	  }

      /* extend with degree mindeg+1 */

      if (mindeg >= 4 && nmindeg <= 2)
	for (i = 0; i < nv; ++i)
	  {
	    if (degree[i] < 2*(mindeg+1)-4) continue;
	    e = ex = firstedge[i];
	    do
	      {
		e1 = e;
		for (j=0;j<(mindeg+1)-2;j++)
		  e1 = e1->next;
		if ((degree[e->end] == mindeg && ncontract[e->end]>0) +
		    (degree[e1->end] == mindeg && ncontract[e1->end]>0) ==
		    nmindeg)
		  {
		    degext[k] = (mindeg+1);
		    ext[k++] = e;
		  }
		e = e->next;
	      } while (e != ex);
	  }
    }

  /* code for nontrivial group */

  else
    {
      nb0 = (EDGE**)numbering[0];
      nboplim = (EDGE**)numbering[numb_pres == 0 ? numb_total : numb_pres];
      nblim = (EDGE**)numbering[numb_total];

      for (i = 0; i < ne; ++i) {
	nb0[i]->index = i;
	nb0[i]->op->index = -1;
      }

      /* extend with degree 3 */

      RESETMARKS;
      for (i = 0; i < ne; ++i) {
	e = nb0[i];
	if (ISMARKEDLO(e)) continue;
	degext[k] = 3; ext[k++] = e;
	
	/* mark first edge */
	
	for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
	  MARKLO(*nb1);
	  MARKLO((*nb1)->op->invers);
	}	
	for (; nb1 < nblim; nb1 += MAXE) {
	  MARKLO((*nb1)->invers);
	  MARKLO((*nb1)->op);
	}	
	
	/*mark second edge */	
	
	e1 = e->invers->prev;
	if (e1->index >= 0) {
	  for (nb1 = nb0 + e1->index; nb1 < nboplim; nb1 += MAXE) {
	    MARKLO(*nb1);
	    MARKLO((*nb1)->op->invers);
	  }
	  for (; nb1 < nblim; nb1 += MAXE) {
	    MARKLO((*nb1)->invers);
	    MARKLO((*nb1)->op);
	  }
	}
	else {
	  for (nb1 = nb0 + e1->op->index; nb1 < nboplim; nb1 += MAXE) {
	    MARKLO((*nb1)->invers);
	    MARKLO((*nb1)->op);
	  }
	  for (; nb1 < nblim; nb1 += MAXE) {
	    MARKLO(*nb1);
	    MARKLO((*nb1)->op->invers);
	  }
	}
	
	/* make third edge */
	
	e1 = e1->invers->prev;
	if (e1->index >= 0) {
	  for (nb1 = nb0 + e1->index; nb1 < nboplim; nb1 += MAXE) {
	    MARKLO(*nb1);
	    MARKLO((*nb1)->op->invers);
	  }
	  for (; nb1 < nblim; nb1 += MAXE) {
	    MARKLO((*nb1)->invers);
	    MARKLO((*nb1)->op);
	  }
	}
	else {
	  for (nb1 = nb0 + e1->op->index; nb1 < nboplim; nb1 += MAXE) {
	    MARKLO((*nb1)->invers);
	    MARKLO((*nb1)->op);
	  }
	  for (; nb1 < nblim; nb1 += MAXE) {
	    MARKLO(*nb1);
	    MARKLO((*nb1)->op->invers);
	  }
	}
	
      }

      /* extend with degree 4 */

      if (deg3 <= 2)
        {
	  RESETMARKS;
	
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e)) continue;
	      e1 = e->next->invers->prev;
	      if ((degree[e->end] == 3) + (degree[e1->end] == 3) != deg3)
		continue;
	      degext[k] = 4; ext[k++] = e;
	
	      for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
		MARKLO(*nb1);
		MARKLO((*nb1)->op->prev->prev);
	      }

	      for (; nb1 < nblim; nb1 += MAXE) {
		MARKLO((*nb1)->prev->prev);
		MARKLO((*nb1)->op);
	      }

	      if (e1->index >= 0) {
		for (nb1 = nb0 + e1->index; nb1 < nboplim; nb1 += MAXE) {
		  MARKLO(*nb1);
		  MARKLO((*nb1)->op->prev->prev);
		}
	
		for (; nb1 < nblim; nb1 += MAXE) {
		  MARKLO((*nb1)->prev->prev);
		  MARKLO((*nb1)->op);
		}
	      }
	      else {
		e1 = e1->op;
		for (nb1 = nb0 + e1->index; nb1 < nboplim; nb1 += MAXE) {
		  MARKLO((*nb1)->prev->prev);
		  MARKLO((*nb1)->op);
		}
		
		for (; nb1 < nblim; nb1 += MAXE) {
		  MARKLO(*nb1);
		  MARKLO((*nb1)->op->prev->prev);
		}
	      }
            }
        }

      /* extend with degree n=5,6,...mindeg */

      for (ideg=5;ideg<=mindeg;ideg++)
        {
	  RESETMARKS;
	
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e) || degree[e->start] < 2*ideg-4) continue;
	
	      degext[k] = ideg; ext[k++] = e;
	
	      for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
		MARKLO(*nb1);
		e1 = (*nb1)->op;
		for (j=0;j<ideg-2;j++)
		  e1 = e1->prev;
		MARKLO(e1);
	      }

	      for (; nb1 < nblim; nb1 += MAXE) {
		e1 = *nb1;
		for (j=0;j<ideg-2;j++)
		  e1 = e1->prev;
		MARKLO(e1);
		MARKLO((*nb1)->op);
	      }
	    }
        }

      /* extend with degree mindeg+1 */

      if (mindeg >= 4 && nmindeg <= 2)
        {
	  RESETMARKS;
	
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e) || degree[e->start] < 2*(mindeg+1)-4) continue;
	
	      e1 = e;
	      for (j=0;j<(mindeg+1)-2;j++)
		e1 = e1->next;
	      if ((degree[e->end] == mindeg && ncontract[e->end]>0) +
		  (degree[e1->end] == mindeg && ncontract[e1->end]>0) != nmindeg)
		continue;
	      degext[k] = mindeg+1; ext[k++] = e;
	
	      for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
		MARKLO(*nb1);
		e1 = (*nb1)->op;
		for (j=0;j<(mindeg+1)-2;j++)
		  e1 = e1->prev;
		MARKLO(e1);
	      }

	      for (; nb1 < nblim; nb1 += MAXE) {
		e1 = *nb1;
		for (j=0;j<(mindeg+1)-2;j++)
		  e1 = e1->prev;
		MARKLO(e1);
		MARKLO((*nb1)->op);
		}
            }
        }
    }
  *numext = k;

}

/**************************************************************************/

static int
destroy_hole(EDGE *e, int degext)

/* This routine is for triangulations with holes.
   Determine if an extention of degext at e would destroy a button hole.
   This would occur if e->start (=a) is on a hole (=abc...yz) and the edge 
   (ac) or (ay) is extended to a 2-path. */
{
  int i,j,c,y,ok_c,ok_y;
  EDGE *e1,*e2;

  e1 = e;
  for (i=0; i<degext-2; i++)
    {
      if (RIGHTHOLE(e1)) {
	c = e1->invers->prev->end;
	y = e1->next->invers->next->end;
	ok_c = FALSE;
	ok_y = FALSE;
	e2 = e;
	for (j=0; j<degext-1 && !(ok_c && ok_y); j++)
	  {
	    ok_c = ok_c || (e2->end == c);
	    ok_y = ok_y || (e2->end == y);
	    e2 = e2->next;
	  }
	if (!(ok_c && ok_y)) return TRUE;
      }
      e1 = e1->next;
    }
  
  do
    {
      if (RIGHTHOLE(e1)) {
	c = e1->invers->prev->end;
	y = e1->next->invers->next->end;
	e2 = e->next;
	for (j=0; j<degext-3; j++)
	  {
	    if (e2->end == c) return TRUE;
	    if (e2->end == y) return TRUE;
	    e2 = e2->next;
	  }
      }
      e1 = e1->next;
    } while (e1 != e);
  
  return FALSE;
}
 
/**************************************************************************/

static void
find_extensions_holes(int numb_total, int numb_pres,
                EDGE *ext[], int degext[], int *numext)

/* This routine is for triangulations with holes.
   Find all nonequivalent places for extension.
   These are listed in ext with the degree of the future new vertex in degext.
   The number of cases is returned in numext. */
{
  register int i,j,k,ideg;
  int deg3;
  int mindeg;        /* minimum degree */
  int nmindeg;       /* number of vertices with degree of mindeg */
  register EDGE *e,*e1,*ex;
  EDGE **nb0,**nb1,**nboplim,**nblim;

  mindeg = nv;
  nmindeg = 0;
  for (i = 0; i < nv; ++i)
    if (ncontract[i]>0) {
      if (degree[i] == mindeg)
	++nmindeg;
      else if (degree[i] < mindeg)
	{
	  mindeg = degree[i];
	  nmindeg = 1;
	}
    }

  if (mindeg == 3)
    deg3 = nmindeg;
  else
    deg3 = 0;

  /* code for trivial group */

  k = 0;
  if (numb_total == 1)
    {

      /* extend with degree 3 */

      for (i = 0; i < nv; ++i)
        {
	  e = ex = firstedge[i];
	  do
            {
	      if (!RIGHTHOLE(e)) {
		if (e->end > i)
		  if (e->next->end > i) {degext[k] = 3; ext[k++] = e;}
	      }
	      else {
		if (e->invers->prev->end == e->next->end &&
		    e->next->invers->next->end == e->end) 
		  {degext[k] = 3; ext[k++] = e;}
	      }
	      e = e->next;
            } while (e != ex);
        }

      /* extend with degree 4 */

      if (deg3 <= 2)
	for (i = 0; i < nv; ++i)
	  {
	    if (degree[i] == 3 && ncontract[i]>0) continue;
	    e = ex = firstedge[i];
	    do
	      {
		if (!destroy_hole(e,4))
		  if ((e->next->end > i) || RIGHTHOLE(e) || RIGHTHOLE(e->next))
		    if ((degree[e->end] == 3 && ncontract[e->end]>0) +
			(degree[e->next->next->end] == 3 && 
			 ncontract[e->next->next->end]>0) == deg3)
		      {degext[k] = 4; ext[k++] = e;}
		e = e->next;
	      } while (e != ex);
	  }
      
      /* extend with degree n=5,6,...mindeg */

      for (ideg=5;ideg<=mindeg;ideg++)
	for (i = 0; i < nv; ++i)
	  {
	    if (degree[i] < 2*ideg-4) continue;
	    e = ex = firstedge[i];
	    do
	      {
		if (!destroy_hole(e,ideg))
		  {
		    degext[k] = ideg;
		    ext[k++] = e;
		  }
		e = e->next;
	      } while (e != ex);
	  }
      
      /* extend with degree mindeg+1 */
      
      if (mindeg >= 4 && nmindeg <= 2)
	for (i = 0; i < nv; ++i)
	  {
	    if (degree[i] < 2*(mindeg+1)-4) continue;
	    e = ex = firstedge[i];
	    do
	      {
		if (!destroy_hole(e,mindeg+1))
		  {
		    e1 = e;
		    for (j=0;j<(mindeg+1)-2;j++)
		      e1 = e1->next;
		    
		    if ((degree[e->end] == mindeg && ncontract[e->end]>0) +
			(degree[e1->end] == mindeg && ncontract[e1->end]>0) ==
			nmindeg)
		      {
			degext[k] = (mindeg+1);
			ext[k++] = e;
		      }
		  }
		e = e->next;
	      } while (e != ex);
	  }
    }

  /* code for nontrivial group */

  else
    {
      nb0 = (EDGE**)numbering[0];
      nboplim = (EDGE**)numbering[numb_pres == 0 ? numb_total : numb_pres];
      nblim = (EDGE**)numbering[numb_total];

      for (i = 0; i < ne; ++i) {
	nb0[i]->index = i;
	nb0[i]->op->index = -1;
      }

      /* extend with degree 3 */

      RESETMARKS;
      for (i = 0; i < ne; ++i) {
	e = nb0[i];
	if (ISMARKEDLO(e)) continue;
	if (RIGHTHOLE(e) &&
	    (e->invers->prev->end != e->next->end ||
	     e->next->invers->next->end != e->end)) continue;
	
	degext[k] = 3; ext[k++] = e;
	
	/* mark first edge */
	
	for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
	  MARKLO(*nb1);
	  MARKLO((*nb1)->op->invers);
	}	
	for (; nb1 < nblim; nb1 += MAXE) {
	  MARKLO((*nb1)->invers);
	  MARKLO((*nb1)->op);
	}	
	
	/*mark second edge */	
	
	e1 = e->next->op;
	if (e1->index >= 0) {
	  for (nb1 = nb0 + e1->index; nb1 < nboplim; nb1 += MAXE) {
	    MARKLO(*nb1);
	    MARKLO((*nb1)->op->invers);
	  }
	  for (; nb1 < nblim; nb1 += MAXE) {
	    MARKLO((*nb1)->invers);
	    MARKLO((*nb1)->op);
	  }
	}
	else {
	  for (nb1 = nb0 + e1->op->index; nb1 < nboplim; nb1 += MAXE) {
	    MARKLO((*nb1)->invers);
	    MARKLO((*nb1)->op);
	  }
	  for (; nb1 < nblim; nb1 += MAXE) {
	    MARKLO(*nb1);
	    MARKLO((*nb1)->op->invers);
	  }
	}
	
	/* make third edge */
	
	if (!RIGHTHOLE(e)) {

	  e1 = e->invers->prev;
	  if (e1->index >= 0) {
	    for (nb1 = nb0 + e1->index; nb1 < nboplim; nb1 += MAXE) {
	      MARKLO(*nb1);
	      MARKLO((*nb1)->op->invers);
	    }
	    for (; nb1 < nblim; nb1 += MAXE) {
	      MARKLO((*nb1)->invers);
	      MARKLO((*nb1)->op);
	    }
	  }
	  else {
	    for (nb1 = nb0 + e1->op->index; nb1 < nboplim; nb1 += MAXE) {
	      MARKLO((*nb1)->invers);
	      MARKLO((*nb1)->op);
	    }
	    for (; nb1 < nblim; nb1 += MAXE) {
	      MARKLO(*nb1);
	      MARKLO((*nb1)->op->invers);
	    }
	  }
	}
      }

      /* extend with degree 4 */

      if (deg3 <= 2)
        {
	  RESETMARKS;
	
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e)) continue;
	      if ((degree[e->end] == 3 && ncontract[e->end]>0) + 
		  (degree[e->next->next->end] == 3 && 
		   ncontract[e->next->next->end]>0) != deg3)
		continue;
	      if (!destroy_hole(e,4))
		{
		  degext[k] = 4; ext[k++] = e;
		}

	      for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
		MARKLO(*nb1);
		MARKLO((*nb1)->op->prev->prev);
	      }
	      
	      for (; nb1 < nblim; nb1 += MAXE) {
		MARKLO((*nb1)->prev->prev);
		MARKLO((*nb1)->op);
	      }
	      
	      if (!RIGHTHOLE(e) && !RIGHTHOLE(e->next))
		{
		  e1 = e->next->invers->prev;
		  if (e1->index >= 0) {
		    for (nb1 = nb0 + e1->index; nb1 < nboplim; nb1 += MAXE) {
		      MARKLO(*nb1);
		      MARKLO((*nb1)->op->prev->prev);
		    }
		    
		    for (; nb1 < nblim; nb1 += MAXE) {
		      MARKLO((*nb1)->prev->prev);
		      MARKLO((*nb1)->op);
		    }
		  }
		  else {
		    e1 = e1->op;
		    for (nb1 = nb0 + e1->index; nb1 < nboplim; nb1 += MAXE) {
		      MARKLO((*nb1)->prev->prev);
		      MARKLO((*nb1)->op);
		    }
		    
		    for (; nb1 < nblim; nb1 += MAXE) {
		      MARKLO(*nb1);
		      MARKLO((*nb1)->op->prev->prev);
		    }
		  }
		}
	    }
        }
      
      /* extend with degree n=5,6,...mindeg */
      
      for (ideg=5;ideg<=mindeg;ideg++)
        {
	  RESETMARKS;
	  
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e) || degree[e->start] < 2*ideg-4) continue;
	      
	      if (!destroy_hole(e,ideg))
		  {
		    degext[k] = ideg; 
		    ext[k++] = e;
		  }

	      for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
		MARKLO(*nb1);
		e1 = (*nb1)->op;
		for (j=0;j<ideg-2;j++)
		  e1 = e1->prev;
		MARKLO(e1);
	      }
	      
	      for (; nb1 < nblim; nb1 += MAXE) {
		e1 = *nb1;
		for (j=0;j<ideg-2;j++)
		  e1 = e1->prev;
		MARKLO(e1);
		MARKLO((*nb1)->op);
	      }
	    }
        }
      
      /* extend with degree mindeg+1 */
      
      if (mindeg >= 4 && nmindeg <= 2)
        {
	  RESETMARKS;
	  
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e) || degree[e->start] < 2*(mindeg+1)-4) continue;
	      
	      if (!destroy_hole(e,mindeg+1))
		{
		  e1 = e->next;
		  for (j=1;j<(mindeg+1)-2;j++)
		    e1 = e1->next;
		  
		  if ((degree[e->end] == mindeg && ncontract[e->end]>0) +
		      (degree[e1->end] == mindeg && ncontract[e1->end]>0) == nmindeg)
		    {
		      degext[k] = mindeg+1; ext[k++] = e;
		    }
		}
	      
	      for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
		MARKLO(*nb1);
		e1 = (*nb1)->op;
		for (j=0;j<(mindeg+1)-2;j++)
		  e1 = e1->prev;
		MARKLO(e1);
	      }
	      
	      for (; nb1 < nblim; nb1 += MAXE) {
		e1 = *nb1;
		for (j=0;j<(mindeg+1)-2;j++)
		  e1 = e1->prev;
		MARKLO(e1);
		MARKLO((*nb1)->op);
	      }
	    }
	}
    }
  *numext = k;
  
}

/**************************************************************************/

static void
compute_code(unsigned char code[])

/* computes a code by numbering the vertices in a breadth first manner and
   then listing them in code. Code is a sequence of numbers of length ne+nv+1.
   The first entry is the number of vertices.
   Then the numbers of the vertices adjacent to the vertex numbered 1 are
   given -- ended by a 0, and listed in clockwise orientation.
   Then those adjacent to 2, ended by a 0, etc. . In case of no
   double edges, the identification of the corresponding "half edges" leaving
   each vertex is unique. Nevertheless also for this case the following rules
   apply (not in the definition of the code, but in this routine):

   In case of a given starting edge in code_edge, the start of this
   edge is numbered 1 and the end 2.
*/
{
  register EDGE *run;
  register int vertex;
  EDGE *temp;
  EDGE *startedge[MAXN+1];
  int number[MAXN+1], i;
  int last_number, actual_number;
  EDGE *givenedge;

  for (i = 0; i < nv; i++) number[i] = 0;

  *code=nv; code++;
  if (code_edge==NULL) givenedge=firstedge[0];
  else { givenedge=code_edge; number[nv]=0; }
  number[givenedge->start] = 1;
  if (givenedge->start != givenedge->end)
    {
      number[givenedge->end] = 2;
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else  last_number = 1;

  actual_number = 1;
  temp = givenedge;

  while (last_number < nv)
    {   *code=number[temp->end]; code++;
    for (run = temp->next; run != temp; run = run->next)
      { vertex = run->end;
      if (!number[vertex])
	{ startedge[last_number] = run->invers;
	last_number++; number[vertex] = last_number;
	*code = last_number; }
      else *code = number[vertex];
      code++;
      }
    *code = 0;  code++;
    temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nv)
    {  	*code=number[temp->end]; code++;
    for (run = temp->next; run != temp; run = run->next)
      {
	*code = number[run->end]; code++;
      }
    *code = 0;
    code++;
    temp = startedge[actual_number];  actual_number++;
    }

}

/****************************************************************************/

static void
compute_code_mirror(unsigned char code[])

/* In the case of no double edges -- that is when the identifications are
   clear -- there is no problem in just returning the inverse order of
   the previously computed code. Nevertheless in the case where edge
   identifications must be made via the spanning tree described in the code,
   this is not that easy, so we can as well just compute mirror code just
   like the normal code computed before.

   In case code_edge is not NULL, its start is numbered 1 and its
   end is numbered 0.
*/
{
  register EDGE *run;
  register int vertex;
  EDGE *temp;
  EDGE *startedge[MAXN+1];
  int number[MAXN+1], i;
  int last_number, actual_number;
  EDGE *givenedge;

  for (i = 0; i < nv; i++) number[i] = 0;

  *code=nv; code++;
  if (code_edge==NULL) givenedge=firstedge[0];
  else { givenedge=code_edge; number[nv]=0; }
  number[givenedge->start] = 1;
  if (givenedge->start != givenedge->end)
    {
      number[givenedge->end] = 2;
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else  last_number = 1;

  actual_number = 1;
  temp = givenedge;

  while (last_number < nv)
    {   *code=number[temp->end]; code++;
    for (run = temp->prev; run != temp; run = run->prev)
      { vertex = run->end;
      if (!number[vertex])
	{ startedge[last_number] = run->invers;
	last_number++; number[vertex] = last_number;
	*code = last_number; }
      else *code = number[vertex];
      code++;
      }
    *code = 0;  code++;
    temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nv)
    {  	*code=number[temp->end]; code++;
    for (run = temp->prev; run != temp; run = run->prev)
      {
	*code = number[run->end]; code++;
      }
    *code = 0;
    code++;
    temp = startedge[actual_number];  actual_number++;
    }
}

/****************************************************************************/

static void
compute_dual_code(unsigned char code[])

/* works like compute_code -- only for the dual */

{
  register EDGE *run, *run2;
  EDGE *temp;
  EDGE *startedge[MAXF+1];
  int last_number, actual_number;
  int nf;
  EDGE *givenedge;

  nf=eulerc+(ne/2)-nv;
  *code=nf; code++;
  RESETMARKS; /* The face on the right has already been numbered if
		 and only if it is marked. */

  if (code_edge==NULL) givenedge=firstedge[0];
  else givenedge=code_edge->invers;

  run=givenedge;
  do { MARKLO(run); run->rightface=1; run=run->invers->prev;
  } while (run!=givenedge);
  if (!ISMARKED(givenedge->invers)) /* no loop in the dual at this point */
    { run2=run=givenedge->invers;
    do { MARKLO(run2); run2->rightface=2; run2=run2->invers->prev;
    } while (run2!=run);
    last_number = 2;
    startedge[1] = givenedge; /* the startedge has the face it belongs
				 to on the LEFT */
    }
  else  last_number = 1;

  actual_number = 1;
  temp = givenedge->invers;
  while (last_number < nf)
    {
      *code= temp->rightface; code++;
      for (run = temp->prev->invers; run != temp; run = run->prev->invers)
	/* run also has the face it runs around on the left */
	{ if (!ISMARKED(run))
	  { startedge[last_number] = run->invers;
	  last_number++;
	  run2=run;
	  do
	    { MARKLO(run2);
	    run2->rightface=last_number;
	    run2=run2->invers->prev;
	    } while (run2!=run);
	  *code = last_number;
	  }
	else *code=run->rightface;
	code++;
	}
      *code = 0;  code++;
      temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nf)
    {  	*code=temp->rightface; code++;
    for (run = temp->prev->invers; run != temp; run = run->prev->invers)
      {
	*code = run->rightface; code++;
      }
    *code = 0;
    code++;
    temp = startedge[actual_number];  actual_number++;
    }
}

/****************************************************************************/

static void
compute_dual_code_mirror(unsigned char code[])

/* works like compute_code_mirror -- only for the dual */

{
  register EDGE *run, *run2;
  EDGE *temp;
  EDGE *startedge[MAXF+1];
  int last_number, actual_number;
  int nf;
  EDGE *givenedge;

  nf=eulerc+(ne/2)-nv;
  *code=nf; code++;
  RESETMARKS; /* The face on the right has already been numbered
		 if and only if it is marked. */

  if (code_edge==NULL) givenedge=firstedge[0];
  else givenedge=code_edge->invers;

  run=givenedge;
  do { MARKLO(run); run->rightface=1; run=run->invers->prev;
  } while (run!=givenedge);
  if (!ISMARKED(givenedge->invers)) /* no loop in the dual at this point */
    { run2=run=givenedge->invers;
    do { MARKLO(run2); run2->rightface=2; run2=run2->invers->prev;
    } while (run2!=run);
    last_number = 2;
    startedge[1] = givenedge; /* the startedge has the face it
				 belongs to on the LEFT */
    }
  else  last_number = 1;

  actual_number = 1;
  temp = givenedge->invers;

  while (last_number < nf)
    {
      *code= temp->rightface; code++;
      for (run = temp->invers->next; run != temp; run = run->invers->next)
	/* run also has the face it runs around on the left */
	{ if (!ISMARKED(run))
	  { startedge[last_number] = run->invers;
	  last_number++;
	  run2=run;
	  do
	    { MARKLO(run2); run2->rightface=last_number;
	    run2=run2->invers->prev;
	    } while (run2!=run);
	  *code = last_number;
	  }
	else *code=run->rightface;
	code++;
	}
      *code = 0;  code++;
      temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nf)
    {  	*code=temp->rightface; code++;
    for (run = temp->invers->next; run != temp; run = run->invers->next)
      {
	*code = run->rightface; code++;
      }
    *code = 0;
    code++;
    temp = startedge[actual_number];  actual_number++;
    }
}

/**************************************************************************/

static void
compute_code_non_or(unsigned char code[])

/* computes a code by numbering the vertices in a breadth first manner and
   then listing them in code. Code is a sequence of numbers of length ne+nv+1.
   The first entry is the number of vertices.
   Then the numbers of the vertices adjacent to the vertex numbered 1 are
   given -- ended by a 0, and listed in clockwise orientation.
   Then those adjacent to 2, ended by a 0, etc. . In case of no
   double edges, the identification of the corresponding "half edges" leaving
   each vertex is unique. Nevertheless also for this case the following rules
   apply (not in the definition of the code, but in this routine):

   In case of a given starting edge in code_edge, the start of this
   edge is numbered 1 and the end 2.
*/
{
  register EDGE *run;
  register int vertex;
  EDGE *temp;
  EDGE *startedge[MAXN+1];
  int number[MAXN+1], i;
  int last_number, actual_number;
  EDGE *givenedge;
  EDGE *run1,*end1;

  /* when an edge e is encountered in the breadth first search for which 
     e->invers has not been marked (first time e->end has been encountered), 
     MARK e->invers and all ->next of e->invers. */

  RESETMARKS;

  for (i = 0; i < nv; i++) number[i] = 0;

  *code=nv; code++;
  if (code_edge==NULL) givenedge=firstedge[0];
  else { givenedge=code_edge; number[nv]=0; }
  number[givenedge->start] = 1;
  if (givenedge->start != givenedge->end)
    {
      number[givenedge->end] = 2;
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else  last_number = 1;

  actual_number = 1;

  run1 = end1 = givenedge;
  do
    {	
      MARKHI(run1);
      run1 = run1->next;
    } while (run1 != end1);
  
  if (givenedge->start != givenedge->end)
    {
      run1 = end1 = givenedge->invers;
      do
	{	
	  MARKHI(run1);
	  run1 = run1->next;
	} while (run1 != end1);
    }

  temp = givenedge;

  while (last_number < nv)
    { 
      *code=number[temp->end];
      code++;
      for (run = temp->next; run != temp; run = run->next)
	{ vertex = run->end;
	  if (!number[vertex])
	    { startedge[last_number] = run->invers;
	      last_number++; number[vertex] = last_number;
	      run1 = end1 = run->invers;
	      do
		{	
		  MARKHI(run1);
		  run1 = run1->next;
		} while (run1 != end1);
	    }
	  if (ISMARKEDHI(run) == ISMARKEDHI(run->invers))
	    *code=number[vertex];
	  else
	    *code=MAXN+number[vertex];
	  code++;
	}
      *code = 0;  code++;
      temp = startedge[actual_number];  actual_number++;
    }
  
  while (actual_number <= nv)
    {  	*code=number[temp->end]; code++;
      for (run = temp->next; run != temp; run = run->next)
	{
	  if (ISMARKEDHI(run) == ISMARKEDHI(run->invers))
	    *code=number[run->end];
	  else
	    *code=MAXN+number[run->end];
	  code++;
	}
      *code = 0;
      code++;
      temp = startedge[actual_number];  actual_number++;
    }
  
}

/**************************************************************************/

static void
compute_dual_code_non_or(unsigned char code[])

/* works like compute_code_non_or -- only for the dual */

{
  register EDGE *run, *run2;
  EDGE *temp;
  EDGE *startedge[MAXF+1];
  int last_number, actual_number;
  int nf;
  EDGE *givenedge;

  nf=eulerc+(ne/2)-nv;
  *code=nf; code++;
  RESETMARKS; /* The face on the right has already been numbered if
		 and only if it is marked. */

  if (code_edge==NULL) givenedge=firstedge[0];
  else givenedge=code_edge->invers;

  run=givenedge;
  do { MARKLO(run); run->rightface=run->op->invers->rightface=1; run=run->invers->prev;
  } while (run!=givenedge);
  if (!ISMARKED(givenedge->invers)) /* no loop in the dual at this point */
    { run2=run=givenedge->invers;
    do { MARKLO(run2); run2->rightface=run2->op->invers->rightface=2; run2=run2->invers->prev;
    } while (run2!=run);
    last_number = 2;
    startedge[1] = givenedge; /* the startedge has the face it belongs
				 to on the LEFT */
    }
  else  last_number = 1;

  actual_number = 1;
  temp = givenedge->invers;
  while (last_number < nf)
    {
      *code= temp->rightface; code++;
      for (run = temp->prev->invers; run != temp; run = run->prev->invers)
	/* run also has the face it runs around on the left */
	{ 
	  if (!ISMARKED(run) && !ISMARKED(run->op->invers))
	    { startedge[last_number] = run->invers;
	      last_number++;
	      run2=run;
	      do
		{ MARKLO(run2);
		  run2->rightface=run2->op->invers->rightface=last_number;
		  run2=run2->invers->prev;
		} while (run2!=run);
	    }
	  if (ISMARKED(run) == ISMARKED(run->invers))
	    *code = run->rightface;
	  else
	    *code=MAXN + run->rightface;
	  code++;
	}
      *code = 0;  code++;
      temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nf)
    {  	*code=temp->rightface; code++;
    for (run = temp->prev->invers; run != temp; run = run->prev->invers)
      {
	if (ISMARKED(run) == ISMARKED(run->invers))
	  *code = run->rightface;
	else
	  *code=MAXN + run->rightface;
	code++;
      }
    *code = 0;
    code++;
    temp = startedge[actual_number];  actual_number++;
    }
}

/**************************************************************************/

static void
compute_code_mirror_non_or(unsigned char code[])

/* In the case of no double edges -- that is when the identifications are
   clear -- there is no problem in just returning the inverse order of
   the previously computed code. Nevertheless in the case where edge
   identifications must be made via the spanning tree described in the code,
   this is not that easy, so we can as well just compute mirror code just
   like the normal code computed before.

   In case code_edge is not NULL, its start is numbered 1 and its
   end is numbered 0.
*/
{
  register EDGE *run;
  register int vertex;
  EDGE *temp;
  EDGE *startedge[MAXN+1];
  int number[MAXN+1], i;
  int last_number, actual_number;
  EDGE *givenedge;
  EDGE *run1,*end1;

  /* when an edge e is encountered in the breadth first search for which 
     e->invers has not been marked (first time e->end has been encountered), 
     MARK e->invers and all ->next of e->invers. */

  RESETMARKS;

  for (i = 0; i < nv; i++) number[i] = 0;

  *code=nv; code++;
  if (code_edge==NULL) givenedge=firstedge[0];
  else { givenedge=code_edge; number[nv]=0; }
  number[givenedge->start] = 1;
  if (givenedge->start != givenedge->end)
    {
      number[givenedge->end] = 2;
      last_number = 2;
      startedge[1] = givenedge->invers;
    }
  else  last_number = 1;

  actual_number = 1;

  run1 = end1 = givenedge;
  do
    {	
      MARKHI(run1);
      run1 = run1->next;
    } while (run1 != end1);
  
  if (givenedge->start != givenedge->end)
    {
      run1 = end1 = givenedge->invers;
      do
	{	
	  MARKHI(run1);
	  run1 = run1->next;
	} while (run1 != end1);
    }

  temp = givenedge;

  while (last_number < nv)
    {   
      *code=number[temp->end]; code++;
      for (run = temp->prev; run != temp; run = run->prev)
	{ vertex = run->end;
	  if (!number[vertex])
	    { startedge[last_number] = run->invers;
	      last_number++; number[vertex] = last_number;
	      run1 = end1 = run->invers;
	      do
		{	
		  MARKHI(run1);
		  run1 = run1->next;
		} while (run1 != end1);
	    }
	  if (ISMARKEDHI(run) == ISMARKEDHI(run->invers))
	    *code=number[vertex];
	  else
	    *code=MAXN+number[vertex];
	  code++;
	}
      *code = 0;  code++;
      temp = startedge[actual_number];  actual_number++;
    }

  while (actual_number <= nv)
    {  	*code=number[temp->end]; code++;
      for (run = temp->prev; run != temp; run = run->prev)
	{
	  if (ISMARKEDHI(run) == ISMARKEDHI(run->invers))
	    *code=number[run->end];
	  else
	    *code=MAXN+number[run->end];
	  code++;
	}
      *code = 0;
      code++;
      temp = startedge[actual_number];  actual_number++;
    }
}

/**************************************************************************/

static void
write_planar_code(FILE *f, int doflip, int doprev)

/* Write in planar_code format.  Always write in next direction,
   and if doflip != 0 also write in prev direction. */
{
  size_t length;
  unsigned char code[MAXN+MAXE+1];

  length=nv+ne+1;
  compute_code(code);
  if (fwrite(code,sizeof(unsigned char),length,f) != length)
    {
      fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
      perror(">E ");
      exit(1);
    }
  if (doflip)
    { compute_code_mirror(code);
    if (fwrite(code,sizeof(unsigned char),length,f) != length)
      {
	fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
	perror(">E ");
	exit(1);
      }
    }
}

/**************************************************************************/

static void
write_dual_planar_code(FILE *f, int doflip, int doprev)

/* Write the dual in planar_code format.  Always write in next direction,
   and if doflip != 0 also write in prev direction. */
{
  size_t length;
  unsigned char code[MAXF+MAXE+1];

  length=eulerc+ne+(ne/2)-nv;
  compute_dual_code(code);
  if (fwrite(code,sizeof(unsigned char),length,f) != length)
    {
      fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
      perror(">E ");
      exit(1);
    }
  if (doflip)
    { compute_dual_code_mirror(code);
    if (fwrite(code,sizeof(unsigned char),length,f) != length)
      {
	fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
	perror(">E ");
	exit(1);
      }
    }
}

/**************************************************************************/

static void
write_digits(FILE *f, int doflip, int doprev)
 
/* Write in alphabetic format.  Always write in next direction,
   and if doflip != 0 also write in prev direction.  This output
   procedure uses the internal numbering of vertices and is 
   intended for debugging purposes. */
{
    int i,k;
    EDGE *ex,*e;
    unsigned char code[2*MAXN+2*MAXE+9];
    int nvsize;
    int lastnum,w;
    int nbop,nbtot,nbpart;
    EDGE **nb0,**nb,**nblim;
#define CODE0(x) ((x)<10 ? '0'+(x) : 'a'+(x)-10)
 
    if (oneswitch)
    {
	nbop = gotone_nbop;
	nbtot = gotone_nbtot;
	nbpart = (nbop==0 || nbop==nbtot ? nbtot : nbop);
	fprintf(f,"%d %d %d %d %d\n",nv,ne,(doflip?2:1),nbpart,nbtot);

	nb0 = (EDGE**) numbering[0];
	nblim = (EDGE**) numbering[doflip ? nbpart : nbtot];

	if (nblim != (EDGE**) numbering[1])
	    for (nb = nb0; nb < nblim; nb += MAXE)
	    {
		for (i = 0; i < ne; ++i)
		    fprintf(f," %c%c",CODE0(nb[i]->start),CODE0(nb[i]->end));
		fprintf(f,"\n");
	    }
    }
    
    if (nv >= 10) 
    {
        code[0] = '0' + nv/10;
        code[1] = '0' + nv%10;
        code[2] = ' ';
        nvsize = k = 3;
    }
    else
    {
        code[0] = '0' + nv;
        code[1] = ' ';
        nvsize = k = 2;
    }

    if (missing_vertex < 0) lastnum = nv-1;
    else                    lastnum = nv;
 
    for (i = 0; i <= lastnum; ++i)
    {
	if (i != missing_vertex)
	{
            e = ex = firstedge[i];
            do
            {
		w = e->end;
		code[k++] = CODE0(w);
                e = e->next;
            }
            while (e != ex);
            code[k++] = ',';
	}
    }
    code[k-1] = '\n';
 
    if (doflip)
    {
        for (i = 0; i <nvsize; ++i) code[k++] = code[i];
    
        for (i = 0; i <= lastnum; ++i)
        {
	    if (i != missing_vertex)
            {
                e = ex = firstedge[i];
                do
                {
                    w = e->end;
		    code[k++] = CODE0(w);
                    e = e->prev;
                }
                while (e != ex);
                code[k++] = ',';
	    }
        }
    }
    code[k-1] = '\n';
 
    if (fwrite(code,(size_t)1,(size_t)k,f) != (size_t)k)
    {
        fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
        perror(">E ");
        exit(1);
    }
}

/**************************************************************************/

static void
write_alpha(FILE *f, int doflip, int doprev)

/* Write in alphabetic format.  Write in next direction if doprev == 0,
   write in prev direction if doprev != 0,
   and if doflip != 0  write in next and prev direction. */
{
  int i, j, start;
  unsigned char code[MAXN+MAXE+4];
  unsigned char precode[MAXN+MAXE+1];
  size_t length;

  length=nv+ne;

  if (nv >= 10)
    {
      code[0] = '0' + nv/10;
      code[1] = '0' + nv%10;
      code[2] = ' ';
      length += 3;
      start=3;
    }
  else
    {
      code[0] = '0' + nv;
      code[1] = ' ';
      length += 2;
      start=2;
    }

  if (doflip || !doprev)
    { compute_code(precode);
      for (i = 1, j=start; j < length; ++i, ++j)
	if (precode[i]==0) code[j]=',';
	else code[j]=precode[i]-1+'a';
      code[j-1]='\n';
      if (fwrite(code,sizeof(unsigned char),length,f) != length)
	{
	  fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
	  perror(">E ");
	  exit(1);
	}
    }
  if (doflip || doprev)
    { compute_code_mirror(precode);
      for (i = 1, j=start; j < length; ++i, ++j)
	if (precode[i]==0) code[j]=',';
      else code[j]=precode[i]-1+'a';
      code[j-1]='\n';
      if (fwrite(code,sizeof(unsigned char),length,f) != length)
	{
	  fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
	  perror(">E ");
	  exit(1);
	}
    }
}

/**************************************************************************/

static void
write_dual_alpha(FILE *f, int doflip, int doprev)

/* Write the dual in alphabetic format.  Always write in next direction,
   and if doflip != 0 also write in prev direction. */
{
  int i,j,start;
  unsigned char code[MAXN+MAXE+4];
  unsigned char precode[MAXN+MAXE+1];
  size_t length;

  length=eulerc+ne+(ne/2)-nv;
  if (orient)
    compute_dual_code(precode);
  else
    compute_dual_code_non_or(precode);

  if (precode[0] >= 10)
    {
      code[0] = '0' + precode[0]/10;
      code[1] = '0' + precode[0]%10;
      code[2] = ' ';
      length += 3;
      start=3;
    }
  else
    {
      code[0] = '0' + precode[0];
      code[1] = ' ';
      length += 2;
      start=2;
    }

  for (i = 1, j=start; j < length; ++i, ++j)
    if (precode[i]==0) code[j]=',';
    else if (precode[i]<=MAXN)
      code[j]=precode[i]-1+'a';
    else
      code[j]=precode[i]-MAXN-1+'a';
  
  code[j-1]='\n';
  if (fwrite(code,sizeof(unsigned char),length,f) != length)
    {
      fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
      perror(">E ");
      exit(1);
    }
  if (doflip)
    { compute_dual_code_mirror(precode);
    for (i = 1, j=start; j < length; ++i, ++j)
      if (precode[i]==0) code[j]=',';
      else if (precode[i]<=MAXN)
	code[j]=precode[i]-1+'a';
      else
	code[j]=precode[i]-MAXN-1+'a';
    code[j-1]='\n';
    if (fwrite(code,sizeof(unsigned char),length,f) != length)
      {
	fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
	perror(">E ");
	exit(1);
      }
    }
}

/**************************************************************************/

static void
write_ualpha(FILE *f, int doflip, int doprev)

/* Write in alphabetic format with uppercase.  
   Write in next direction if doprev == 0,
   write in prev direction if doprev != 0,
   and if doflip != 0  write in next and prev direction. */
{
  int i, j, start;
  unsigned char code[MAXN+MAXE+4];
  unsigned char precode[MAXN+MAXE+1];
  size_t length;

  length=nv+ne;

  if (nv >= 10)
    {
      code[0] = '0' + nv/10;
      code[1] = '0' + nv%10;
      code[2] = ' ';
      length += 3;
      start=3;
    }
  else
    {
      code[0] = '0' + nv;
      code[1] = ' ';
      length += 2;
      start=2;
    }

  if (doflip || !doprev)
    { if (orient)
	compute_code(precode);
      else
	compute_code_non_or(precode);
    for (i = 1, j=start; j < length; ++i, ++j)
      if (precode[i]==0) code[j]=',';
      else if (precode[i]<=MAXN)
	code[j]=precode[i]-1+'a';
      else
	code[j]=precode[i]-MAXN-1+'A';
    code[j-1]='\n';
    if (fwrite(code,sizeof(unsigned char),length,f) != length)
      {
	fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
	perror(">E ");
	exit(1);
      }
    }
  if (doflip || doprev)
    { if (orient)
	compute_code_mirror(precode);
      else
	compute_code_mirror_non_or(precode);
    for (i = 1, j=start; j < length; ++i, ++j)
      if (precode[i]==0) code[j]=',';
      else if (precode[i]<=MAXN)
	code[j]=precode[i]-1+'a';
      else
	code[j]=precode[i]-MAXN-1+'A';
    code[j-1]='\n';
    if (fwrite(code,sizeof(unsigned char),length,f) != length)
      {
	fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
	perror(">E ");
	exit(1);
      }
    }
}

/**************************************************************************/

static void
write_dual_ualpha(FILE *f, int doflip, int doprev)

/* Write the dual in alphabetic format with uppercase.  
   Always write in next direction,
   and if doflip != 0 also write in prev direction. */
{
  int i,j,start;
  unsigned char code[MAXN+MAXE+4];
  unsigned char precode[MAXN+MAXE+1];
  size_t length;

  length=eulerc+ne+(ne/2)-nv;
  if (orient)
    compute_dual_code(precode);
  else
    compute_dual_code_non_or(precode);
  if (precode[0] >= 10)
    {
      code[0] = '0' + precode[0]/10;
      code[1] = '0' + precode[0]%10;
      code[2] = ' ';
      length += 3;
      start=3;
    }
  else
    {
      code[0] = '0' + precode[0];
      code[1] = ' ';
      length += 2;
      start=2;
    }

  for (i = 1, j=start; j < length; ++i, ++j)
    if (precode[i]==0) code[j]=',';
    else if (precode[i]<=MAXN)
      code[j]=precode[i]-1+'a';
    else
      code[j]=precode[i]-MAXN-1+'A';

  code[j-1]='\n';
  if (fwrite(code,sizeof(unsigned char),length,f) != length)
    {
      fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
      perror(">E ");
      exit(1);
    }
  if (doflip)
    { 
      /* with doflip must be orientable */

      compute_dual_code_mirror(precode);

      for (i = 1, j=start; j < length; ++i, ++j)
	if (precode[i]==0) code[j]=',';
	else if (precode[i]<=MAXN)
	  code[j]=precode[i]-1+'a';
	else
	  code[j]=precode[i]-MAXN-1+'A';
      code[j-1]='\n';
      if (fwrite(code,sizeof(unsigned char),length,f) != length)
	{
	  fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
	  perror(">E ");
	  exit(1);
	}
    }
}

/**************************************************************************/

static void
write_code_as_sparse6(FILE *f, unsigned char code[])

/* Write a graph represented in planar_code to a file in sparse6 format.
   The imbedding is lost but the vertex numbering is the same.  This
   goes not use any global variables and works to 255 vertices. */

{
  register unsigned char *pin,*pout;
  unsigned char s6[20+2*MAXE+2*MAXF];
  int n,nb,i,j,lastj,x,k,r,rr,topbit;
  int loopcount;

  pin = code;
  n = *pin++;
  pout = s6;
  *pout++ = ':';

  if (n <= 62)
    *pout++ = 63 + n;
  else
    {
      *pout++ = 63 + 63;
      *pout++ = 63 + 0;
      *pout++ = 63 + (n >> 6);
      *pout++ = 63 + (n & 0x3F);
    }

  for (i = n-1, nb = 0; i != 0 ; i >>= 1, ++nb) {}
  topbit = 1 << (nb-1);
  k = 6;
  x = 0;

  lastj = 0;
  for (j = 0; j < n; ++j)
    {
      loopcount = 0;   /* The input code shows loops once from each end,
			  but we want each loop just once in sparse6. */
      while ((i = *pin++) != 0)
        {
	  --i;
	  if (i < j || (i == j && ((++loopcount)&1)))
            {
	      if (j == lastj)
                {
		  x <<= 1;
		  if (--k == 0)
                    {
		      *pout++ = 63 + x;
		      k = 6;
		      x = 0;
                    }
                }
	      else
                {
		  x = (x << 1) | 1;
		  if (--k == 0)
                    {
		      *pout++ = 63 + x;
		      k = 6;
		      x = 0;
                    }
		  if (j > lastj+1)
                    {
		      for (r = 0, rr = j; r < nb; ++r, rr <<= 1)
                        {
			  if (rr & topbit) x = (x << 1) | 1;
			  else             x <<= 1;
			  if (--k == 0)
                            {
			      *pout++ = 63 + x;
			      k = 6;
			      x = 0;
                            }
                        }
		      x <<= 1;
		      if (--k == 0)
                        {
			  *pout++ = 63 + x;
			  k = 6;
			  x = 0;
                        }
                    }
		  lastj = j;
                }
	      for (r = 0, rr = i; r < nb; ++r, rr <<= 1)
                {
		  if (rr & topbit) x = (x << 1) | 1;
		  else             x <<= 1;
		  if (--k == 0)
                    {
		      *pout++ = 63 + x;
		      k = 6;
		      x = 0;
                    }
                }
            }
        }
    }

  if (k != 6) *pout++ = 63 + ((x << k) | ((1 << k) - 1));

  *pout++ = '\n';
  k = pout - s6;

  if (fwrite(s6,sizeof(unsigned char),(size_t)k,f) != k)
    {
      fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
      perror(">E ");
      exit(1);
    }
}

/**************************************************************************/

static void
write_sparse6(FILE *f, int doflip, int doprev)

/* Write in sparse6 format.  doflip is ignored. */
{
  unsigned char code[MAXN+MAXE+1];

  compute_code(code);
  write_code_as_sparse6(f,code);
}

/**************************************************************************/

static void
write_dual_sparse6(FILE *f, int doflip, int doprev)

/* Write dual cubic graph in sparse6 format.  doflip is ignored. */
{
  unsigned char code[MAXF+MAXE+1];

  compute_dual_code(code);
  write_code_as_sparse6(f,code);
}


/**************************************************************************/

static void
write_code_as_graph6(FILE *f, unsigned char code[])

/* Write a graph represented in planar_code to a file in graph6 format.
   The imbedding is lost, loops are lost, and multiple edges are changed
   to one edge.  The vertex numbering is the same.  This goes not use any
   global variables and works to 255 vertices. */

{
  unsigned char g6[20+MAXF*(MAXF-1)/12];
  register unsigned char *pin,*pout;
  int n,nlen,bodylen,i,j,org;
  static unsigned char g6bit[] = {32,16,8,4,2,1};

  pin = code;
  n = *pin++;

  if (n <= 62)
    {
      g6[0] = 63 + n;
      nlen = 1;
    }
  else
    {
      g6[0] = 63 + 63;
      g6[1] = 63 + 0;
      g6[2] = 63 + (n >> 6);
      g6[3] = 63 + (n & 0x3F);
      nlen = 4;
    }

  pout = g6 + nlen;
  bodylen = ((n * (n-1)) / 2 + 5) / 6;
  for (i = 0; i < bodylen; ++i) pout[i] = 0;
  pout[bodylen] = '\n';

  for (i = 0, org = -1; i < n; org += i, ++i)
    {
      while ((j = *pin++) != 0)
        {
	  if (j <= i)
            {
	      j += org;
	      pout[j/6] |= g6bit[j%6];
            }
        }
    }

  for (i = 0; i < bodylen; ++i) pout[i] += 63;

  j = nlen + bodylen + 1;
  if (fwrite(g6,sizeof(unsigned char),(size_t)j,f) != j)
    {
      fprintf(stderr,">E %s: fwrite() failed\n",cmdname);
      perror(">E ");
      exit(1);
    }
}

/**************************************************************************/

static void
write_graph6(FILE *f, int doflip, int doprev)

/* Write in graph6 format.  doflip is ignored. */
{
  unsigned char code[MAXN+MAXE+1];

  compute_code(code);
  write_code_as_graph6(f,code);
}

/**************************************************************************/

static void
write_dual_graph6(FILE *f, int doflip, int doprev)

/* Write dual cubic graph in graph6 format.  doflip is ignored. */
{
  unsigned char code[MAXF+MAXE+1];

  compute_dual_code(code);
  write_code_as_graph6(f,code);
}

/**************************************************************************/

static void
check_it(int code)

/* Checks these properties:
   1. Degrees are >= 3 and correct.
   2. Faces are triangles or quadrangles.
   3. start and end fields are correct.

   5. vertex numbers are in range.
   6. adj array is correct.
   7. nnei array is correct.
   8. ncontract arrary is correct.
   9. STARN(i), i>=nv are correct.
*/

/* general: code 1-299 for general graph scan after startpolyscan */
/* general_split: code 300-399 for triangulations with holes 
   during vertex splitting */

{
  int i,j,k,nadj,nn,ncont,general,general_split;
  EDGE *e, *e1;
  register EDGE *si;

  //if (FALSE) return;
  if (TRUE) return;

  //debug_count++;
  //fprintf(outfile,"debug_count %d, code = %d\n",debug_count,code);

  /* dump triangulation */

  if (FALSE) {
    //if (TRUE) {
    fprintf(outfile,"adjacency matrix:\n");
    for (i=0; i<nv; i++) {
      for (j=0; j<nv; j++) 
	fprintf(outfile,"%d",adj[i][j]);
      fprintf(outfile,"\n");
    }

    for (i=0; i<nv; i++) {
      fprintf(outfile,"vertex %d, degree %d\n",i,degree[i]);
      fprintf(outfile,"clockwise:\n");
      e = firstedge[i];
      for (j = 0; j < degree[i]+1; ++j)
	{
	  fprintf(outfile,"neighor %d, start %d, end %d\n",
		  j,e->start,e->end);
	  e = e->next;
	}

      fprintf(outfile,"counterclockwise:\n");
      e = firstedge[i];
      for (j = 0; j < degree[i]+1; ++j)
	{
	  fprintf(outfile,"neighor %d, start %d, end %d\n",
		  j,e->start,e->end);
	  e = e->prev;
	}
      fprintf(outfile,"oposite clockwise:\n");
      e = firstedge[i]->op;
      for (j = 0; j < degree[i]+1; ++j)
	{
	  fprintf(outfile,"neighor %d, start %d, end %d\n",
		  j,e->start,e->end);
	  e = e->next;
	}

      fprintf(outfile,"oposite counterclockwise:\n");
      e = firstedge[i]->op;
      for (j = 0; j < degree[i]+1; ++j)
	{
	  fprintf(outfile,"neighor %d, start %d, end %d\n",
		  j,e->start,e->end);
	  e = e->prev;
	}
    }
  }

  /* do checks */

  general = (code<300);
  general_split = (300<=code&&code<400);

  for (i = 0; i < nv; ++i)
    {
      if ((degree[i] < 3 && !qswitch && !general && !general_split) || 
	  (degree[i] < 2 && nv > 3 && !general && !general_split) || 
	  (degree[i] < 1))
        {
	  fprintf(stderr,">E degree error, code=%d\n",code);
	  exit(1);
        }

      e = firstedge[i];
      for (j = 0; j < degree[i]; ++j) e = e->next;
      if (e != firstedge[i])
        {
	  fprintf(stderr,">E next error, code=%d\n",code);
	  exit(1);
        }

      e = firstedge[i];
      for (j = 0; j < degree[i]; ++j) e = e->prev;
      if (e != firstedge[i])
        {
	  fprintf(stderr,">E prev error, code=%d\n",code);
	  exit(1);
        }

      e = firstedge[i];
      for (j = 0; j < degree[i]; ++j)
        {
	  e = e->next;
	  if (e->start != i)
            {
	      fprintf(stderr,">E start error, code=%d\n",code);
	      exit(1);
            }
	  if (e->end < 0 || e->end >= nv)
            {
	      fprintf(stderr,">E end label error, code=%d\n",code);
	      exit(1);
            }
	  if (e->end != e->invers->start)
            {
	      fprintf(stderr,">E invers label error, code=%d\n",code);
	      exit(1);
	    }
	  if (e->invers->end != i)
            {
	      fprintf(stderr,">E end error, code=%d\n",code);
	      exit(1);
            }
	  if (qswitch) {
	    if (e->invers->next == e
		|| e->invers->next->invers->next == e
		|| e->invers->next->invers->next->invers->next == e
		|| e->invers->next->invers->next->invers->next->invers->next != e)
	      {
		fprintf(stderr,">E quad face error, code=%d\n",code);
		exit(1);
	      }
	  }
	  else if (!general && !general_split) {
	    if (e->invers->next == e
		|| e->invers->next->invers->next == e
		|| e->invers->next->invers->next->invers->next != e)
	      {
		fprintf(stderr,">E tri face error, code=%d\n",code);
		exit(1);
	      }
	  }
	  else if (!general_split) {
	    if (e->left_facesize < 3)
	      {
		fprintf(stderr,">E small left face error, code=%d\n",code);
		exit(1);
	      }
	    e1 = e;
	    for (k=0; k<(e->left_facesize)-1; k++) {
	      e1 = e1->invers->next;
	      if (e1 == e)
		{
		  fprintf(stderr,">E left face error 1, code=%d\n",code);
		  exit(1);
		}
	      if (e1->left_facesize != e->left_facesize)
		{
		  fprintf(stderr,">E left face error 2, code=%d\n",code);
		  exit(1);
		}
	    }
	    e1 = e1->invers->next;
	    if (e1 != e)
	      {
		fprintf(stderr,">E left face error 3, code=%d\n",code);
		exit(1);
	      }
	  }
	  
	  if (!adj[i][e->end])
	    {
	      fprintf(stderr,">E adj[%d][%d] array error, code=%d\n",
		      i,e->end,code);
	      exit(1);
	    }
        }
    }

  nadj = 0;
  for (i=0;i<nv;i++)
    {
      ncont = 0;
      for (j=0;j<nv;j++)
	{
	  if (adj[i][j]) nadj++;
	  if (adj[i][j] && nnei[i][j] == 2) {
	    if (!irrhole)
	      ncont++;
	    else {
	      for (e = firstedge[i], k = 0;
		   e->end != j && k < degree[i]; 
		   e = e->next, k++) {}
	      if (e->end != j)
		{
		  fprintf(stderr,">E adj[%d][%d] array error 2, code=%d\n",
			  i,j,code);
		  exit(1);
		}
	      if (!RIGHTHOLE(e) && !RIGHTHOLE(e->invers))
		ncont++;
	    }
	  }
	  if (i != j)
	    {
	      nn = 0;
	      for (k=0;k<nv;k++)
		if (adj[i][k] && adj[j][k]) nn++;
	      if (irrhole && adj[i][j])
		{
		  for (e = firstedge[i], k = 0;
		       e->end != j && k < degree[i]; 
		       e = e->next, k++) {}
		  if (e->end != j)
		    {
		      fprintf(stderr,">E adj[%d][%d] array error 3, code=%d\n",
			      i,j,code);
		      exit(1);
		    }
		  if (RIGHTHOLE(e)) nn++;
		  if (RIGHTHOLE(e->invers)) nn++;
		}
	      if (nn != nnei[i][j] && !general)
		{
		  fprintf(stderr,">E nnei[%d][%d] = %d (nn=%d) error, code=%d\n",
			  i,j,nnei[i][j],nn,code);
		  exit(1);
		}
	    }
	}
      if (ncont != ncontract[i] && !qswitch && !general)
	{
	  fprintf(stderr,">E ncontract error (ncont=%d) (ncontract[%d]=%d), code=%d\n",
		  ncont,i,ncontract[i],code);
	  exit(1);
	}
    }
  if (nadj != ne && !general)
    {
      fprintf(stderr,">E adj count (%d) != ne (%d) error, code=%d\n",
	      nadj,ne,code);
      exit(1);
    }


  if (minimumdeg == 3 && !qswitch)
    for (i=nv;i<MAXN;i++)
      {
	si = STARN(i);
	for (j=0;j<MAXN;++j)
	  {
	    e = si + j;
	    e1 = si + j + MAXN;
	    if (e->end != i ||
		e1->start != i ||
		e->invers != e1 ||
		e1->invers != e ||
		e1->next != si + MAXN + (j+1) % MAXN ||
		e1->prev != si + MAXN + (j+MAXN-1) % MAXN)
	      {
		fprintf(stderr,">E STARN(%d) (j=%d) error, code=%d\n",
			i,j,code);
		exit(1);
	      }
	  }
      }
  
  
  for (i = 0; i < nv; ++i)
    {
      e = firstedge[i]->op;
      for (j = 0; j < degree[i]; ++j) e = e->next;
      if (e != firstedge[i]->op)
        {
	  fprintf(stderr,">E next error (op), code=%d\n",code);
	  exit(1);
        }

      e = firstedge[i]->op;
      for (j = 0; j < degree[i]; ++j) e = e->prev;
      if (e != firstedge[i]->op)
        {
	  fprintf(stderr,">E prev error (op), code=%d\n",code);
	  exit(1);
        }

      e = firstedge[i]->op;
      for (j = 0; j < degree[i]; ++j)
        {
	  e = e->next;
	  if (e->start != i)
            {
	      fprintf(stderr,">E start error (op), code=%d\n",code);
	      exit(1);
            }
	  if (e->end < 0 || e->end >= nv)
            {
	      fprintf(stderr,">E end label error (op), code=%d\n",code);
	      exit(1);
            }
	  if (e->end != e->invers->start)
            {
	      fprintf(stderr,">E invers label error (op), code=%d\n",code);
	      exit(1);
	    }
	  if (e->invers->end != i)
            {
	      fprintf(stderr,">E end error (op), code=%d\n",code);
	      exit(1);
            }
	  
	  if (qswitch) {
	    if (e->invers->next == e
		|| e->invers->next->invers->next == e
		|| e->invers->next->invers->next->invers->next == e
		|| e->invers->next->invers->next->invers->next->invers->next != e)
	      {
		fprintf(stderr,">E quad face error (op), code=%d\n",code);
		exit(1);
	      }
	  }
	  else if (!general && !general_split) {
	    if (e->invers->next == e
		|| e->invers->next->invers->next == e
		|| e->invers->next->invers->next->invers->next != e)
	      {
		fprintf(stderr,">E tri face error (op), code=%d\n",code);
		exit(1);
	      }
	  }
	  else if (!general_split) {
	    if (e->left_facesize < 3)
	      {
		fprintf(stderr,">E small left face error (op), code=%d\n",code);
		exit(1);
	      }
	    e1 = e;
	    for (k=0; k<(e->left_facesize)-1; k++) {
	      e1 = e1->invers->next;
	      if (e1 == e)
		{
		  fprintf(stderr,">E left face error 1 (op), code=%d\n",code);
		  exit(1);
		}
	      if (e1->left_facesize != e->left_facesize)
		{
		  fprintf(stderr,">E left face error 2 (op), code=%d\n",code);
		  exit(1);
		}
	    }
	    e1 = e1->invers->next;
	    if (e1 != e)
	      {
		fprintf(stderr,">E left face error 3 (op), code=%d\n",code);
		exit(1);
	      }
	  }
	  
	  if (!adj[i][e->end])
	    {
	      fprintf(stderr,">E adj[%d][%d] array error (op), code=%d\n",i,e->end,code);
	      exit(1);
	    }
        }
    }
  
  
  if (minimumdeg == 3 && !qswitch)
    for (i=nv;i<MAXN;i++)
      {
	si = STARN(i) + NUMEDGES;
	for (j=0;j<MAXN;++j)
	  {
	    e = si + j;
	    e1 = si + j + MAXN;
	    if (e->end != i ||
		e1->start != i ||
		e->invers != e1 ||
		e1->invers != e ||
		e1->prev != si + MAXN + (j+1) % MAXN ||
		e1->next != si + MAXN + (j+MAXN-1) % MAXN)
	      {
		fprintf(stderr,">E STARN(%d) (j=%d) error (op), code=%d\n",
			i,j,code);
		exit(1);
	      }
	  }
      }

  for (i=nv;i<maxnv;i++) {
    /*
    if (degree[i] != 0) {
      fprintf(stderr,">E degree[%d]=%d not 0 error, code=%d\n",i,degree[i],code);
      exit(1);
    }
    */
    for (j=0;j<maxnv;j++) {
      if (adj[i][j] != 0) {
	fprintf(stderr,">E adj[%d][%d]=%d not 0 error, code=%d\n",i,j,adj[i][j],code);
	exit(1);
      }
      /*
      if (nnei[i][j] != 0) {
	fprintf(stderr,">E nnei[%d][%d]=%d not 0 error, code=%d\n",i,j,nnei[i][j],code);
	exit(1);
      }
      */
    }
  }
}

/**************************************************************************/

static void
got_one(int nbtot, int nbop, int connec)
  
/* This is called when a complete triangulation is formed.  The main
   purpose is to write the triangulation and to collect some stats. */
{
  int doflip,wt;
#ifdef STATS
  int numroot;
#ifdef STATS2
  int num2,i;
#endif
#endif
  
  if (xswitch && connec != xconnec) return;
  
  doflip = oswitch && (nbop == nbtot || nbop == 0);
  wt = doflip ? 2 : 1;
  
#ifdef FILTER
  if (!FILTER(nbtot,nbop,doflip,connec)) return;
#endif
  
  ADDBIG(nout[connec],1);
  if (oswitch) ADDBIG(nout_op[connec],wt);
  
  if (pswitch)
    {
      if (oswitch) ADDBIG(nout_e_op[ne/2],wt);
      ADDBIG(nout_e[ne/2],1);
    }
  
#ifdef STATS
  if (polygonsize < 0)
    {
      numroot = wt * numedgeorbits(nbtot,nbop);
      ADDBIG(numrooted,numroot);
      if (pswitch) ADDBIG(numrooted_e[ne/2],numroot);
      if (degree[nv-1] < 6) ADDBIG(nummindeg[degree[nv-1]],wt);
    }
  else
    {
      numroot = wt * numorbitsonface(nbtot,nbop,code_edge);
      ADDBIG(numrooted,numroot);
      ADDBIG(numbigface[degree[missing_vertex]],numroot);
    }
  if (needgroup && nbtot == 1) ADDBIG(ntriv,wt);
#ifdef STATS2
  num2 = 0;
  for (i = 0; i < nv; ++i) if (degree[i] == 2) ++num2;
  ADDBIG(numtwos[num2],wt);
#endif
#endif

  if (!uswitch)
    {
      if (qswitch)
	code_edge = firstedge[nv-1];
      gotone_nbop = nbop;
      gotone_nbtot = nbtot;
      if (dswitch) (*write_dual_graph)(outfile,doflip,nbop==0);
      else         (*write_graph)(outfile,doflip,nbop==0);
    }
  
}

/***************************************************************/

static void
insert_edge_tri(EDGE *e)

/* inserts the edge previously deleted -- e may not have been modified in
   the meantime and the map must look exactly like it looked before deleting 
   it. On the left hand side of e there must be a triangle after inserting e */
{
    int newfacesize;
    EDGE *run;

    ne +=2;

    adj[e->start][e->end] = adj[e->end][e->start] = 1;

    e->prev->next=e;
    e->op->next->prev=e->op;
    e->next->prev=e;
    e->op->prev->next=e->op;
    (degree[e->start])++;

    e=e->invers;
    
    e->prev->next=e;
    e->op->next->prev=e->op;
    e->next->prev=e;
    e->op->prev->next=e->op;
    (degree[e->start])++;

    run=e->prev->invers;
	/* now the face on the right of e when entering this function is
           on the left of run */
    newfacesize=e->left_facesize;

    for (run=e->prev->invers; run != e; run=run->prev->invers) {
      run->left_facesize=newfacesize; 
      run->op->invers->left_facesize=newfacesize; 
    }

    run=e->next;
    run->left_facesize=3;
    run->op->invers->left_facesize=3;
    run->invers->next->left_facesize=3;
    run->invers->next->op->invers->left_facesize=3;

    check_it(112);

}

/************************************************************************/

static int
threeconn(EDGE *e)

/* tests whether the graph obtained by deleting EDGE e is still 3-connected.
   The edge e may have been deleted or not, but the values in e must be
   as before it was (possibly) deleted. The map must have been 3-connected
   before -- so especially there weren't any vertices of degree 2.
   degree[] is not assumed correct for the endvertices of e.

   On the left hand side of e there must be a triangle
   (e->left_facesize==3) and it is assumed that it is checked before
   that the endvertices of e have degree at least 3 after the deletion.

   If there is a 2-cut, e->start and e->end cannot be contained, but they
   must be in different components, so v=e->prev->end MUST be contained.
   It is checked whether v is contained in a face that shares yet another
   vertex with the face formerly on the right hand side of e (the new face
   obtained by deleting e).

   Returns 1 if it is 3-connected after deleting e, 0 else.  */

{
    EDGE *run, *start, *end;

    start=e->prev->invers;
    if (degree[start->start]==3) return 1;

    RESETMARKS_V;

    /* The endvertices of e need not be marked */
    for ( run=e->next, end=e->invers->prev->invers ;
          run != end; run=run->invers->next) MARK_V(run->end);

    
    end=start->prev->prev; /* stops the running around the vertex before
			      the last face */

/* The first face and the last face contain also one of the endvertices of e,
   so if they also contain marked vertices, then there already was a 2-cut. */

    start=start->next;

    while (start != end)
      { run=start->invers;
        start=start->next;
        for ( ; run != start; run=run->prev->invers)
	    if (ISMARKED_V(run->start)) return 0; 
      }

    return 1;
}

/************************************************************************/

static int
twoconn(EDGE *e)

/* tests whether the graph obtained by deleting EDGE e is still 2-connected.
   The edge e may have been deleted or not, but the values in e must be
   as before it was (possibly) deleted. The map must have been 2-connected
   before.  degree[] is not assumed correct for the endvertices of e.

   On the left hand side of e there must be a triangle (e->left_facesize==3).

   If there is a 1-cut, it cannot be e->start or e->end (otherwise
   it was a 1-cut before), but they must be in different components, (same
   reason), so v=e->prev->end MUST be the cutvertex.
   It is checked whether v is contained in the face on the right hand side
   of e (before deleting). 

   Returns 1 if it is 2-connected after deleting e, 0 else.
*/

{
    EDGE *run, *end;
    int v;

    end = e->next;
    if (end == e->prev) return 0;
    end = end->invers;

    v = e->prev->end;
    if (degree[v] == 2) return 1;

    for (run = e->invers->prev; run != end; run = run->invers->prev)
        if (run->end == v) return 0;

   return 1;
}

/************************************************************************/

static void
prune_poly_edgelist(EDGE *old[], int numold, EDGE *newe[], int *numnew)

/* Copy from old[0..numold-1] to newe[0..*numnew-1] each edge e with
   these two properties:
   1. Both ends of e have degree >= minpolydeg+1.
   2. e is contained in a 3-face.
   It is legal that old and newe and &numold and numnew are the same.

   17Aug2005: Changed "new" to "newe" to avoid Codewarrior bug.
*/

{
    int i,counter=0;
    EDGE *e;

    for (i=0; i<numold; i++, old++)
      { e = *old; 
        if (degree[e->start] > minpolydeg && degree[e->end] > minpolydeg
	     && (e->left_facesize == 3 || e->invers->left_facesize == 3))
            newe[counter++]=e;
      }

    *numnew=counter;
}

/**************************************************************************/

static int
make_colours(int col[], EDGE *e3)
/* Make better colours for maxdeg=3, supposing that expand3() has been
   performed at position e3 (though it hasn't been yet).
   If the virtual new vertex, nv, is not best, return 0.
   Otherwise, return the number of vertices with the same
   colour as nv (including itself).  Note that for correct
   operation, col[nv-1] must have the smallest value in col[]
   and all colours must be positive. */
{
  register int i,c,c0,nc;
  register EDGE *e;
  register int v0,v1,v2;

  v0 = e3->start;
  v1 = e3->end;
  v2 = e3->next->end;

  c0 =  (1 << ((++degree[v0])&7))
    + (1 << ((++degree[v1])&7))
    + (1 << ((++degree[v2])&7));

  col[nv] = 2;
  nc = 1;

  for (i = nv; --i >= 0;)
    {
      if (degree[i] != 3)
	col[i] = degree[i];
      else
        {
	  e = firstedge[i];
	  c = (1 << (degree[e->end]&7))
	    + (1 << (degree[e->next->end]&7))
	    + (1 << (degree[e->next->next->end]&7));
	  if (c > c0)
            {
	      --degree[v0];
	      --degree[v1];
	      --degree[v2];
	      return 0;
            }
	  else if (c == c0)
            {
	      col[i] = 2;
	      ++nc;
            }
	  else
	    col[i] = 3;
        }
    }

  --degree[v0];
  --degree[v1];
  --degree[v2];

  return nc;
}

/*************************************************************************/

static void
mark_edge_orbits(EDGE *edge[], int count, int minimal[], int nbtot)

/* edge[0..count-1] is a list of edges.  Put minimal[i] = 1 if
   edge[i] is the first appearance of its orbit, and  minimal[i] = 0
   if not.  Orbits on undirected edges are considered. */
{
    int i;
    EDGE **nb0,**nb,**nblim;
 
    if (nbtot == 1) 
        for (i = 0; i < count; ++i) minimal[i] = 1;
    else
    {
        nb0 = (EDGE**)numbering[0];
        nblim = (EDGE**)numbering[nbtot];
        for (i = 0; i < ne; ++i) nb0[i]->op->index = nb0[i]->index = i;
 
        RESETMARKS;

        for (i = 0; i < count; ++i)
        {
            nb = nb0 + edge[i]->index;
            if (!ISMARKEDLO((*nb)) &&
		!ISMARKEDLO((*nb)->invers) &&
		!ISMARKEDLO((*nb)->op) &&
		!ISMARKEDLO((*nb)->op->invers))
            {
                minimal[i] = 1;
                for ( ; nb < nblim; nb += MAXE) MARKLO((*nb));
            }
            else
                minimal[i] = 0;
        }

    }
}

/*************************************************************************/

static int
maybe_delete_c3(EDGE *edel, int oldmaxface, int oldmaxlist0, int oldmaxlist1,
                int *newmaxface,int *newmaxlist0,int *newmaxlist1,
                EDGE *good_or[],int *ngood_or,int *ncan_or,
                EDGE *good_inv[],int *ngood_inv,int *ncan_inv)

/* Assumes there is a 3-face on the left of *edel, and that *edel can
   be deleted while keeping degrees >= 3.  Also, the new face created will
   be a largest face.  oldmaxface is the size of the largest face so far,
   and inmaxface[oldmaxlist0..oldmaxlist1-1] are the edges with a face of
   maximum size on the left.

   This procedure deletes *edel provided the result is 3-connected, and that
   the procedure believes the re-insertion of *edel might be a canonical
   edge insertion.  The latter decision is based on four combinatorial
   invariates: the three vertex degrees at the ends of *edel and on its left,
   and the size of the face to the left of edel->prev.

   In case *edel passes the test and is deleted, the edges which may 
   represent the re-insertion of *edel and optimise the four-part invariant
   mentioned above are put in good_or[0..*ncan_or-1] and
   good_inv[0..*ncan_inv-1].  This will include at least one of the edges
   edel->prev, edel->invers->next and (if there is also a 3-face on the
   right of *edel) edel->next and edel->invers->prev.  Then all other edges
   which might possibly represent canonical edge insertions are put in 
   good_or[*ncan_or..*ngood_or-1] or good_inv[*ncan_inv..*ngood_inv-1].
   The *_or edges are those for which the inserted edge will be in the
   next direction, and the *_inv edges .. the prev direction.

   In addition, if *edel is deleted, inmaxface[*newmaxlist0..*newmaxlist1-1]
   will have all edges with a maximum face on the left (that size being put
   into *newmaxface).  This list may overlap
   inmaxface[oldmaxlist0..oldmaxlist1-1] if the max face size does not
   increase.

   Return values:   0 = ok
                    1 = rejected by threeconn()
		    2 = rejected by colour

   This version is for 3-connected only!  maybe_delete() is the more
   general version that works for connectivity 1 or more.
*/

{
#define DEGENDC3(ed) (degree[ed->end])
#define REJECTC3 {++degree[v1]; ++degree[v2]; return 2;}
#define ORTESTC3(e) {col = DEGENDC3(e); \
 if (col > maxcol) REJECTC3 else if (col == maxcol) good_or[ng_or++] = e;}
#define INVTESTC3(e) {col = DEGENDC3(e); \
 if (col > maxcol) REJECTC3 else if (col == maxcol) good_inv[ng_inv++] = e;}
#define ORCOLC3(ed) ((degree[ed->invers->prev->end] << 10) + ed->left_facesize)
#define INVCOLC3(ed) \
 ((degree[ed->invers->next->end] << 10) + ed->invers->left_facesize)

    EDGE *e1,*e2,*e3,*e4,*ee,*e,*ex;
    long col,maxcol;
    int k,i,v,ng_or,ng_inv,nc_or,nc_inv;
    int maxface,v1,v2,v3,v4,maxdeg,nml1,nml2,i0,i1;


    maxface = *newmaxface = edel->invers->left_facesize + 1;

    /* check if deleting edel would transform button holes into non button 
       holes.
       suppose edel=(a,b) is a non-border edge of a button hole, {a,b,c} is a 
       3-cycle, and (a,c) and (b,c) are border edges of button hole. if edel 
       is deleted then it could be reinserted inside the former button hole.
       let maxtransface be the maximum size of these transformed faces.
       if maxtransface < maxface then any transformed button holes may be 
       ignored.  
       if maxtransface > maxface then reject edel.  
       if maxtransface == maxface then any edges of transformed faces with 
       maxface edges must be added to inmaxface for checking later in this 
       routine.  when maxface==oldmaxface there is no easy way to tell if edel
       is the first edge to transform a button hole.  so we add all edges of 
       button holes of size maxface when maxface is first increased 
       (maxface>oldmaxface).
    */
    
    if (irrhole) {
      e1 = edel->next;
      do 
	{
	  if (e1->invers->left_facesize > 3) {
	    if (e1->invers->left_facesize > maxface)
	      if (e1->invers->prev->end == edel->end ||
		  e1->next->invers->next->end == edel->end)
		return 2;
	  }
	  e1 = e1->next;
	} while (e1 != edel);
    }
    
    if (maxface > oldmaxface) oldmaxlist0 = oldmaxlist1;
    *newmaxlist0 = oldmaxlist0;
    
    *ngood_or = *ngood_inv = 0;
    
 /* Now old edges that will be on a largest face (if *edel is deleted)
    are in places oldmaxlist0..oldmaxlist1-1.  New edges will be from
    oldmaxlist1 to *newmaxlist1-1, but *newmaxlist1 is not set yet. */

    v1 = edel->start;  v2 = edel->end;
    maxdeg = (degree[v1] > degree[v2] ? degree[v1] : degree[v2]) - 1;

    e1 = edel->prev;  v3 = e1->end;
    e4 = edel->next; v4 = e4->end;

    if (degree[v3] > maxdeg || degree[v4] > maxdeg) return 2;


    --degree[v1]; --degree[v2];
    e2 = edel->invers->next;
    e3 = edel->invers->prev;

    maxcol = nc_or = nc_inv = 0;

    if (degree[v1] == maxdeg)
    {
        col = DEGENDC3(e1);
	maxcol = col;
        good_or[nc_or++] = e1;

        if (maxface == 4)
	{
	    col = DEGENDC3(e4);
	    if (col > maxcol)
	    {
		nc_or = nc_inv = 0;
		good_inv[nc_inv++] = e4;
		maxcol = col;
	    }
	    else if (col == maxcol)
	        good_inv[nc_inv++] = e4;
	}
    }

    if (degree[v2] == maxdeg)
    {
	col = DEGENDC3(e2);
	if (col > maxcol)
  	{
	    nc_or = nc_inv = 0;
            good_inv[nc_inv++] = e2;
	    maxcol = col;
	}
	else if (col == maxcol)
	    good_inv[nc_inv++] = e2;
	
	if (maxface == 4)
	{
            col = DEGENDC3(e3);
	    if (col > maxcol)
	    {
		nc_or = nc_inv = 0;
		good_or[nc_or++] = e3;
		maxcol = col;
	    }
	    else if (col == maxcol)
		good_or[nc_or++] = e3;
	}
    }

    ng_or = nc_or;  ng_inv = nc_inv;

    if (maxface > 4)
    {
	if (degree[v2] == maxdeg) ORTESTC3(e3);
	if (degree[v1] == maxdeg) INVTESTC3(e4);
    }

    if (degree[v3] == maxdeg)
    {
	ORTESTC3(e2->invers);
	INVTESTC3(e1->invers);
    }

    nml1 = oldmaxlist1;

    v = e3->end;
    ee = e3->invers;
    for (;;)
    {
	inmaxface[nml1++] = ee;
	if (degree[v] > maxdeg)
	{
	    REJECTC3
	}
	else if (degree[v] == maxdeg)
	{
	    INVTESTC3(ee);
	    ee = ee->prev;
	    ORTESTC3(ee);
	}
	else
	    ee = ee->prev;

	if (v == v4) break;

	v = ee->end;
	ee = ee->invers;
    }
     
    inmaxface[nml1++] = e2;
    inmaxface[nml1++] = e1->invers;
    inmaxface[nml1++] = e4;

    /* add edges of button holes of size maxface if not already on list */

    nml2 = nml1;
    if (maxface > oldmaxface && irrhole == 1) 
      for (i = 0; i < nv; ++i)
	{
	  e = ex = firstedge[i];
	  do
	    {
	      if (e->left_facesize == maxface)
		inmaxface[nml2++] = e;
	      e = e->next;
	    } while (e != ex);
	}

    /* Now test old edges still on max faces */
    /* Or test newly added button hole edges */
 
    if (maxface > oldmaxface && irrhole == 1) {
      i0 = nml1;
      i1 = nml2;
    }
    else {
      i0 = oldmaxlist0;
      i1 = oldmaxlist1;
    }

    for (i = i0; i < i1; ++i)
    {
	ee = inmaxface[i];
	if (degree[ee->start] > maxdeg)
	{
	    REJECTC3
	}
	else if (degree[ee->start] == maxdeg)
	{
	    INVTESTC3(ee);
	    ee = ee->prev;
	    ORTESTC3(ee);
	}
    }

    if (!threeconn(edel))
    {
	++degree[v1];
	++degree[v2];
	return 1;
    }

 /* Now we have complete success!  Delete edel! */

    *newmaxlist1 = nml1;
    *ngood_or = ng_or;
    *ngood_inv = ng_inv;
    *ncan_or = nc_or;
    *ncan_inv = nc_inv;

    for (i = oldmaxlist1; i < nml1; ++i) {
      inmaxface[i]->left_facesize = maxface;
      inmaxface[i]->op->invers->left_facesize = maxface;
    }

    edel->prev->next = edel->next;
    edel->op->next->prev = edel->op->prev;
    edel->next->prev = edel->prev;
    edel->op->prev->next = edel->op->next;
    edel = edel->invers;
    edel->prev->next = edel->next;
    edel->op->next->prev = edel->op->prev;
    edel->next->prev = edel->prev;
    edel->op->prev->next = edel->op->next;
    firstedge[v1] = e1;
    firstedge[v2] = e2;
    ne -= 2;

    adj[v1][v2] = adj[v2][v1] = 0;

    if (ng_or + ng_inv == 1) return 0;

    maxcol = 0;
    for (i = 0; i < nc_or; ++i)
        if (ORCOLC3(good_or[i]) > maxcol) maxcol = ORCOLC3(good_or[i]);
    for (i = 0; i < nc_inv; ++i)
        if (INVCOLC3(good_inv[i]) > maxcol) maxcol = INVCOLC3(good_inv[i]);

    for (i = 0, k = 0; i < nc_or; ++i)
        if (ORCOLC3(good_or[i]) == maxcol) good_or[k++] = good_or[i];
    *ncan_or = k;
    for (; i < ng_or; ++i)
    {
	col = ORCOLC3(good_or[i]);
        if (col == maxcol)
	    good_or[k++] = good_or[i];
        else if (col > maxcol)
        {
            insert_edge_tri(edel->invers);
            *ngood_or = *ngood_inv = 0;
	    return 2;
        }
    }
    *ngood_or = k;

    for (i = 0, k = 0; i < nc_inv; ++i)
        if (INVCOLC3(good_inv[i]) == maxcol) good_inv[k++] = good_inv[i];
    *ncan_inv = k;
    for (; i < ng_inv; ++i)
    {
	col = INVCOLC3(good_inv[i]);
        if (col == maxcol)
            good_inv[k++] = good_inv[i];
        else if (col > maxcol) 
        {
            insert_edge_tri(edel->invers);
            *ngood_or = *ngood_inv = 0;
	    return 2;
        }
    }
    *ngood_inv = k;

    return 0; 
}
 
/*************************************************************************/

static void
scanpoly_c3(int nbtot, int nbop, EDGE *oldfeas[], int noldfeas,
            int oldmaxface, int oldmaxlist0, int oldmaxlist1)

/* This is the recursive search procedure for polytopes.
   oldfeas[0..noldfeas-1] are the edges which can be removed
   without violating connectivity.  nbtot/nbop represent the
   group, as usual.   oldmaxface is the size of the largest face.
   inmaxface[oldmaxlist0..oldmaxlist1-1] lists the edges whose
   left face has greatest size (unless that is 3).

   This version is for 3-connected only!  scanpoly() is the more
   general version that works for connectivity 1 or more.
*/

{
    EDGE *newfeas[MAXE/2],*good_or[MAXE],*good_inv[MAXE],*e,*esave;
    int i,nnewfeas,minimal[MAXE/2],newmaxface;
    int code,xnbtot,xnbop;
    int ngood_or,ncan_or,ngood_inv,ncan_inv;
    int newmaxlist0,newmaxlist1;

    if (ne <= edgebound[1]) got_one(nbtot,nbop,3);
    if (ne == edgebound[0]) return;
    if (ne - 2*noldfeas > edgebound[1]) return;

#ifdef PRE_FILTER_POLY
    if (!(PRE_FILTER_POLY)) return;
#endif

    mark_edge_orbits(oldfeas,noldfeas,minimal,nbtot);

    for (i = 0; i < noldfeas; ++i)
    {
        if (!minimal[i]) continue;

        e = oldfeas[i];
        if (e->left_facesize != 3) e = e->invers;
	if (e->invers->left_facesize < oldmaxface-1
                      || e->invers->left_facesize == maxfacesize)
            continue;

	check_it(109);
        code = maybe_delete_c3(e,oldmaxface,oldmaxlist0,oldmaxlist1,
		               &newmaxface,&newmaxlist0,&newmaxlist1,
		               good_or,&ngood_or,&ncan_or,
		               good_inv,&ngood_inv,&ncan_inv);
	check_it(110);
        if (code == 0)
        {
#ifdef FAST_FILTER_POLY
	    if (FAST_FILTER_POLY)
#endif
	    {
	        if (ngood_or + ngood_inv == 1)
	        {
                    esave = oldfeas[i];
                    oldfeas[i] = oldfeas[noldfeas-1];
                    prune_poly_edgelist(oldfeas,noldfeas-1,newfeas,&nnewfeas);
                    oldfeas[i] = esave;
                    scanpoly_c3(1,1,newfeas,nnewfeas,
                             newmaxface,newmaxlist0,newmaxlist1);
                }
                else if (canon_edge_oriented(good_or,ngood_or,ncan_or,
                    		             good_inv,ngood_inv,ncan_inv,
                    		             degree,numbering,&xnbtot,&xnbop))
                {
		    esave = oldfeas[i];
                    oldfeas[i] = oldfeas[noldfeas-1];
                    prune_poly_edgelist(oldfeas,noldfeas-1,newfeas,&nnewfeas);
                    oldfeas[i] = esave;
                    scanpoly_c3(xnbtot,xnbop,newfeas,nnewfeas,
			     newmaxface,newmaxlist0,newmaxlist1);
                }
	    }
            insert_edge_tri(e);
        }
        else if (code == 1)
	{
	    oldfeas[i] = oldfeas[noldfeas-1];
	    minimal[i] = minimal[noldfeas-1];
	    --i;
	    --noldfeas;
	}
    }
}
 
/*************************************************************************/

static int
maybe_delete(EDGE *edel, int oldmaxface, int oldmaxlist0, int oldmaxlist1,
             int *newmaxface, int *newmaxlist0, int *newmaxlist1,
             EDGE *good_or[], int *ngood_or, int *ncan_or,
             EDGE *good_inv[], int *ngood_inv, int *ncan_inv, int *connec)

/* Assumes there is a 3-face on the left of *edel, and that *edel can be
   deleted while keeping degrees >= minpolydeg.  Also, the new face created
   will be a largest face.  oldmaxface is the size of the largest face so far,
   and inmaxface[oldmaxlist0..oldmaxlist1-1] are the edges with a face of
   oldmaxface on the left.  *connec is the current connectivity (or 3,
   whichever is smaller).

   This procedure deletes *edel provided the result is at least as connected
   as minpolyconn, and that the procedure believes the re-insertion of *edel
   might be a canonical edge insertion.  The latter decision is based on
   four combinatorial invariates: the three vertex degrees at the ends of
   *edel and on its left, and the size of the face to the left of edel->prev.

   In case *edel passes the test and is deleted, the edges which may 
   represent the re-insertion of *edel and optimise the four-part invariant
   mentioned above are put in good_or[0..*ncan_or-1] and
   good_inv[0..*ncan_inv-1].  This will include at least one of the edges
   edel->prev, edel->invers->next and (if there is also a 3-face on the
   right of *edel) edel->next and edel->invers->prev.  Then all other edges
   which might possibly represent canonical edge insertions are put in 
   good_or[*ncan_or..*ngood_or-1] or good_inv[*ncan_inv..*ngood_inv-1].
   The *_or edges are those for which the inserted edge will be in the
   next direction, and the *_inv edges .. the prev direction.

   In addition, if *edel is deleted, inmaxface[*newmaxlist0..*newmaxlist1-1]
   will have all edges with a maximum face on the left (that size being put
   into *newmaxface).  This list may overlap
   inmaxface[oldmaxlist0..oldmaxlist1-1] if the max face size does not
   increase.

   In the case of value 0, *connec is changed to represent the new
   connectivity (or 3, whichever is smaller).

   Return values:   0 = ok
                    1 = rejected as connectivity will be too small
		    2 = rejected by colour
*/

{
#define DEGEND(e) (degree[e->end])
#define OROK(e) ISNEQADJ(e->start,e->invers->prev->end)
#define INVOK(e) ISNEQADJ(e->start,e->invers->next->end)
#define REJECT(x) {++degree[v1]; ++degree[v2]; return x;}
#define ORTEST(e) {col = DEGEND(e); if (col > maxcol) \
  {if (OROK(e)) REJECT(2)} else if (col==maxcol&&OROK(e)) good_or[ng_or++]=e;}
#define INVTEST(e) {col = DEGEND(e); if (col > maxcol) \
  {if (INVOK(e)) REJECT(2)} \
   else if (col==maxcol&&INVOK(e)) good_inv[ng_inv++]=e;}
#define ORCOL(e) (((long)degree[e->invers->prev->end] << 10)  \
                      + e->left_facesize)
#define INVCOL(e) \
 (((long)degree[e->invers->next->end] << 10) + e->invers->left_facesize)

  EDGE *e1,*e2,*e3,*e4,*ee,*ea,*eb,*e,*ex;
    long col,maxcol,col1,col2;
    int k,i,ng_or,ng_inv,nc_or,nc_inv;
    int maxface,v1,v2,v3,v4,v5;
    int da,db,mindeg,maxdeg,nml1,nml2,i0,i1;
    int newconnec;

    maxface = *newmaxface = edel->invers->left_facesize + 1;

    /* check if deleting edel would transform button holes into non button 
       holes.
       suppose edel=(a,b) is a non-border edge of a button hole, {a,b,c} is a 
       3-cycle, and (a,c) and (b,c) are border edges of button hole. if edel 
       is deleted then it could be reinserted inside the former button hole.
       let maxtransface be the maximum size of these transformed faces.
       if maxtransface < maxface then any transformed button holes may be 
       ignored.  
       if maxtransface > maxface then reject edel.  
       if maxtransface == maxface then any edges of transformed faces with 
       maxface edges must be added to inmaxface for checking later in this 
       routine.  when maxface==oldmaxface there is no easy way to tell if edel
       is the first edge to transform a button hole.  so we add all edges of 
       button holes of size maxface when maxface is first increased 
       (maxface>oldmaxface).
    */

    if (irrhole) {
      e1 = edel->next;
      do 
	{
	  if (e1->invers->left_facesize > 3) {
	    if (e1->invers->left_facesize > maxface)
	      if (e1->invers->prev->end == edel->end ||
		  e1->next->invers->next->end == edel->end)
		return 2;
	  }
	  e1 = e1->next;
	} while (e1 != edel);
    }
    
    if (maxface > oldmaxface) oldmaxlist0 = oldmaxlist1;
    *newmaxlist0 = oldmaxlist0;

    *ngood_or = *ngood_inv = 0;

 /* Now old edges that will be on a largest face (if *edel is deleted)
    are in places oldmaxlist0..oldmaxlist1-1.  New edges will be from
    oldmaxlist1 to *newmaxlist1-1, but *newmaxlist1 is not set yet. */

    v1 = edel->start;  v2 = edel->end;
 /* maxdeg = (degree[v1] > degree[v2] ? degree[v1] : degree[v2]) - 1;  BDM 2/5/09 */

 /* mindeg and maxdeg are the degrees of the ends of edel after deletion */
    if (degree[v1] < degree[v2])
    {
	mindeg = degree[v1]-1; maxdeg = degree[v2]-1;
    }
    else
    {
	mindeg = degree[v2]-1; maxdeg = degree[v1]-1;
    }

    e1 = edel->prev; v3 = e1->end;
    e4 = edel->next; v4 = e4->end;
    e3 = edel->invers->prev;

   /* The following is an efficiency short-cut, but it is assumed
      to have been done below. */

    if (maxface == 4)
    {
	if (ISNEQADJ(v3,v4) && (degree[v3] > maxdeg || degree[v4] > maxdeg))
	    return 2;
    }
    else if (degree[v3] > maxdeg)
    {
	if (ISNEQADJ(v3,e3->end) || ISNEQADJ(v3,v4)) return 2;
    }

    e2 = edel->invers->next;
    --degree[v1];
    --degree[v2];

    maxcol = nc_or = nc_inv = 0;

    if (degree[v1] == maxdeg)
    {
        col = DEGEND(e1);
	maxcol = col;
        good_or[nc_or++] = e1;

        if (maxface == 4)
	{
	    col = DEGEND(e4);
	    if (col > maxcol)
	    {
		nc_or = nc_inv = 0;
		good_inv[nc_inv++] = e4;
		maxcol = col;
	    }
	    else if (col == maxcol)
	        good_inv[nc_inv++] = e4;
	}
    }

    if (degree[v2] == maxdeg)
    {
	col = DEGEND(e2);
	if (col > maxcol)
  	{
	    nc_or = nc_inv = 0;
            good_inv[nc_inv++] = e2;
	    maxcol = col;
	}
	else if (col == maxcol)
	    good_inv[nc_inv++] = e2;
	
	if (maxface == 4)
	{
            col = DEGEND(e3);
	    if (col > maxcol)
	    {
		nc_or = nc_inv = 0;
		good_or[nc_or++] = e3;
		maxcol = col;
	    }
	    else if (col == maxcol)
		good_or[nc_or++] = e3;
	}
    }

    ng_or = nc_or;  ng_inv = nc_inv;

    if (maxface == 4)
    {
     /* Recall from above that degree[v3,v4] <= maxdeg */

	if (ISNEQADJ(v3,v4))
	{
	    col1 = degree[v1];
	    col2 = degree[v2];

	    if (degree[v3] == maxdeg)
	    {
		if (col1 > maxcol) REJECT(2)
	        else if (col1 == maxcol) good_inv[ng_inv++] = e1->invers;
		if (col2 > maxcol) REJECT(2)
		else if (col2 == maxcol) good_or[ng_or++] = e2->invers;
	    }

	    if (degree[v4] == maxdeg)
	    {
		if (col1 > maxcol) REJECT(2)
	        else if (col1 == maxcol) good_or[ng_or++] = e4->invers;
		if (col2 > maxcol) REJECT(2)
		else if (col2 == maxcol) good_inv[ng_inv++] = e3->invers;
	    }
	}

        nml1 = oldmaxlist1;
        inmaxface[nml1++] = e1->invers;
        inmaxface[nml1++] = e2;
        inmaxface[nml1++] = e3->invers;
        inmaxface[nml1++] = e4;
    }
    else
    {
	v5 = e3->end;

	/* Recall from above that degree[v3] <= maxdeg */

	if (ISNEQADJ(v3,v5))
	{
	    if (degree[v5] > maxdeg) REJECT(2);
	    col = degree[v2];
	    if (degree[v3] == maxdeg)
	    {
		if (col > maxcol) REJECT(2)
                else if (col == maxcol) good_or[ng_or++] = e2->invers;
	    }
	    if (degree[v5] == maxdeg)
	    {
		if (col > maxcol) REJECT(2)
                else if (col == maxcol) good_inv[ng_inv++] = e3->invers;
            }
	}
	if (ISNEQADJ(v3,v4))
	{
	    if (degree[v4] > maxdeg) REJECT(2);
	    col = degree[v1];
	    if (degree[v3] == maxdeg)
            {
                if (col > maxcol) REJECT(2)
                else if (col == maxcol) good_inv[ng_inv++] = e1->invers;
            }
            if (degree[v4] == maxdeg)
            {
                if (col > maxcol) REJECT(2)
                else if (col == maxcol) good_or[ng_or++] = e4->invers;
            }
	}

        nml1 = oldmaxlist1;

        ea = e3->invers;
	eb = ea->prev;
	do
	{
	    if (ISNEQADJ(ea->end,eb->end))
	    {
		da = DEGEND(ea);
		db = DEGEND(eb);
		if (da > maxdeg || db > maxdeg) REJECT(2);
		col = degree[ea->start];
		if (da == maxdeg)
		{
		    if (col > maxcol) REJECT(2)
		    else if (col == maxcol) good_or[ng_or++] = ea->invers;
		}
		if (db == maxdeg)
		{
		    if (col > maxcol) REJECT(2)
		    else if (col == maxcol) good_inv[ng_inv++] = eb->invers;
		}
	    }
	    inmaxface[nml1++] = ea;
	    ea = eb->invers;
	    eb = ea->prev;
	} while (ea != e4);
         
        inmaxface[nml1++] = e2;
        inmaxface[nml1++] = e1->invers;
        inmaxface[nml1++] = e4;
    }

    /* add edges of button holes of size maxface if not already on list */

    nml2 = nml1;
    if (maxface > oldmaxface && irrhole == 1) 
      for (i = 0; i < nv; ++i)
	{
	  e = ex = firstedge[i];
	  do
	    {
	      if (e->left_facesize == maxface)
		inmaxface[nml2++] = e;
	      e = e->next;
	    } while (e != ex);
	}

    /* Now test old edges still on max faces */
    /* Or test newly added button hole edges */
    
    if (maxface > oldmaxface && irrhole == 1) {
      i0 = nml1;
      i1 = nml2;
    }
    else {
      i0 = oldmaxlist0;
      i1 = oldmaxlist1;
    }

    for (i = i0; i < i1; ++i)
    {
	ee = inmaxface[i];
	if (degree[ee->start] > maxdeg)
	{
	    if (INVOK(ee)) REJECT(2);
            ee = ee->prev;
            if (OROK(ee)) REJECT(2);
	}
	else if (degree[ee->start] == maxdeg)
	{
	    INVTEST(ee);
	    ee = ee->prev;
	    ORTEST(ee);
	}
    }

    if (mindeg < *connec)
	newconnec = mindeg;
    else if (*connec == 3)
	newconnec = 2 + threeconn(edel);
    else if (*connec == 2)
	newconnec = 1 + twoconn(edel);
    else
	newconnec = 1;

    if (newconnec < minpolyconnec) REJECT(1);

 /* Now we have complete success!  Just prune the lists a bit and
    complete the deletion of edel. */

    edel->prev->next = edel->next;
    edel->op->next->prev = edel->op->prev;
    edel->next->prev = edel->prev;
    edel->op->prev->next = edel->op->next;
    ee = edel->invers;
    ee->prev->next = ee->next;
    ee->op->next->prev = ee->op->prev;
    ee->next->prev = ee->prev;
    ee->op->prev->next = ee->op->next;

    for (i = oldmaxlist1; i < nml1; ++i) {
      inmaxface[i]->left_facesize = maxface;
      inmaxface[i]->op->invers->left_facesize = maxface;
    }

    firstedge[v1] = e1;
    firstedge[v2] = e2;
    ne -= 2;

    adj[v1][v2] = adj[v2][v1] = 0;

    *connec = newconnec;

    *newmaxlist1 = nml2;
    *ngood_or = ng_or;
    *ngood_inv = ng_inv;
    *ncan_or = nc_or;
    *ncan_inv = nc_inv;

    if (ng_or + ng_inv == 1) return 0;

    maxcol = 0;
    for (i = 0; i < nc_or; ++i)
        if (ORCOL(good_or[i]) > maxcol) maxcol = ORCOL(good_or[i]);
    for (i = 0; i < nc_inv; ++i)
        if (INVCOL(good_inv[i]) > maxcol) maxcol = INVCOL(good_inv[i]);

    for (i = 0, k = 0; i < nc_or; ++i)
        if (ORCOL(good_or[i]) == maxcol) good_or[k++] = good_or[i];
    *ncan_or = k;
    for (; i < ng_or; ++i)
    {
	col = ORCOL(good_or[i]);
        if (col == maxcol)
	    good_or[k++] = good_or[i];
        else if (col > maxcol)
        {
            insert_edge_tri(edel);
            *ngood_or = *ngood_inv = 0;
	    return 2;
        }
    }
    *ngood_or = ng_or = k;

    for (i = 0, k = 0; i < nc_inv; ++i)
        if (INVCOL(good_inv[i]) == maxcol) good_inv[k++] = good_inv[i];
    *ncan_inv = k;
    for (; i < ng_inv; ++i)
    {
	col = INVCOL(good_inv[i]);
        if (col == maxcol)
            good_inv[k++] = good_inv[i];
        else if (col > maxcol) 
        {
            insert_edge_tri(edel);
            *ngood_or = *ngood_inv = 0;
	    return 2;
        }
    }
    *ngood_inv = ng_inv = k;

    return 0; 
}

/*************************************************************************/

static void
scanpoly(int nbtot, int nbop, EDGE *oldfeas[], int noldfeas,
         int oldmaxface, int oldmaxlist0, int oldmaxlist1, int connec)

/* This is the recursive search procedure for polytopes.
   oldfeas[0..noldfeas-1] are the edges which can be removed without
   violating the degree bound minpolydeg, with some (but not necessarily
   all) missing because they are known to violate the connectivity
   bound minpolyconnec.  nbtot/nbop represent the group, as usual.
   oldmaxface is the size of the largest face.
   inmaxface[oldmaxlist0..oldmaxlist1-1] lists the edges whose
   left face has greatest size (unless that is 3).  connec is
   the actual connectivity, except that values greater than 3
   are given as 3. */
{
    EDGE *newfeas[MAXE/2],*good_or[MAXE],*good_inv[MAXE],*e,*esave;
    int i,nnewfeas,minimal[MAXE/2],newmaxface;
    int code,xnbtot,xnbop;
    int ngood_or,ncan_or,ngood_inv,ncan_inv;
    int newmaxlist0,newmaxlist1,newconnec;

    if (ne <= edgebound[1]) got_one(nbtot,nbop,connec);
    if (ne == edgebound[0]) return;
    if (ne - 2*noldfeas > edgebound[1]) return;

#ifdef PRE_FILTER_POLY
    if (!(PRE_FILTER_POLY)) return;
#endif

    mark_edge_orbits(oldfeas,noldfeas,minimal,nbtot);

    for (i = 0; i < noldfeas; ++i)
    {
        if (!minimal[i]) continue;

        e = oldfeas[i];
        if (e->left_facesize != 3) e = e->invers;
	if (e->invers->left_facesize < oldmaxface-1
                      || e->invers->left_facesize == maxfacesize)
            continue;

	AMDELEDGE(e->start,e->end);

	newconnec = connec;
        code = maybe_delete(e,oldmaxface,oldmaxlist0,oldmaxlist1,
		            &newmaxface,&newmaxlist0,&newmaxlist1,
		            good_or,&ngood_or,&ncan_or,
		            good_inv,&ngood_inv,&ncan_inv,&newconnec);
	check_it(111);
        if (code == 0)
        {
#ifdef FAST_FILTER_POLY
            if (FAST_FILTER_POLY)
#endif  
            {
	        if (ngood_or + ngood_inv == 1)
	        {
                    esave = oldfeas[i];
                    oldfeas[i] = oldfeas[noldfeas-1];
                    prune_poly_edgelist(oldfeas,noldfeas-1,newfeas,&nnewfeas);
                    oldfeas[i] = esave;
                    scanpoly(1,1,newfeas,nnewfeas,
                             newmaxface,newmaxlist0,newmaxlist1,newconnec);
                }
                else if (canon_edge_oriented(good_or,ngood_or,ncan_or,
                    		             good_inv,ngood_inv,ncan_inv,
                    		             degree,numbering,&xnbtot,&xnbop))
                {
                 /* The following line corrects for the fact that canon*()
                    finds each automorphism twice if the maximum degree is
                    at most 2.  However, it interferes with -o so is disabled.
    
	            if (ne <= 2*nv && maxdegree() <= 2) xnbtot = xnbop;
                 */
		    esave = oldfeas[i];
                    oldfeas[i] = oldfeas[noldfeas-1];
                    prune_poly_edgelist(oldfeas,noldfeas-1,newfeas,&nnewfeas);
                    oldfeas[i] = esave;
                    scanpoly(xnbtot,xnbop,newfeas,nnewfeas,
			     newmaxface,newmaxlist0,newmaxlist1,newconnec);
                }
	    }
            insert_edge_tri(e);
        }
        else if (code == 1)
	{
	    oldfeas[i] = oldfeas[noldfeas-1];
	    minimal[i] = minimal[noldfeas-1];
	    --i;
	    --noldfeas;
	}
	AMADDEDGE(e->start,e->end);
    }
}

/**************************************************************************/

static void
startpolyscan(int nbtot, int nbop)

/* This routine begins the scan for general maps formed
   by removing edges from face irreducible maps.  The current graph
   is such a face irreducible map, and the group is known 
   (parameters nbtot,nbop).
*/

{
  EDGE *feasible[MAXE/2];
  EDGE *e,*ex,*e1;
  EDGE **nb0,**nb1;
  int xnbtot,xnbop;
  int i,j,nfeas;
  int ok,facesize;
  int newmaxface;

  /* check minimum degree (minpolydeg) */
  
  for (i = 0; i < nv; ++i)
    if (degree[i] < minpolydeg)
      return;
  
  /* In the -m5 case, the group must be saved and restored. */
  
  if (minimumdeg == 5 && nbtot > 1)
    {
      nb0 = (EDGE**)numbering[0];
      nb1 = (EDGE**)saved_numbering[0];
      for (i = 0; i < nbtot; ++i, nb0 += MAXE, nb1 += MAXE)
	for (j = 0; j < ne; ++j) nb1[j] = nb0[j];
    }
  
  if (minpolyconnec < 3)
    {
      for (i = 0; i < nv; ++i)
        {
	  for (j = 0; j < nv; ++j) am2[i][j] = 0;
	  am2[i][i] = 1;
        }
    }
  
  /* compute left_facesize while making feasability list */
  
  ok = TRUE;
  nfeas = 0;
  newmaxface = 3;
  for (i = 0; i < nv && ok; ++i)
    {
      e = ex = firstedge[i];
      do
	{
	  if (e->start > e->end && degree[e->start] > minpolydeg
	      && degree[e->end] > minpolydeg)
	    feasible[nfeas++] = e;
	  facesize = 0;
	  e1 = e;
	  do
	    {
	      facesize++;
	      e1 = e1->invers->next;
	    } while (e1 != e);
	  if (facesize > 3) {
	    if (facesize > maxfacesize) {
	      ok = FALSE;
	      break;
	    }
	    newmaxface = MAX(newmaxface,facesize);
	  }
	  
	  e->left_facesize = facesize;
	  e->op->next->left_facesize = facesize;
	  am2[e->start][e->end] = 1;
	  e = e->next;
	} while (e != ex);
    }

  /* test to see if connectivity has increased as vertices are split */

  if (ok && (polyconnec == 1)) {

    polyconnec = 2;
    for (i = 0; i < nv && polyconnec == 2; ++i) {
      e = ex = firstedge[i];
      do
	{
	  e1 = e;
	  do
	    {
	      e1 = e1->invers->next;
	      if (e1 != e) {
		if (e1->start == e->start)
		  polyconnec = 1;
	      }
	    } while (polyconnec == 2 && e1 != e);
	  e = e->next;
	} while (polyconnec == 2 && e != ex);
    }
  }

  if (ok) {
    check_it(113);
    canon(degree,numbering,&xnbtot,&xnbop);
    
    prune_poly_edgelist(feasible,nfeas,feasible,&nfeas);
    
    neroot = ne;
    if (minpolyconnec == 3) scanpoly_c3(nbtot,nbop,feasible,nfeas,3,0,0);
    else                    scanpoly(nbtot,nbop,feasible,nfeas,3,0,0,polyconnec);
  }

  if (minimumdeg == 5 && nbtot > 1)
    {
      nb0 = (EDGE**)saved_numbering[0];
      nb1 = (EDGE**)numbering[0];
      for (i = 0; i < nbtot; ++i, nb0 += MAXE, nb1 += MAXE)
	for (j = 0; j < ne; ++j) nb1[j] = nb0[j];
    }
  check_it(316);
}

/*************************************************************************/

static void
scansimple(int nbtot, int nbop)
  
/* The main node of the recursion for triangulations without double
   edges or loops.  As this procedure is entered,
   nv,ne,degree etc are set for some graph, and nbtot/nbop are the
   values returned by canon() for that graph. */
{
  EDGE *ext[MAXN*MAXN];
  int degext[MAXN*MAXN];
  int numext;
  EDGE *save_list[2*(MAXN-3)];
  register int i;
  register EDGE *e1,*e2,**nb,**nblim;
  int nc,xnbtot,xnbop;
  int colour[MAXN];
#ifndef SPLITTEST
  EDGE *firstedge_save[MAXN];
#endif

  if (nv == splitlevel && nv != nvroot)
    {
#ifdef SPLITTEST
      ADDBIG(splitcases,1);
      return;
#else
      if (splitcount-- != 0) return;
      splitcount = mod - 1;
      
      for (i = 0; i < nv; ++i) firstedge_save[i] = firstedge[i];
#endif
    }
  
#ifdef PRE_FILTER_SIMPLE
  if (!(PRE_FILTER_SIMPLE)) {
    if (nv == splitlevel && nv != nvroot)
      for (i = 0; i < nv; ++i) firstedge[i] = firstedge_save[i];
    return;
  }
#endif

  if (nv == maxnv)
    {
      if (pswitch)               startpolyscan(nbtot,nbop);
      else                       got_one(nbtot,nbop,3);
      return;
    }
  
#ifndef FIND_EXTENSIONS_SIMPLE
#define FIND_EXTENSIONS_SIMPLE find_extensions
#endif
  
#ifndef FIND_EXTENSIONS_HOLES
#define FIND_EXTENSIONS_HOLES find_extensions_holes
#endif
  
  if (!irrhole)
    FIND_EXTENSIONS_SIMPLE(nbtot,nbop,ext,degext,&numext);
  else
    FIND_EXTENSIONS_HOLES(nbtot,nbop,ext,degext,&numext);
  
  for (i = 0; i < numext; ++i) {
    if (degext[i] == 3)
      {
	nc = make_colours(colour,ext[i]);
	if (nc > 0)
	  {
	    if (!irrhole)
	      extend3(ext[i]);
	    else
	      extend3_holes(ext[i]);
	    
	    if (nc == 1 && nv == maxnv && !needgroup)
	      got_one(1,1,3);
	    else if (canon(colour,numbering,&xnbtot,&xnbop))
	      scansimple(xnbtot,xnbop);
	    
	    if (!irrhole)
	      reduce3(ext[i]);
	    else
	      reduce3_holes(ext[i]);
	  }
      }
    else if (degext[i] == 4)
      {
	if (!irrhole)
	  extend4(ext[i],save_list);
	else
	  extend4_holes(ext[i],save_list);
	
	if (canon(degree,numbering,&xnbtot,&xnbop))
	  {
	    e1 = ext[i]->next->invers;
	    e2 = e1->next->next;
	    nblim = (EDGE**)numbering[xnbtot];
	    for (nb = (EDGE**)numbering[0]; nb < nblim; nb += MAXE)
	      if (*nb == e1 || *nb == e2) break;
	    
	    if (nb < nblim) scansimple(xnbtot,xnbop);
	  }
	if (!irrhole)
	  reduce4(ext[i],save_list);
	else
	  reduce4_holes(ext[i],save_list);
      }
    else
      {
	if (!irrhole)
	  extendn(degext[i],ext[i],save_list);
	else
	  extendn_holes(degext[i],ext[i],save_list);
	
	if (canon(degree,numbering,&xnbtot,&xnbop))
	  {
	    e1 = ext[i]->next->invers;
	    e2 = numbering[0][0];
	    if (xnbtot == 1)
	      {
		if (e2 == e1) scansimple(xnbtot,xnbop);
	      }
	    else
	      {
		nblim = (EDGE**)numbering[xnbtot];
		for (nb = (EDGE**)numbering[0]; nb < nblim; nb += MAXE)
		  if (*nb == e1) break;
		
		if (nb < nblim) scansimple(xnbtot,xnbop);
	      }
	  }
	if (!irrhole)
	  reducen(degext[i],ext[i],save_list);
	else
	  reducen_holes(degext[i],ext[i],save_list);
      }
  }
  
#ifndef SPLITTEST
  if (nv == splitlevel && nv != nvroot)
    for (i = 0; i < nv; ++i) firstedge[i] = firstedge_save[i];
#endif
}


/**************************************************************************/

static void
prune_oriented_lists(EDGE *good_or[], int *ngood_or, int *ngood_ref,
                     EDGE *good_mir[], int *ngood_mir, int *ngood_mir_ref)

/* Try to prune the edge lists (of the form required by
   canon_edge_oriented()) by using the degrees of a couple of
   extra vertices.  The result is returned in the same form.
   As always, if *ngood_ref==*ngood_mir_ref==0 on output
   (which must not be true on input), all else is undefined.

   Used only for quadrangulations.
*/
{
    int i,k,kref;
    long code_or[MAXE],code_mir[MAXE],maxcode;
#define PRUNE_OR(e) ((degree[(e)->invers->prev->prev->end] << 10) \
		      + degree[(e)->next->invers->prev->end])
#define PRUNE_MIR(e) ((degree[(e)->invers->next->next->end] << 10) \
                      + degree[(e)->prev->invers->next->end])

    maxcode = 0;
    for (i = 0; i < *ngood_or; ++i)
    {
	code_or[i] = PRUNE_OR(good_or[i]);
	if (code_or[i] > maxcode) maxcode = code_or[i];
    }

    for (i = 0; i < *ngood_mir; ++i)
    {
        code_mir[i] = PRUNE_MIR(good_mir[i]);
        if (code_mir[i] > maxcode) maxcode = code_mir[i];
    }

    k = kref = 0;
    for (i = 0; i < *ngood_or; ++i)
        if (code_or[i] == maxcode)
	{
	    if (i < *ngood_ref) ++kref;
	    good_or[k++] = good_or[i];
	}
    *ngood_or = k;
    *ngood_ref = kref;

    k = kref = 0;
    for (i = 0; i < *ngood_mir; ++i)
        if (code_mir[i] == maxcode)
        {
            if (i < *ngood_mir_ref) ++kref;
            good_mir[k++] = good_mir[i];
        }
    *ngood_mir = k;
    *ngood_mir_ref = kref;
}

/**************************************************************************/

static void
min4_four_legal(EDGE *ref, EDGE *good_or[], int *ngood_or, int *ngood_ref,
                EDGE *good_mir[], int *ngood_mir, int *ngood_mir_ref)

/* The four-operation with reference edge *ref has just been performed.
   Make a list in good_or[0..*ngood_or-1] of the reference edges of
   legal four-reductions (oriented editions) that might be canonical,
   with the first *ngood_ref of those being ref.
   Make a list in good_mir[0..*ngood_mir-1] of the
   reference edges of legal four-reductions (mirror-image editions)
   that might be canonical, with the first *ngood_mir_ref of those being
   ref->next.
   *ngood_ref and *ngood_mir_ref might each be 0-2.  If they are
   both 0, nothing else need be correct.
   All the edges in good_or[] and good_mir[] must start with the same
   vertex degree and end with the same vertex degree (actually, colour
   as passed to canon_edge_oriented).
   Four-reductions have a priority: (maxdeg, mindeg) where those are
   the two degrees on the subdivided diagonal.  Bigger = better.
*/
{
    EDGE *e,*er;
    long deg1,deg2,deg3,deg4;
#define FOURTYPE(d3,d4) (((d3)<<10)+(d4))
    long besttype=0,etype1=0,etype2=0,ertype1=0,ertype2=0,etype=0,ertype=0;
    int bestdeg=0,nextdeg=0,nor,nmir,i,w;
    EDGE *ei,*eilast;

    e = ref;
    er = ref->next->next;
    deg1 = degree[e->end];
    deg2 = degree[er->end];
    deg3 = degree[e->prev->end];
    deg4 = degree[e->next->end];

    nor = nmir = 0;

    if (deg1 >= deg2)
    {
        bestdeg = deg1; nextdeg = deg2;
        etype1 = FOURTYPE(deg3,deg4);
        etype2 = FOURTYPE(deg4,deg3);
        besttype = etype = etype1 > etype2 ? etype1 : etype2;
    }

    if (deg1 <= deg2)
    {
        bestdeg = deg2; nextdeg = deg1;
        ertype1 = FOURTYPE(deg4,deg3);
        ertype2 = FOURTYPE(deg3,deg4);
	besttype = ertype = ertype1 > ertype2 ? ertype1 : ertype2;
    }

    if (deg1 == deg2) besttype = etype > ertype ? etype : ertype;

    if (deg1 >= deg2)
    {
	if (etype1 == besttype) good_or[nor++] = e;
	if (etype2 == besttype) good_mir[nmir++] = e;
    }
    if (deg1 <= deg2)
    {
        if (ertype1 == besttype) good_or[nor++] = er;
        if (ertype2 == besttype) good_mir[nmir++] = er;
    }

    *ngood_ref = nor;
    *ngood_mir_ref = nmir;

    RESETMARKS;
    MARKLO(e->invers);
    MARKLO(er->invers);
    MARKLO(e->op->invers);
    MARKLO(er->op->invers);

    for (i = 0; i < nv; ++i)
    if (degree[i] > bestdeg)
    {
	ei = eilast = firstedge[i];
	do
	{
	    if (degree[ei->end] == 4 && is_min4_four(ei->invers))
	    {
		*ngood_ref = *ngood_mir_ref = 0;
		return;
	    }
	    ei = ei->next;
	} while (ei != eilast);
    }
    else if (degree[i] == bestdeg)
    {
        ei = eilast = firstedge[i];
        do
        {
            if (degree[ei->end] == 4 && !ISMARKEDLO(ei))
	    {
		e = ei->invers;
	  	er = e->next->next;
		w = er->end;
		deg2 = degree[w];
		if (deg2 > nextdeg && is_min4_four(e))
		{
                    *ngood_ref = *ngood_mir_ref = 0;
                    return;
                }
		else if (deg2 == nextdeg && (bestdeg > nextdeg || i < w)
					 && is_min4_four(e))
		{
		    deg3 = degree[e->prev->end];
		    deg4 = degree[e->next->end];
                    etype1 = FOURTYPE(deg3,deg4);
                    etype2 = FOURTYPE(deg4,deg3);

                    if (etype1 > besttype || etype2 > besttype)
                    {
                        *ngood_ref = *ngood_mir_ref = 0;
                        return;
                    }
                    if (etype1 == besttype) good_or[nor++] = e;
                    if (etype2 == besttype) good_mir[nmir++] = e;

		    if (nextdeg == bestdeg)
		    {
                	ertype1 = FOURTYPE(deg4,deg3);
                	ertype2 = FOURTYPE(deg3,deg4);

                	if (ertype1 > besttype || ertype2 > besttype)
                	{
                    	    *ngood_ref = *ngood_mir_ref = 0;
                    	    return;
                	}

                	if (ertype1 == besttype) good_or[nor++] = er;
                	if (ertype2 == besttype) good_mir[nmir++] = er;
		    }
		}
	    }
	    ei = ei->next;
	} while (ei != eilast);
    }

    if (nor > *ngood_ref || nmir > *ngood_mir_ref)
        prune_oriented_lists(good_or,&nor,ngood_ref,
                             good_mir,&nmir,ngood_mir_ref);

    *ngood_or = nor;
    *ngood_mir = nmir;    
}

/**************************************************************************/

static int
min4_four_centre()

/* If there is a four-reduction, return the central vertex.
   If not, return -1.
*/
{
    int i;
    EDGE *e;

    for (i = 0; i < nv; ++i)
    if (degree[i] == 4)
    {
	e = firstedge[i];
	if ((degree[e->prev->end] >= 5 && degree[e->next->end] >= 5 &&
	     !adj[e->end][e->next->next->end]) ||
	    (degree[e->end] >= 5 && degree[e->next->next->end] >= 5 &&
	     !adj[e->prev->end][e->next->end]))
	  return i;
    }

    return -1;
}

/*************************************************************************/

static int
is_min4_four_centre(int v)

/* return 1 if v is the centre of a four-reduction, 0 if not.
*/
{
    EDGE *e;

    if (degree[v] != 4) return 0;

    e = firstedge[v];
    if ((degree[e->prev->end] >= 5 && degree[e->next->end] >= 5 &&
	 !adj[e->end][e->next->next->end]) ||
	(degree[e->end] >= 5 && degree[e->next->next->end] >= 5 &&
	 !adj[e->prev->end][e->next->end]))
        return 1;
    else
	return 0;
}

/*************************************************************************/

static int
degree5_vertex()

/* Return a vertex of degree 5 which has a 4-contractable edge, 
   -1 if there are none */
{
    int i,j;
    EDGE *e;

    for (i = 0; i < nv; ++i)
      if (degree[i] == 5 && ncontract[i] > 0) {
	e = firstedge[i];
	for (j = 0; j < 5; j++) {
	  if (nnei[i][e->end] == 2)
	    if (degree[e->next->end] > 4 &&
		degree[e->prev->end] > 4)
	      return i;
	  e = e->next;
	}
      }
    return -1;
}

/*************************************************************************/

static int
degree5plus_vertex(int n)
     
/* Return a vertex of degree 4..(n-1) which has a 4-contractable edge, or
   return a vertex of degree 4 which part of a triangle of three vertices of 
   degree 4, or 
   return -1 if there are none */
{
  int i,j;
  EDGE *e;
  
  for (i = 0; i < nv; ++i) {
    if (degree[i] < n && ncontract[i] > 0) {
      e = firstedge[i];
      for (j = 0; j < degree[i]; j++) {
	if (nnei[i][e->end] == 2)
	  if (degree[e->next->end] > 4 &&
	      degree[e->prev->end] > 4)
	    return i;
	e = e->next;
      }
    }
    if (degree[i] == 4) {
      e = firstedge[i];
      for (j = 0; j < 4; j++) {
	if (degree[e->end] == 4 &&
	    degree[e->next->end] == 4)
	  return i;
	e = e->next;
      }
    }
  }
  return -1;
}
  
/*************************************************************************/

static void
extend_five_plus(EDGE *ref,int n)

/* This routine is another implementation of the five-operation
   in Batagelj's paper. */
{
    EDGE *a,*b,*start,*dummy,*e;
    EDGE *oa,*ob,*ostart,*odummy,*oref;
    int v[MAXN],vend,i,j,k;

    oref=ref->op;
    start=five_op(nv);
    degree[nv]=n; firstedge[nv]=start;


    b=ref->prev;
    a=ref->next;
    dummy=start+1;
    oa=a->op;
    ob=b->op;
    ostart=start->op;
    odummy=dummy->op;
    
    b->next=a->prev=dummy; 
    ob->prev=oa->next=odummy;
    dummy->next=a; dummy->prev=b;
    odummy->prev=oa; odummy->next=ob;
    v[n-2]=a->start;
    dummy->start=start->end=v[n-2];
    odummy->start=ostart->end=v[n-2];
    firstedge[v[n-2]]=a;

    a=b->invers; b=a->prev;
    start +=2; /*2*/ dummy=start+1;
    oa=a->op;
    ob=b->op;
    ostart=start->op;
    odummy=dummy->op;
    b->next=a->prev=dummy; 
    ob->prev=oa->next=odummy;
    dummy->next=a; dummy->prev=b;
    odummy->prev=oa; odummy->next=ob;
    v[n-1]=a->start;
    dummy->start=start->end=v[n-1];
    odummy->start=ostart->end=v[n-1];
    (degree[v[n-1]])++;

    a=b->invers; 
    b=a->prev->prev->prev; for (i=0;i<n-5;i++) b=b->prev;
    dummy=ref->invers;
    oa=a->op;
    ob=b->op;
    odummy=dummy->op;
    b->next=dummy;
    ob->prev=odummy;
    dummy->prev=b;
    odummy->next=ob;
    v[0]=a->start;
    firstedge[v[0]]=a;
    degree[v[0]]-=n-4;

    a=b->invers; b=a->prev;
    start +=2; /*4*/ dummy=start+1;
    oa=a->op;
    ob=b->op;
    ostart=start->op;
    odummy=dummy->op;
    b->next=a->prev=dummy;
    ob->prev=oa->next=odummy;
    dummy->next=a; dummy->prev=b;
    odummy->prev=oa; odummy->next=ob;
    v[1]=a->start;
    dummy->start=start->end=v[1];
    odummy->start=ostart->end=v[1];
    (degree[v[1]])++;

    ref->start=ref->invers->end=nv;
    oref->start=oref->invers->end=nv;
    ref->next=start; ref->prev=start-2;
    oref->prev=ostart; oref->next=ostart-2;
    start->prev=ref; (start-2)->next=ref;
    ostart->next=oref; (ostart-2)->prev=oref;

    dummy=b->invers->prev->invers; /* the other remaining edge of the "v" */
    odummy=dummy->op;

    dummy->start=dummy->invers->end=nv;
    odummy->start=odummy->invers->end=nv;
    v[2]=dummy->end;
    dummy->prev=start;
    odummy->next=ostart;
    start->next=dummy;
    ostart->prev=odummy;

    for (i=3;i<n-2;i++) {
      dummy=dummy->next;
      odummy=dummy->op;
      dummy->start=dummy->invers->end=nv;
      odummy->start=odummy->invers->end=nv;
      v[i]=dummy->end;
    }

    dummy->next=start-4;
    odummy->prev=ostart-4;
    (start-4)->prev=dummy;
    (ostart-4)->next=odummy;

  /* do bookkeeping for removing edges (diagonals) */

  /* remove each diagonal edge.
     if nnei of its ends was 2 then it was contractible so reduce
     ncontract for its ends. */

  for (j=2;j<n-1;j++)
    {
      if (nnei[v[0]][v[j]] == 2)
	{
	  ncontract[v[0]]--;
	  ncontract[v[j]]--;
	}
      adj[v[0]][v[j]] = adj[v[j]][v[0]] = FALSE;
    }

  /* for each neighbor, vend, of v[0] reduce nnei of vend and
     each other end (v[j],j=2,..,n-2) of the diagonals.
     if vend and v[j] are adjacent and nnei is now 2 then increase
     ncontract for vend and v[j].
     don't bother with vend=(v[i],i=1,..,n-1) since nnei[v[i]][v[j]]
     would then need to be increased later
     when radial edges are added. (iou A) */

  e = ref->invers->next;
  for (k=0;k<degree[v[0]]-3;k++)
    {
      e = e->next;
      vend = e->end;
      for (j=2;j<n-1;j++)
	{
	  nnei[v[j]][vend] = --nnei[vend][v[j]];
	  if (adj[v[j]][vend] && nnei[v[j]][vend]==2)
	    {
	      ncontract[vend]++;
	      ncontract[v[j]]++;
	    }
	}
    }

  /* for each neighbor, vend, of each non-v[0] end of the diagonals
     (v[j],j=2,..,n-2) reduce nnei of vend and v[0].
     if vend and v[0] are adjacent and nnei is now 2 then increase
     ncontract for vend and v[0].
     don't bother with edges on path around v[nv] will deal with them later
     (iou B) */

  start=b->invers;
  for (j=2;j<n-1;j++)
    {
      e = start;
      for (k=0;k<degree[v[j]]-3;k++)
	{
	  e = e->next;
	  vend = e->end;
	  nnei[v[0]][vend] = --nnei[vend][v[0]];
	  if (adj[v[0]][vend] && nnei[v[0]][vend]==2)
	    {
	      ncontract[vend]++;
	      ncontract[v[0]]++;
	    }
	}
      start = start->prev->prev->invers;
    }

  /* do bookkeeping for adding edges (radial from new vertex) */

  /* for each old vertex, i, compute nnei[nv][i] by counting number of new
     neighbors of nv (v[j],j=0,..,n-1) which are adjacent to i. */

  for (i=0;i<nv;i++)
    {
      nnei[nv][i] = 0;
      for (j=0;j<n;j++)
	nnei[nv][i] += adj[v[j]][i];
      nnei[i][nv] = nnei[nv][i];
    }

  /* for each new edge if nnei is 2 increase ncontract of ends */

  ncontract[nv] = 0;
  for (j=0;j<n;j++)
    {
      adj[v[j]][nv] = adj[nv][v[j]] = TRUE;
      if (nnei[v[j]][nv] == 2)
	{
	  ncontract[v[j]]++;
	  ncontract[nv]++;
	}
    }

  /* update nnei[v[i]][v[j]] because of addition of nv.
     would need to increase for each pair (((i,j),i=0,..,n-2),j=i+1,..,n-1).
     from iou A above we didn't decrease (((i,j),i=1,n-2),j=i+1,..,n-1)
     except (1,n-1).
     from iou B above we didn't decrease (((0,j),i=2,..,n-2),j=i-1,i+1).
     so still need to decrease ((0,j),j=3,..,n-3) and  increase (1,n-1).
  */

  /* decrease ((0,j),j=3,..,n-3). */

  for (j=3;j<n-2;j++)
    nnei[v[0]][v[j]] = --nnei[v[j]][v[0]];

  /* increase (1,n-1). */

  nnei[v[1]][v[n-1]] = ++nnei[v[n-1]][v[1]];
  if (adj[v[1]][v[n-1]] && nnei[v[1]][v[n-1]]==3)
    {
      ncontract[v[1]]--;
      ncontract[v[n-1]]--;
    }


    nv++;
    ne+=6;

    check_it(800+n);
}

/**************************************************************************/

static void
reduce_five_plus(EDGE *ref,int n)

/* The reverse of extend_five_plus(). */
{
    EDGE *a,*b,*dummy;
    EDGE *oref,*oa,*ob,*odummy,*e,*start;
    int v[MAXN],i,j,k,vend;

    a=ref->next->next; dummy=ref->invers; b=dummy->prev;
    oa=a->op;
    ob=b->op;
    odummy=dummy->op;
    oref=ref->op;
    b->next=a; a->prev=b;
    ob->prev=oa; oa->next=ob;
    v[0]=dummy->start;
    a->start=a->invers->end=v[0];
    oa->start=oa->invers->end=v[0];
    (degree[v[0]])+=n-4;

    a=b->invers; b=a->prev->prev;
    oa=a->op;
    ob=b->op;
    a->prev=b; b->next=a; 
    oa->next=ob; ob->prev=oa;
    v[1]=a->start;
    (degree[v[1]])--;
    firstedge[v[1]]=a;

    v[2]=b->end;

    a=ref->next->next;

    for (i=3;i<n-2;i++) {
      a=a->next;
      oa=a->op;
      a->start=a->invers->end=v[0];
      oa->start=oa->invers->end=v[0];
      v[i]=a->end;
    }

    oa=a->op;
    dummy->prev=a; a->next=dummy;
    odummy->next=oa; oa->prev=odummy;

    a=a->invers->prev->invers;
    b=a->prev->prev;
    oa=a->op;
    ob=b->op;
    a->prev=ref; ref->next=a; b->next=ref; ref->prev=b;
    oa->next=oref; oref->prev=oa; ob->prev=oref; oref->next=ob;
    v[n-2]=a->start;
    ref->start=ref->invers->end=v[n-2];
    oref->start=oref->invers->end=v[n-2];
    firstedge[v[n-2]]=a;

    a=b->invers; b=a->prev->prev;
    oa=a->op;
    ob=b->op;
    a->prev=b; b->next=a; 
    oa->next=ob; ob->prev=oa;
    v[n-1]=a->start;
    (degree[v[n-1]])--;
    firstedge[v[n-1]]=a;

  /* do bookkeeping for removing edges (radial from new vertex) */

  for (j=0;j<n;j++)
    {
      adj[v[j]][nv-1] = adj[nv-1][v[j]] = FALSE;
      if (nnei[v[j]][nv-1] == 2)
	ncontract[v[j]]--;
    }

  for (j=3;j<n-2;j++)
    nnei[v[0]][v[j]] = ++nnei[v[j]][v[0]];

  nnei[v[1]][v[n-1]] = --nnei[v[n-1]][v[1]];

  if (adj[v[1]][v[n-1]] && nnei[v[1]][v[n-1]] == 2)
    {
      ncontract[v[1]]++;
      ncontract[v[n-1]]++;
    }

  /* do bookkeeping for adding edges (diagonals) */

  e = ref->invers->next;
  for (k=0;k<degree[v[0]]-(n-1);k++)
    {
      e = e->next;
      vend = e->end;
      for (j=2;j<n-1;j++)
	{
	  nnei[v[j]][vend] = ++nnei[vend][v[j]];
	  if (adj[v[j]][vend] && nnei[v[j]][vend]==3)
	    {
	      ncontract[vend]--;
	      ncontract[v[j]]--;
	    }
	}
    }

  start = ref->next;
  for (j=n-2;j>1;j--)
    {
      e = start;
      for (k=0;k<degree[v[j]]-3;k++)
	{
	  e = e->next;
	  vend = e->end;
	  nnei[v[0]][vend] = ++nnei[vend][v[0]];
	  if (adj[v[0]][vend] && nnei[v[0]][vend]==3)
	    {
	      ncontract[vend]--;
	      ncontract[v[0]]--;
	    }
	}
      start = start->invers->next->next;
    }

  for (j=2;j<n-1;j++)
    {
      if (nnei[v[0]][v[j]] == 2)
	{
	  ncontract[v[0]]++;
	  ncontract[v[j]]++;
	}
      adj[v[0]][v[j]] = adj[v[j]][v[0]] = TRUE;
    }

    nv--;
    ne-=6;

    check_it(890+n);
}

/**************************************************************************/

static void
min4_five_plus_legal(int n, EDGE *ref, EDGE *good_or[], int *ngood_or, 
		     int *ngood_ref, EDGE *good_mir[], int *ngood_mir, 
		     int *ngood_mir_ref)

/* The five_plus-operation with reference edge *ref has just been performed.
   Make a list in good_or[0..*ngood_or-1] of the reference edges of
   legal five-reductions (oriented editions) that might be canonical,
   with the first *ngood_ref of those being ref.
   Make a list in good_mir[0..*ngood_mir-1] of the
   reference edges of legal five-reductions (mirror-image editions)
   that might be canonical, with the first *ngood_mir_ref of those being
   ref.
   *ngood_ref and *ngood_mir_ref might each be 0 or 1.  If they are
   both 0, nothing else need be correct.
   All the edges in good_or[] and good_mir[] must start with the same
   vertex degree and end with the same vertex degree (actually, colour
   as passed to canon_edge_oriented).
*/
{
    EDGE *e;
    long deg1,deg2,deg3,bestdeg;
#define FIVETYPE(d1,d2,d3) (((d1)<<21)+((d2)<<11)+(d3))
    long besttype,type1,type2;
    int nor,nmir,i,j;

    e = ref;
    bestdeg = degree[e->end];
    deg2 = degree[e->prev->end];
    deg3 = degree[e->next->end];

    type1 = FIVETYPE(bestdeg,deg2,deg3);
    type2 = FIVETYPE(bestdeg,deg3,deg2);
    besttype = type1 > type2 ? type1 : type2;

    nor = nmir = 0;
    if (type1 == besttype) good_or[nor++] = e;
    if (type2 == besttype) good_mir[nmir++] = e;

    *ngood_ref = nor;
    *ngood_mir_ref = nmir;

    for (i = 0; i < nv; ++i)
    if (degree[i] == n)
    {
	for (e = firstedge[i], j = 0; j < n; ++j, e = e->next)
        {
	    if (e == ref) continue;
            deg1 = degree[e->end];
	    if (deg1 < bestdeg) continue;
	    deg2 = degree[e->prev->end];
	    if (deg2 == 4) continue;
	    deg3 = degree[e->next->end];
	    if (deg3 == 4) continue;

	    if (nnei[i][e->end] > 2) continue;

	    if (deg1 > bestdeg)
	    {
		*ngood_ref = *ngood_mir_ref = 0;
		return;
	    }

	    type1 = FIVETYPE(deg1,deg2,deg3);
    	    type2 = FIVETYPE(deg1,deg3,deg2);

	    if (type1 > besttype || type2 > besttype)
	    {
                *ngood_ref = *ngood_mir_ref = 0;
                return;
            }
    	    if (type1 == besttype) good_or[nor++] = e;
    	    if (type2 == besttype) good_mir[nmir++] = e;
        }
    }

    *ngood_or = nor;
    *ngood_mir = nmir;    
}

/**************************************************************************/

static void
extend_S(EDGE *ref)

/* This routine is the implementation of the S-operation in Batagelj's paper.
   It inserts 3 new vertices in triangular form into a triangle on the right
   hand side of ref */

{
    register EDGE *a,*b,*start;
    EDGE *oa,*ob,*ostart,*oref;
    int v[3];  /* v[i] is oposite (not adjacent to) nv+i */
    int i,j,k;

    start=S_op(nv);

    a=ref->invers->prev;
    b=a->invers->prev;
    
    oref=ref->op;
    oa=a->op;
    ob=b->op;
    ostart=start->op;

    degree[nv]=degree[nv+1]=degree[nv+2]=4;
    
    v[2]=ref->start;
    degree[v[2]]+=2; 
    (start+10)->start=(start+11)->end=(start+8)->start=(start+9)->end=v[2];
    (ostart+10)->start=(ostart+11)->end=(ostart+8)->start=(ostart+9)->end=v[2];

    v[1]=ref->end;
    degree[v[1]]+=2; 
    (start+12)->start=(start+13)->end=(start+14)->start=(start+15)->end=v[1];
    (ostart+12)->start=(ostart+13)->end=(ostart+14)->start=(ostart+15)->end=v[1];

    v[0]=a->end;
    degree[v[0]]+=2; 
    (start+6)->start=(start+7)->end=(start+16)->start=(start+17)->end=v[0];
    (ostart+6)->start=(ostart+7)->end=(ostart+16)->start=(ostart+17)->end=v[0];

    firstedge[nv]=start+1;
    firstedge[nv+1]=start;
    firstedge[nv+2]=start+2;

    ref->next=start+10; (start+10)->prev=ref;
    oref->prev=ostart+10; (ostart+10)->next=oref;
    a->next=start+14; (start+14)->prev=a;
    oa->prev=ostart+14; (ostart+14)->next=oa;
    b->next=start+6; (start+6)->prev=b;
    ob->prev=ostart+6; (ostart+6)->next=ob;

    ref=ref->invers;
    oref=oref->invers;
    ref->prev=start+12; (start+12)->next=ref;
    oref->next=ostart+12; (ostart+12)->prev=oref;
    a=a->invers;
    oa=oa->invers;
    a->prev=start+16; (start+16)->next=a;
    oa->next=ostart+16; (ostart+16)->prev=oa;
    b=b->invers;
    ob=ob->invers;
    b->prev=start+8; (start+8)->next=b;
    ob->next=ostart+8; (ostart+8)->prev=ob;
    
    for (i=0;i<3;i++) 
      for (k=0;k<nv;k++)
	nnei[nv+i][k] = nnei[k][nv+i] = 0;
    
    for (i=0;i<3;i++) {
      for (j=0;j<3;j++) {
	if (i != j) {
	  adj[v[i]][nv+j]=adj[nv+j][v[i]]=TRUE;
	  nnei[v[i]][nv+j]=nnei[nv+j][v[i]]+=1;
	  adj[nv+i][nv+j]=TRUE;
	  nnei[nv+i][nv+j]=2;
	  nnei[v[i]][v[j]]++;
	  if (nnei[v[i]][v[j]] == 3) {
	    ncontract[v[i]]--;
	  }
	  for (k=0;k<nv;k++)
	    nnei[k][nv+i] = nnei[nv+i][k] += adj[v[j]][k];
	}
      }
      nnei[v[i]][nv+i]=nnei[nv+i][v[i]]+=2;
      ncontract[nv+i]=4;
      ncontract[v[i]]+=2;
    }
    
    nv+=3;
    ne+=18;
    
    check_it(800);
}

/**************************************************************************/

static void
reduce_S(EDGE *ref)

/* This routine is the implementation of the inverse S-operation in
   Batagelj's paper. It removes the triangle. 
*/
{
    register EDGE *a,*b;
    EDGE *oa,*ob,*oref;
    int v[3];  /* v[i] is oposite (not adjacent to) nv+i */
    int i,j;

    a=ref->invers->prev->prev->prev;
    b=a->invers->prev->prev->prev;

    v[0] = b->start;
    v[1] = a->start;
    v[2] = ref->start;

    oref=ref->op;
    oa=a->op;
    ob=b->op;

    firstedge[ref->start]=ref;
    degree[ref->start]-=2;
    firstedge[a->start]=a;
    degree[a->start]-=2;
    firstedge[b->start]=b;
    degree[b->start]-=2;

    ref->next=b->invers; b->invers->prev=ref;
    oref->prev=ob->invers; ob->invers->next=oref;
    a->next=ref->invers; ref->invers->prev=a;
    oa->prev=oref->invers; oref->invers->next=oa;
    b->next=a->invers; a->invers->prev=b;
    ob->prev=oa->invers; oa->invers->next=ob;

    nv-=3;
    ne-=18;

    for (i=0;i<3;i++) {
      for (j=0;j<3;j++) {
	if (i != j) {
	  adj[v[i]][nv+j]=adj[nv+j][v[i]]=FALSE;
	  adj[nv+i][nv+j]=FALSE;
	  nnei[v[i]][v[j]]--;
	  if (nnei[v[i]][v[j]] == 2) ncontract[v[i]]++;
	}
      }
      ncontract[v[i]]-=2;
    }

    check_it(890);
}

/**************************************************************************/

static void
min4_S_legal(EDGE *ref, EDGE *good_or[], int *ngood_or, int *ngood_ref,
             EDGE *good_mir[], int *ngood_mir, int *ngood_mir_ref)

/* The S-operation with reference edge *ref has just been performed.
   Make a list in good_or[0..*ngood_or-1] of the reference edges of
   legal S-reductions (oriented editions) that might be canonical,
   with the first *ngood_ref of those being ref.
   Make a list in good_mir[0..*ngood_mir-1] of the
   reference edges of legal S-reductions (mirror-image editions)
   that might be canonical, with the first *ngood_mir_ref of those
   being ref.
   *ngood_ref and *ngood_mir_ref might each be 0 or 1.  If they are
   both 0, nothing else need be correct.
   All the edges in good_or[] and good_mir[] must start with the same
   vertex degree and end with the same vertex degree (actually, colour
   as passed to canon_edge_oriented).

   Hopefully this routine is used rarely, so we are going for simple
   rather than fast.  If the S operation ever becomes a critical part
   of some generation, optimising this routine would be necesary.
*/
{
    EDGE *e,*ee;
    int nor,nmir,i,j;

    RESETMARKS;
    e = ref;

    nor = nmir = 0;
    good_or[nor++] = e;
    good_mir[nmir++] = e->invers;
    MARKLO(e); 

    ee = e->invers->prev->prev->prev;
    good_or[nor++] = ee;
    good_mir[nmir++] = ee->invers;
    MARKLO(ee);

    ee = ee->invers->prev->prev->prev;
    good_or[nor++] = ee;
    good_mir[nmir++] = ee->invers;
    MARKLO(ee);

    *ngood_ref = nor;
    *ngood_mir_ref = nmir;

    for (i = 0; i < nv; ++i)
    if (degree[i] >= 6)
    {
	for (e = firstedge[i], j = degree[i]; --j >= 0; e = e->next)
        {
	    if (ISMARKEDLO(e)) continue;
            if (degree[e->end] < 6 || degree[e->next->end] != 4 ||
		degree[e->next->next->end] != 4 || 
		degree[e->next->next->next->end] < 6) continue;
	    if (degree[e->next->invers->next->next->end] != 4) continue;

	    good_or[nor++] = e;
            good_mir[nmir++] = e->invers;
            MARKLO(e);
	    MARKLO(e->next->next->next->op);
        
            ee = e->invers->prev->prev->prev;
            good_or[nor++] = ee;
            good_mir[nmir++] = ee->invers;
            MARKLO(ee);
	    MARKLO(ee->next->next->next->op);

            ee = ee->invers->prev->prev->prev;
            good_or[nor++] = ee;
            good_mir[nmir++] = ee->invers;
            MARKLO(ee);
	    MARKLO(ee->next->next->next->op);
        }
    }

    *ngood_or = nor;
    *ngood_mir = nmir;    
}

/**************************************************************************/

static void
find_extensions_min4_four(int nbtot, EDGE *ext4[], int *next4, EDGE *known)

/* nbtot is the number of automorphisms.  If known!=NULL, it is the
   reference edge of a canonical four-extension just used to make the
   graph.  See min4_four_legal() for important comments about what makes
   a four-extension canonical.  This procedure makes a list in
   ext4[0..*next4-1] of undirected edges whose use in a four-extension
   might possible be canonical.
*/
{
    int deg1,deg2=0,deg3,dega,degb;
    int i,v,k,kk;
    EDGE *e,*elast,*ez,*ezlim,*e2=NULL;
    EDGE **nb,**nb0,**nblim;

    RESETMARKS;
    k = 0;

    if (known) 
    {
	deg1 = degree[known->end];
	deg2 = degree[known->next->next->end];
	if (deg1 < deg2) 
        {
	    i = deg1; deg1 = deg2; deg2 = i;
	    known = known->next->next;
	}

	ez = known->invers;
	ezlim = ez->prev;
	deg3 = 0;
	for (ez = ez->next->next; ez != ezlim; ez = ez->next)
	{
   	    if (is_min4_four(ez->invers))
	    {
		kk = degree[ez->invers->next->next->end];
		if (kk > deg3)
		{
		    deg3 = kk;
		    e2 = ez;
		}
	    }
	}

	if (deg3 == 0)
	{
          /* edges around the outside */
	    e = known->invers->next;
	    ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	    e = e->invers->next->next; 
	    ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	    e = e->invers->next->next; 
	    ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	    e = e->invers->next->next; 
	    ext4[k++] = e; MARKLO(e); MARKLO(e->invers);

          /* edges in the middle */
	    e = known;
	    ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	    e = e->next;
	    ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
            e = e->next;
            if (deg2 == deg1)
            {
                ext4[k++] = e; MARKLO(e); MARKLO(e->invers);    
            }
            e = e->next;
            ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	}
	else
        {
	    dega = degree[known->prev->end];
	    degb = degree[known->next->end];

          /* edges around the outside */
            e = known->invers->next;
	    if (dega >= deg3 || e->next == e2)
	    {
                ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	    }

            e = e->invers->next->next;
	    if (deg1 == deg2 && dega >= deg3)
	    {
                ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	    }

            e = e->invers->next->next;
	    if (deg1 == deg2 && degb >= deg3)
	    {
                ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	    }

            e = e->invers->next->next;
	    if (degb >= deg3 || e->invers->prev == e2)
	    {
                ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	    }

          /* edges in the middle */
            e = known;
            if (deg3 == 4)
            {
                ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
            }

            e = e->next;
	    if (degb > deg1+1 || (degb == deg1+1 && deg3 == 4))
	    {
                ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	    }

            e = e->next;
            if (deg1 == deg2 && deg3 == 4)
            {
                ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
            }

            e = e->next;
	    if (dega > deg1+1 || (dega == deg1+1 && deg3 == 4))
	    {
                ext4[k++] = e; MARKLO(e); MARKLO(e->invers);
	    }
        }
    }
    else
	deg1 = 0;

    for (v = 0; v < nv; ++v)
    if (degree[v] > deg1)
    {
	e = elast = firstedge[v];
	do
	{
	    if (!ISMARKEDLO(e) && !ISMARKEDLO(e->op))
            {
                ext4[k++] = e;
                MARKLO(e); MARKLO(e->invers);
            }
	    e = e->next;
	} while (e != elast);
    }
    else if (degree[v] == deg1)
    {
	e = elast = firstedge[v];
        do
        {
            if (!ISMARKEDLO(e) && !ISMARKEDLO(e->op) && degree[e->end] >= deg2)
	    {
		ext4[k++] = e;
		MARKLO(e); MARKLO(e->invers);
	    }
            e = e->next;
        } while (e != elast);
    }

    if (nbtot > 1)
    {
	nb0 = (EDGE**)numbering[0];
	nblim = (EDGE**)numbering[nbtot];
	for (i = 0; i < ne; ++i) {
	  nb0[i]->index = i;
	  nb0[i]->op->index = i;
	}

	kk = 0;
	RESETMARKS;
	for (i = 0; i < k; ++i)
	if (!ISMARKEDLO(ext4[i]) && !ISMARKEDLO(ext4[i]->op))
	{
	    ext4[kk++] = ext4[i];
	    for (nb = nb0 + ext4[i]->index + MAXE; nb < nblim; nb += MAXE)
	    {
		MARKLO(*nb);
		MARKLO((*nb)->invers);
	    }
	}
	k = kk;
    }

    *next4 = k;
}

/**************************************************************************/

static void
find_extensions_min4(int nbtot, int nbop, EDGE *ext4[], int *next4,
		     EDGE *ext5plus[], int degext5plus[], int *next5plus, 
		     EDGE *extS[], int *nextS, EDGE *known)

/* Determine the inequivalent places to make extensions, for the
   ordinary triangulations of minimum degree 4.  The results are
   put in the arrays ext4[0..*next4-1], etc..
   nbtot and nbop are the usual group parameters.

   If known!=NULL, it is the reference edge of a known four-reduction.
*/
{
  register int i,j,k;
  register EDGE *e,*e1,*elast;
  EDGE **nb,**nb0,**nblim,**nboplim;
  int manycentres,ideg;
  int bmindeg;        /* minimum degree (before extension) */
  int nbmindeg;       /* number of vertices with degree of bmindeg */
  int amindeg;        /* largest possible minimum degree (after extension) */
  int n4contract;
  
  bmindeg = nv-1;
  nbmindeg = 0;
  for (i = 0; i < nv; ++i)
    if (ncontract[i]>0 && degree[i] <= bmindeg) {
      n4contract = ncontract[i];
      e = firstedge[i];
      for (j=0; j<degree[i] && n4contract>0; j++, e=e->next)
	if (nnei[i][e->end] == 2) {
	  if (degree[e->prev->end] > 4 && degree[e->next->end] > 4)
	    j = degree[i];
	  else
	    n4contract--;
	}
      if (n4contract > 0) {
	if (degree[i] == bmindeg)
	  ++nbmindeg;
	else if (degree[i] < bmindeg)
	  {
	    bmindeg = degree[i];
	    nbmindeg = 1;
	  }
      }
    }
  if (nbmindeg < 3) amindeg = bmindeg + 1;
  else amindeg = bmindeg;
  
  find_extensions_min4_four(nbtot,ext4,next4,known);
  
  manycentres = amindeg == 4;
  
  if (nbtot == 1)
    {
      /* five_plus-extensions, trivial group */
      
      if (manycentres)
	*next5plus = 0;
      else if (known)
	{
	  k = 0;
	  for (i = 0, e = known; i < 4; ++i, e = e->next)
	    if (degree[e->end] >= 6)
	      {
		ext5plus[k] = e->invers->prev->invers;
		degext5plus[k++] = 5;
		ext5plus[k] = e->invers->next->next->invers;
		degext5plus[k++] = 5;
	      }
	  *next5plus = k;
	}
      else
	{
	  k = 0;
	  for (ideg=5;ideg<=amindeg;ideg++)
	    for (i = 0; i < nv; ++i)
	      if (degree[i] >= 2*ideg-4)
		{
		  e = elast = firstedge[i];
		  do
		    {
		      ext5plus[k] = e->invers;
		      degext5plus[k++] = ideg;
		      e = e->next;
		    } while (e != elast);
		}
	  *next5plus = k;
	}
      
      /* S-extensions, trivial group */
      
      k = 0;
      if (nv <= maxnv-3)
	{
	  for (i = 0; i < nv; ++i)
            {
	      e = elast = firstedge[i];
	      do
                {
		  if (e->start > e->end)
                    {
		      if (e->start > e->next->end) extS[k++] = e;
                    }
		  e = e->next;
                } while (e != elast);
            }
	}
      *nextS = k;
    }
  else    
    {
      /* five-extensions, non-trivial group */
      
      if (manycentres)
	*next5plus = 0;
      else
	{
	  k = 0;
	  nboplim = (EDGE**)numbering[nbop==0?nbtot:nbop];
	  nblim = (EDGE**)numbering[nbtot];
	  
	  for (ideg=5;ideg<=amindeg;ideg++) {
            RESETMARKS;
            nb0 = (EDGE**)numbering[0];
            for (i = 0; i < ne; ++i, ++nb0)
	      {
                e = *nb0;
                if (!ISMARKEDLO(e)) {
		  if (degree[e->start] >= 2*ideg-4)
		    {
		      ext5plus[k] = e->next->invers;
		      degext5plus[k++] = ideg;
		    }
		  nb = nb0 + MAXE;
		  for (; nb < nboplim; nb += MAXE) {
		    MARKLO(*nb);
		    e1 = (*nb)->op->prev;
		    for (j=0;j<ideg-5;j++)
		      e1 = e1->next;
		    MARKLO(e1);
		  }
		  
		  for (; nb < nblim; nb += MAXE) {
		    e1 = (*nb)->prev;
		    for (j=0;j<ideg-5;j++)
		      e1 = e1->next;
		    MARKLO(e1);
		    MARKLO((*nb)->op);
		  }
		}
	      }
	  }
	  *next5plus = k;
	}
      
      /* S-extensions, non-trivial group */
      
      k = 0;
      if (nv <= maxnv-3)
        {
	  nb0 = (EDGE**)numbering[0];
	  nboplim = (EDGE**)numbering[nbop==0?nbtot:nbop];
	  nblim = (EDGE**)numbering[nbtot];
	  RESETMARKS;
	  
	  for (i = 0; i < ne; ++i) {
	    nb0[i]->index = i;
	    nb0[i]->op->index = -1;
	  }
	  
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e)) continue;
	      extS[k++] = e;
	      
	      for (nb = nb0 + i; nb < nboplim; nb += MAXE) {
		MARKLO(*nb);
		MARKLO((*nb)->op->invers);
	      }

	      for (; nb < nblim; nb += MAXE) {
		MARKLO((*nb)->invers);
		MARKLO((*nb)->op);
	      }

	      e1 = e->invers->prev;
	      if (e1->index >= 0) {
		for (nb = nb0 + e1->index; nb < nboplim; nb += MAXE) {
		  MARKLO(*nb);
		  MARKLO((*nb)->op->invers);
		}
	      
		for (; nb < nblim; nb += MAXE) {
		  MARKLO((*nb)->invers);
		  MARKLO((*nb)->op);
		}
	      }
	      else {
		for (nb = nb0 + e1->op->index; nb < nboplim; nb += MAXE) {
		  MARKLO((*nb)->invers);
		  MARKLO((*nb)->op);
		}
		for (; nb < nblim; nb += MAXE) {
		  MARKLO(*nb);
		  MARKLO((*nb)->op->invers);
		}
	      }
	      
	      e1 = e1->invers->prev;
	      if (e1->index >= 0) {
		for (nb = nb0 + e1->index; nb < nboplim; nb += MAXE) {
		  MARKLO(*nb);
		  MARKLO((*nb)->op->invers);
		}
	      
		for (; nb < nblim; nb += MAXE) {
		  MARKLO((*nb)->invers);
		  MARKLO((*nb)->op);
		}
	      }
	      else {
		for (nb = nb0 + e1->op->index; nb < nboplim; nb += MAXE) {
		  MARKLO((*nb)->invers);
		  MARKLO((*nb)->op);
		}
		for (; nb < nblim; nb += MAXE) {
		  MARKLO(*nb);
		  MARKLO((*nb)->op->invers);
		}
	      }
            }
	}
      *nextS = k;
    }
}

/**************************************************************************/

static void
scanmin4(int nbtot, int nbop, int dosplit, EDGE *lastfour)

/* The main node of the recursion for triangulations with minimum
   degree at least 4.
   As this procedure is entered, nv,ne,degree etc are set for some graph,
   and nbtot/nbop are the values returned by canon() for that graph.
   If dosplit==TRUE, this is the place to do splitting (if any).
   Splitting is a bit more complicated because the S operation adds
   three vertices.
   If this graph was made with a four-operation, the reference edge of
   that operation is lastfour.  If not, lastfour=NULL.
*/
{
#ifndef SPLITTEST
    EDGE *firstedge_save[MAXN];
#endif
    EDGE *ext4[MAXE/2],*ext5plus[MAXN*MAXN],*extS[MAXF];
    EDGE *good_or[MAXE],*good_mir[MAXE];
    int ngood_or,ngood_mir,ngood_ref,ngood_mir_ref;
    int next4,next5plus,degext5plus[MAXN*MAXN],nextS,i;
    int xnbtot,xnbop;
    int hint;

    /* make compiler happy */
    ngood_mir = ngood_or = 0;

    if (nv == maxnv)
      {

	/* for large minimumdeg, check now */

	if (minimumdeg > 4)
	  for (i = 0; i < nv; ++i)
	    if (degree[i] < minimumdeg)
	      return;
	
	if (pswitch) startpolyscan(nbtot,nbop);
	else         got_one(nbtot,nbop,3);
	return;
      }

    if (dosplit && nv != nvroot)
    {
#ifdef SPLITTEST
        ADDBIG(splitcases,1);
        return;
#else
        if (splitcount-- != 0) return;
        splitcount = mod - 1;

        for (i = 0; i < nv; ++i) firstedge_save[i] = firstedge[i];
#endif
    }

#ifdef PRE_FILTER_MIN4
    if (!(PRE_FILTER_MIN4)) return;
#endif

#ifndef FIND_EXTENSIONS_MIN4
#define FIND_EXTENSIONS_MIN4 find_extensions_min4
#endif

    FIND_EXTENSIONS_MIN4(nbtot,nbop,ext4,&next4,ext5plus,degext5plus,
			 &next5plus,extS,&nextS,lastfour);
    
    for (i = 0; i < next4; ++i)
      {
        extend_four(ext4[i]);
	min4_four_legal(ext4[i],good_or,&ngood_or,&ngood_ref,
			good_mir,&ngood_mir,&ngood_mir_ref);
	
        if (ngood_ref+ngood_mir_ref > 0)
	  {
	    if (nv == maxnv && !needgroup && ngood_or == ngood_ref
		&& ngood_mir == ngood_mir_ref && minimumdeg == 4)
	      got_one(1,1,3);
	    else if (ngood_or+ngood_mir==1)
	      scanmin4(1,1,nv==splitlevel,ext4[i]);
	    else if (canon_edge_oriented(good_or,ngood_or,ngood_ref,
					 good_mir,ngood_mir,ngood_mir_ref,
					 degree,numbering,&xnbtot,&xnbop))
	      scanmin4(xnbtot,xnbop,nv==splitlevel,ext4[i]);
	  }
	reduce_four(ext4[i]);
      }

    /* hint = lastfour ? lastfour->start : -1; */
    hint = -1;
    for (i = 0; i < next5plus; ++i)
      if (degext5plus[i] == 5)
	{
	  extend_five_plus(ext5plus[i],degext5plus[i]);
	  if (hint < 0 || !is_min4_four_centre(hint))
	    hint = min4_four_centre();
	  if (hint < 0)
	    {
	      min4_five_plus_legal(degext5plus[i],ext5plus[i],good_or,
				   &ngood_or,&ngood_ref,good_mir,&ngood_mir,
				   &ngood_mir_ref);
	      if (ngood_ref+ngood_mir_ref > 0)
		{
		  if (canon_edge_oriented(good_or,ngood_or,ngood_ref,
					  good_mir,ngood_mir,ngood_mir_ref,
					  degree,numbering,&xnbtot,&xnbop))
		    scanmin4(xnbtot,xnbop,nv==splitlevel,NULL);
		}
	    }
	  reduce_five_plus(ext5plus[i],degext5plus[i]);
	}
    
    for (i = 0; i < next5plus; ++i)
      if (degext5plus[i] > 5)
	{
	  extend_five_plus(ext5plus[i],degext5plus[i]);
	  
	  if (degree5plus_vertex(degext5plus[i]) < 0)
	    {
	      min4_five_plus_legal(degext5plus[i],ext5plus[i],good_or,
				   &ngood_or,&ngood_ref,good_mir,&ngood_mir,
				   &ngood_mir_ref);
	      if (ngood_ref+ngood_mir_ref > 0)
		{
		  if (canon_edge_oriented(good_or,ngood_or,ngood_ref,
					  good_mir,ngood_mir,ngood_mir_ref,
					  degree,numbering,&xnbtot,&xnbop))
		    scanmin4(xnbtot,xnbop,nv==splitlevel,NULL);
		}
	    }
	  reduce_five_plus(ext5plus[i],degext5plus[i]);
	}
    
    for (i = 0; i < nextS; ++i)
      {
        extend_S(extS[i]);
        if (degree5_vertex() < 0 && min4_four_centre() < 0)
	  {
            min4_S_legal(extS[i],good_or,&ngood_or,&ngood_ref,
                         good_mir,&ngood_mir,&ngood_mir_ref);
            if (ngood_ref+ngood_mir_ref > 0)
	      {
                if (canon_edge_oriented(good_or,ngood_or,ngood_ref,
                                        good_mir,ngood_mir,ngood_mir_ref,
                                        zero,numbering,&xnbtot,&xnbop))
		  scanmin4(xnbtot,xnbop,nv>=splitlevel&&nv<=splitlevel+2,NULL);
	      }
	  }
        reduce_S(extS[i]);
      }

#ifndef SPLITTEST
    if (dosplit && nv != nvroot)
      for (i = 0; i < nv; ++i) firstedge[i] = firstedge_save[i];
#endif
}

/****************************************************************************/

static void
initialize_min3_quadrangulations()

/* initialize edges for the 3-connected  and min3 quadrangulation
   generation, and make the inital cube */

{
    EDGE *run,*start;
    EDGE *orun,*ostart;
    int n,i;

/* First initialize the edges for operation P1 and P4. They look like

       a
    \ / \ 
    ?b   c--  Vertex c is the new point. (a,c) and (d,c) are the
    / \ / ?   new edges with a-->c taken as quadr_P1(n) 
       d

It is assumed that for 5<=n<MAXN after quadr_P1(n) there is room for 4 edges.
*/
    
    for (n=5; n<MAXN; n++)
      { run=start=quadr_P1(n);
	orun=ostart=start+NUMEDGES;
        run->end=n;
        orun->end=n;
        run->invers=start+1; 
        orun->invers=ostart+1; 
	
        run=start+1; run->start=n; 
        orun=ostart+1; orun->start=n; 
        run->invers=start; run->prev=start+3;
        orun->invers=ostart; orun->next=ostart+3;
	
        run=start+2; run->end=n; run->invers=start+3;
        orun=ostart+2; orun->end=n; orun->invers=ostart+3;

        run=start+3; run->start=n; run->invers=start+2;
        orun=ostart+3; orun->start=n; orun->invers=ostart+2;
        run->next=start+1;
        orun->prev=ostart+1;
      }

    /* The edges for operation P3.

    It is assumed that after quadr_P3(n) there is room for 16 edges.
    */

    for (n=4; n<MAXN-4; n++)
      { start=quadr_P3(n);
	ostart=start+NUMEDGES;

      for (i=0; i<4; i++)
	{ run=start+3*i;
	  orun=ostart+3*i;
	  run->next=run+1; run->prev=run+2; run->start=n+i;
	  orun->prev=orun+1; orun->next=orun+2; orun->start=n+i;
	  run++; run->next=run+1; run->prev=run-1; run->start=n+i;
	  orun++; orun->prev=orun+1; orun->next=orun-1; orun->start=n+i;
	  run++; run->next=run-2; run->prev=run-1; run->start=n+i;
	  orun++; orun->prev=orun-2; orun->next=orun-1; orun->start=n+i;
      }

      run=start; run->invers=start+12;
      run=start+1; run->invers=start+3; run->end=n+1;
      run=start+2; run->invers=start+11; run->end=n+3;
      run=start+3; run->invers=start+1; run->end=n;
      run=start+4; run->invers=start+13;
      run=start+5; run->invers=start+6; run->end=n+2;
      run=start+6; run->invers=start+5; run->end=n+1;
      run=start+7; run->invers=start+14;
      run=start+8; run->invers=start+9; run->end=n+3;
      run=start+9; run->invers=start+8; run->end=n+2;
      run=start+10; run->invers=start+15;
      run=start+11; run->invers=start+2; run->end=n;
      run=start+12; run->invers=start; run->end=n;
      run=start+13; run->invers=start+4; run->end=n+1;
      run=start+14; run->invers=start+7; run->end=n+2;
      run=start+15; run->invers=start+10; run->end=n+3;

      orun=ostart; orun->invers=ostart+12;
      orun=ostart+1; orun->invers=ostart+3; orun->end=n+1;
      orun=ostart+2; orun->invers=ostart+11; orun->end=n+3;
      orun=ostart+3; orun->invers=ostart+1; orun->end=n;
      orun=ostart+4; orun->invers=ostart+13;
      orun=ostart+5; orun->invers=ostart+6; orun->end=n+2;
      orun=ostart+6; orun->invers=ostart+5; orun->end=n+1;
      orun=ostart+7; orun->invers=ostart+14;
      orun=ostart+8; orun->invers=ostart+9; orun->end=n+3;
      orun=ostart+9; orun->invers=ostart+8; orun->end=n+2;
      orun=ostart+10; orun->invers=ostart+15;
      orun=ostart+11; orun->invers=ostart+2; orun->end=n;
      orun=ostart+12; orun->invers=ostart; orun->end=n;
      orun=ostart+13; orun->invers=ostart+4; orun->end=n+1;
      orun=ostart+14; orun->invers=ostart+7; orun->end=n+2;
      orun=ostart+15; orun->invers=ostart+10; orun->end=n+3;

      }

    for (i = 0; i < NUMEDGES; ++i)
      {
	edges[i].op = edges + NUMEDGES + i;
	edges[i+NUMEDGES].op = edges + i;
      }

}

/**************************************************************************/

static void
initialize_quadrangulations()

/* initialize edges for the generation of all simple quadrangulations */

{
    EDGE *run,*start;
    EDGE *orun,*ostart;
    int n,i;

/* Initialize the edges for operation P0. They look like

       a
      / \
    ?b   c    Vertex c is the new point. (a,c) and (d,c) are the
      \ /     new edges with a-->c taken as quadr_P0(n) 
       d

It is assumed that for 3<=n<MAXN after quadr_P0(n) there is room for 4 edges.
*/

    for (n=3; n<MAXN; n++)
      { run=start=quadr_P0(n);
	orun=ostart=start+NUMEDGES;
        run->end=n;
	orun->end=n;
        run->invers=start+1; 
        orun->invers=ostart+1; 

        run=start+1; run->start=n; 
        orun=ostart+1; orun->start=n; 
        run->prev=run->next=start+3;
        orun->next=orun->prev=ostart+3;
        run->invers=start; 
        orun->invers=ostart; 

        run=start+2; run->end=n; run->invers=start+3;
        orun=ostart+2; orun->end=n; orun->invers=ostart+3;

        run=start+3; run->start=n; run->invers=start+2;
        orun=ostart+3; orun->start=n; orun->invers=ostart+2;
        run->prev=run->next=start+1;
        orun->next=orun->prev=ostart+1;
      }

/* Then initialize the edges for operation P1 and P4. They look like

       a
    ? / \ 
   --b   c--  Vertex c is the new point. (a,c) and (d,c) are the
    ? \ / ?   new edges with a-->c taken as quadr_P1(n) 
       d

It is assumed that for 5<=n<MAXN after quadr_P1(n) there is room for 4 edges.
*/

    for (n=5; n<MAXN; n++)
      { run=start=quadr_P1(n);
	orun=ostart=start+NUMEDGES;
        run->end=n;
        orun->end=n;
        run->invers=start+1; 
        orun->invers=ostart+1; 

        run=start+1; run->start=n; 
        orun=ostart+1; orun->start=n; 
        run->invers=start; run->prev=start+3;
        orun->invers=ostart; orun->next=ostart+3;

        run=start+2; run->end=n; run->invers=start+3;
        orun=ostart+2; orun->end=n; orun->invers=ostart+3;

        run=start+3; run->start=n; run->invers=start+2;
        orun=ostart+3; orun->start=n; orun->invers=ostart+2;
        run->next=start+1;
        orun->prev=ostart+1;
      }

   for (i = 0; i < NUMEDGES; ++i)
     {
       edges[i].op = edges + NUMEDGES + i;
       edges[i+NUMEDGES].op = edges + i;
     }

}

/*******************************************************************/

static EDGE
*extend_quadr_P0(EDGE *e1)

/* extends a graph in the way given by the type P0 extension for
   quadrangulations. 

   The new path is inserted on the right (next) of e1. 

   It returns the directed edge starting at the new vertex with the
   new face on its left (prev).

   In the picture: e1=b->a, the edge c->a is returned and
   vertex c is the new point nv.

       ?
       a
      / \
    ?b   c    (degree[c] is 2)
      \ /    
       d
       ?
*/

{
  register EDGE *start, *dummy, *dummy2;
  register EDGE *oe1, *ostart, *odummy, *odummy2;
  int buffer;
  int v0, v1, i;

  check_it(401);

  start=quadr_P0(nv);
  ostart=start->op;

  degree[nv]=2;

  buffer=e1->end;
  dummy=e1->invers; dummy2=dummy->prev;
  odummy=dummy->op; odummy2=dummy2->op;
  start->start=buffer;
  ostart->start=buffer;
  start->next=dummy; dummy->prev=start;
  ostart->prev=odummy; odummy->next=ostart;
  start->prev=dummy2; dummy2->next=start;
  ostart->next=odummy2; odummy2->prev=ostart;
  degree[buffer]++;
  adj[buffer][nv] = adj[nv][buffer] = TRUE;
  v0 = buffer;

  start++;
  ostart++;
  firstedge[nv]=start;
  start->end=buffer;
  ostart->end=buffer;

  e1=e1->next;
  oe1=e1->op;
  start++;
  ostart++;
  buffer=e1->end;
  dummy=e1->invers; dummy2=dummy->next;
  odummy=oe1->invers; odummy2=odummy->prev;
  start->start=buffer;
  ostart->start=buffer;
  start->next=dummy2; dummy2->prev=start;
  ostart->prev=odummy2; odummy2->next=ostart;
  start->prev=dummy; dummy->next=start;
  ostart->next=odummy; odummy->prev=ostart;
  degree[buffer]++;
  adj[buffer][nv] = adj[nv][buffer] = TRUE;
  v1 = buffer;

  (start+1)->end=buffer;
  (ostart+1)->end=buffer;

  for (i=0;i<nv;i++)
    nnei[i][nv] = nnei[nv][i] = adj[v0][i] + adj[v1][i];

  nnei[v0][v1] = ++nnei[v1][v0];

  nv++; ne+=4;

  update_ncontract_quad();

  check_it(402);

  return (start-1); 
}

/**************************************************************************/

static void
reduce_quadr_P0(EDGE *e)

/* reduces a graph previously extended by the P0 extension for
   quadrangulations. The edge e must be the reference edge
   returned by the expansion routine.
*/

{
  register EDGE *dummy, *dummy2;
  register EDGE *odummy, *odummy2;
  int buffer;
  int v0, v1;

  check_it(408);

  nv--; ne-=4;

  dummy=e->invers;
  dummy2=dummy->next;
  odummy2=dummy2->op;
  dummy=dummy->prev;
  odummy=dummy->op;

  dummy->next=dummy2; dummy2->prev=dummy;
  odummy->prev=odummy2; odummy2->next=odummy;
  buffer=dummy->start;
  firstedge[buffer]=dummy; degree[buffer]--;

  adj[buffer][nv] = adj[nv][buffer] = FALSE;
  v0 = buffer;

  dummy=e->prev->invers;
  dummy2=dummy->next;
  odummy2=dummy2->op;
  dummy=dummy->prev;
  odummy=dummy->op;

  dummy->next=dummy2; dummy2->prev=dummy;
  odummy->prev=odummy2; odummy2->next=odummy;
  buffer=dummy->start;
  firstedge[buffer]=dummy; degree[buffer]--;
  v1 = buffer;

  adj[buffer][nv] = adj[nv][buffer] = FALSE;
  nnei[v0][v1] = --nnei[v1][v0];
  degree[nv] = 0;
  for (v0=0;v0<nv;v0++)
    nnei[nv][v0]=nnei[v0][nv]=0;
  
  update_ncontract_quad();

  check_it(409);

  return;
}

/*******************************************************************/

static EDGE
*extend_quadr_P1(EDGE *e)

/* extends a graph in the way given by the type P1 extension for
   3-connected quadrangulations. 

   It returns the directed edge characterizing this operation.
*/

{
  register EDGE *e1, *e1i, *e2, *e3, *e3i, *e4, *start, *dummy;
  register EDGE *oe, *oe1, *oe1i, *oe2, *oe3, *oe3i, *oe4, *ostart, *odummy;
  int end1, end2, center, end;
  int v, k;

  check_it(411);

  oe = e->op;
  center=e->start; 
  end=e->end;
  e1=e->prev;
  e1i=e1->invers;
  e2=e1i->prev;
  end1=e2->start;

  oe1 = e1->op;
  oe1i = e1i->op;
  oe2 = e2->op;

  e3=e->next;
  e3i=e3->invers;
  e4=e3i->next;
  end2=e4->start;

  oe3 = e3->op;
  oe3i = e3i->op;
  oe4 = e4->op;

  e1->next=e3; e3->prev=e1; degree[center]--; firstedge[center]=e1;
  adj[center][end] = adj[end][center] = FALSE;
  adj[end][nv] = adj[nv][end] = TRUE;
  oe1->prev=oe3; oe3->next=oe1;

  start=quadr_P1(nv);
  ostart=start->op;
  dummy=start+1;
  odummy=dummy->op;
  start->start=dummy->end=end1;
  ostart->start=odummy->end=end1;
  degree[end1]++;
  adj[end1][nv] = adj[nv][end1] = TRUE;
  start->next=e1i; e1i->prev=start;
  ostart->prev=oe1i; oe1i->next=ostart;
  start->prev=e2; e2->next=start;
  ostart->next=oe2; oe2->prev=ostart;

  dummy->next=e; e->prev=dummy;
  odummy->prev=oe; oe->next=odummy;

  start+=2; dummy=start+1;
  ostart+=2; odummy=ostart+1;
  start->start=dummy->end=end2;
  ostart->start=odummy->end=end2;
  degree[end2]++;
  adj[end2][nv] = adj[nv][end2] = TRUE;
  e3i->next=start; start->prev=e3i;
  oe3i->prev=ostart; ostart->next=oe3i;
  e4->prev=start; start->next=e4;
  oe4->next=ostart; ostart->prev=oe4;

  e->next=dummy; dummy->prev=e;
  oe->prev=odummy; odummy->next=oe;

  degree[nv]=3; e->start=e->invers->end=nv; firstedge[nv]=e;
  oe->start=oe->invers->end=nv;

  for (v=0;v<nv;v++)
    nnei[v][nv] = nnei[nv][v] = adj[end][v] + adj[end1][v] + adj[end2][v];

  for (k=0;k<degree[center]-2;k++)
    {
      e3 = e3->next;
      v = e3->end;
      nnei[end][v] = --nnei[v][end];
      /*
      if (adj[end][v] && nnei[end][v]==2)
	{
	  ncontract[v]++;
	  ncontract[end]++;
	}
      */
    }

  e3 = e->invers;
  for (k=0;k<degree[end]-1;k++)
    {
      e3 = e3->next;
      v = e3->end;
      nnei[center][v] = --nnei[v][center];
      /*
      if (adj[end][v] && nnei[end][v]==2)
	{
	  ncontract[v]++;
	  ncontract[end]++;
	}
      */
    }


  nnei[end1][end2] = ++nnei[end2][end1];

  nv++; ne+=4;

  update_ncontract_quad();

  check_it(412);

  return (e); 
}

/**************************************************************************/

static void
reduce_quadr_P1(EDGE *e)

/* reduces a graph previously extended by the type P1 extension for
   3-connected quadrangulations. The edge e must be the reference edge
*/

{
  register EDGE *e1, *e1i, *e2, *e3, *e3i, *e4;
  register EDGE *oe, *oe1, *oe1i, *oe2, *oe3, *oe3i, *oe4;
  int end1, end2, center, end;
  int k, v;

  check_it(418);

  nv--; ne-=4;

  oe=e->op;
  end=e->end;
  e2=e->prev->invers->prev;
  end1=e2->start;
  e1i=e2->next->next;
  e1=e1i->invers;
  center=e1->start;
  e3=e1->next;
  e3i=e3->invers;
  e4=e3i->next->next;
  end2=e4->start;

  oe2=e2->op;
  oe1i=e1i->op;
  oe1=e1->op;
  oe3=e3->op;
  oe3i=e3i->op;
  oe4=e4->op;

  e2->next=e1i; e1i->prev=e2; firstedge[end1]=e2; degree[end1]--;
  oe2->prev=oe1i; oe1i->next=oe2;
  e1->next=e; e->prev=e1; e->next=e3; e3->prev=e; 
  oe1->prev=oe; oe->next=oe1; oe->prev=oe3; oe3->next=oe; 
  e->start=e->invers->end=center; degree[center]++;
  oe->start=oe->invers->end=center;
  e3i->next=e4; e4->prev=e3i; firstedge[end2]=e4; degree[end2]--;
  oe3i->prev=oe4; oe4->next=oe3i;

  adj[nv][end] = adj[end][nv] = FALSE;
  adj[nv][end1] = adj[end1][nv] = FALSE;
  adj[nv][end2] = adj[end2][nv] = FALSE;
  adj[center][end] = adj[end][center] = TRUE;

  for (k=0;k<degree[center]-3;k++)
    {
      e3 = e3->next;
      v = e3->end;
      nnei[end][v] = ++nnei[v][end];
      /*
      if (adj[end][v] && nnei[end][v]==2)
	{
	  ncontract[v]++;
	  ncontract[end]++;
	}
      */
    }

  e3 = e->invers;
  for (k=0;k<degree[end]-1;k++)
    {
      e3 = e3->next;
      v = e3->end;
      nnei[center][v] = ++nnei[v][center];
      /*
      if (adj[end][v] && nnei[end][v]==2)
	{
	  ncontract[v]++;
	  ncontract[end]++;
	}
      */
    }

  nnei[end1][end2] = --nnei[end2][end1];

  degree[nv] = 0;
  for (v=0;v<nv;v++)
    nnei[nv][v]=nnei[v][nv]=0;

  update_ncontract_quad();

  check_it(419);

  return;
}

/*******************************************************************/

static EDGE
*extend_quadr_P4(EDGE *e, int newdegree)

/* extends a graph in the way given by the type P4 extension.

   It returns the directed edge characterizing this operation.
*/

{
  EDGE *e1, *e1i, *e2, *e3, *e3i, *e4, *e5, *e6, *start, *dummy;
  EDGE *oe, *oe1, *oe1i, *oe2, *oe3, *oe3i, *oe4, *oe5, *ostart, *odummy;
  int end1, end2, center;
  int v, k, l;

  check_it(441);

  oe = e->op;
  center=e->start; 
  e1=e->prev;
  e1i=e1->invers;
  e2=e1i->prev;
  end1=e2->start;

  oe1 = e1->op;
  oe1i = e1i->op;
  oe2 = e2->op;

  e3=e->next;
  for (k=0;k<newdegree-3;k++)
    e3=e3->next;
  e3i=e3->invers;
  e4=e3i->next;
  end2=e4->start;

  oe3 = e3->op;
  oe3i = e3i->op;
  oe4 = e4->op;

  e1->next=e3; e3->prev=e1; 
  degree[center] -= newdegree-2; firstedge[center]=e1;
  oe1->prev=oe3; oe3->next=oe1;

  start=quadr_P1(nv);
  ostart=start->op;
  dummy=start+1;
  odummy=dummy->op;
  start->start=dummy->end=end1;
  ostart->start=odummy->end=end1;
  degree[end1]++;
  adj[end1][nv] = adj[nv][end1] = TRUE;
  start->next=e1i; e1i->prev=start;
  ostart->prev=oe1i; oe1i->next=ostart;
  start->prev=e2; e2->next=start;
  ostart->next=oe2; oe2->prev=ostart;

  dummy->next=e; e->prev=dummy;
  odummy->prev=oe; oe->next=odummy;

  start+=2; dummy=start+1;
  ostart+=2; odummy=ostart+1;
  start->start=dummy->end=end2;
  ostart->start=odummy->end=end2;
  degree[end2]++;
  adj[end2][nv] = adj[nv][end2] = TRUE;
  e3i->next=start; start->prev=e3i;
  oe3i->prev=ostart; ostart->next=oe3i;
  e4->prev=start; start->next=e4;
  oe4->next=ostart; ostart->prev=oe4;

  for (v=0;v<nv;v++)
    nnei[v][nv] = nnei[nv][v] = adj[end1][v] + adj[end2][v];

  e5 = e;
  oe5 = e->op;
  for (k=0;k<newdegree-2;k++) 
    {
      e5->start=e5->invers->end=nv; 
      oe5->start=oe5->invers->end=nv;
      adj[center][e5->end] = adj[e5->end][center] = FALSE;
      adj[e5->end][nv] = adj[nv][e5->end] = TRUE;
      for (v=0;v<nv;v++)
	nnei[v][nv] = nnei[nv][v] += adj[e5->end][v];

      e6 = e3;
      for (l=0;l<degree[center]-2;l++)
	{
	  e6 = e6->next;
	  v = e6->end;
	  nnei[e5->end][v] = --nnei[v][e5->end];
	}

      e6 = e5->invers;
      for (l=0;l<degree[e5->end]-1;l++)
	{
	  e6 = e6->next;
	  v = e6->end;
	  nnei[center][v] = --nnei[v][center];
	}

      if (k!=newdegree-3)
	{
	  e5=e5->next;
	  oe5=e5->op;
	}
    }

  e5->next=dummy; dummy->prev=e5;
  oe5->prev=odummy; odummy->next=oe5;

  degree[nv]=newdegree; e->start=e->invers->end=nv; firstedge[nv]=e;
  oe->start=oe->invers->end=nv;

  nnei[end1][end2] = ++nnei[end2][end1];

  nv++; ne+=4;

  update_ncontract_quad();

  check_it(442);

  return (e); 
}

/**************************************************************************/

static void
reduce_quadr_P4(EDGE *e)

/* reduces a graph previously extended by the type P1 extension. 
   The edge e must be the reference edge
*/

{
  register EDGE *e1, *e1i, *e2, *e3, *e3i, *e4, *e5, *e6, *e7;
  register EDGE *oe, *oe1, *oe1i, *oe2, *oe3, *oe3i, *oe4, *oe5, *oe6;
  int end1, end2, center;
  int k, l, v, degreenv;

  check_it(448);

  nv--; ne-=4;

  degreenv=degree[nv];

  oe=e->op;
  e2=e->prev->invers->prev;
  end1=e2->start;
  e1i=e2->next->next;
  e1=e1i->invers;
  center=e1->start;
  e3=e1->next;
  e3i=e3->invers;
  e4=e3i->next->next;
  end2=e4->start;

  oe2=e2->op;
  oe1i=e1i->op;
  oe1=e1->op;
  oe3=e3->op;
  oe3i=e3i->op;
  oe4=e4->op;

  e5=e->prev->prev->prev;
  oe5=e5->op;

  e2->next=e1i; e1i->prev=e2; firstedge[end1]=e2; degree[end1]--;
  oe2->prev=oe1i; oe1i->next=oe2;
  e1->next=e; e->prev=e1; 

  e5->next=e3; e3->prev=e5; 
  oe1->prev=oe; oe->next=oe1; oe5->prev=oe3; oe3->next=oe5; 

  e6=e;
  oe6=e6->op;
  for (k=2;k<degreenv;k++)
    {
      e6->start=e6->invers->end=center; 
      oe6->start=oe6->invers->end=center;
      adj[nv][e6->end] = adj[e6->end][nv] = FALSE;
      adj[center][e6->end] = adj[e6->end][center] = TRUE;
      e7=e3;
      for (l=0;l<degree[center]-2;l++)
	{
	  e7 = e7->next;
	  v = e7->end;
	  nnei[e6->end][v] = ++nnei[v][e6->end];
	}
      
      e7 = e6->invers;
      for (l=0;l<degree[e6->end]-1;l++)
	{
	  e7 = e7->next;
	  v = e7->end;
	  nnei[center][v] = ++nnei[v][center];
	}
      
      e6=e6->next;
      oe6=e6->op;
    }
  degree[center] += degreenv-2;

  e3i->next=e4; e4->prev=e3i; firstedge[end2]=e4; degree[end2]--;
  oe3i->prev=oe4; oe4->next=oe3i;

  adj[nv][end1] = adj[end1][nv] = FALSE;
  adj[nv][end2] = adj[end2][nv] = FALSE;

  nnei[end1][end2] = --nnei[end2][end1];

  degree[nv] = 0;
  for (v=0;v<nv;v++)
    nnei[nv][v]=nnei[v][nv]=0;

  update_ncontract_quad();

  check_it(449);

  return;
}

/*******************************************************************/

static EDGE
*extend_quadr_P3(EDGE *e)

/* extends a graph in the way given by the type P3 extension for
   3-connected quadrangulations. That is: It inserts a square on the
   right hand side of e.

   It returns the directed edge characterizing this operation.
*/

{
  register EDGE *run, *start, *dummy, *e1;
  register EDGE *orun, *ostart, *odummy, *oe;
  int vertex;

  check_it(450);

  start=quadr_P3(nv);
  ostart=start->op;

  oe=e->op;
  run=e->next;
  orun=run->op;
  vertex=e->start;
  dummy=start+12;
  odummy=dummy->op;
  start->end=dummy->start=vertex; degree[vertex]++;
  ostart->end=odummy->start=vertex;
  e->next=run->prev=dummy; dummy->next=run; dummy->prev=e;
  oe->prev=orun->next=odummy; odummy->prev=orun; odummy->next=oe;
  adj[vertex][nv] = adj[nv][vertex] = 1;
  nnei[nv][e->end]=nnei[e->end][nv]=2;
  nnei[nv][run->end]=nnei[run->end][nv]=2;
  e1=run->next;
  while (e1 != e) {
    nnei[nv][e1->end]=nnei[e1->end][nv]=1;
    e1=e1->next;
  }

  e=run->invers;
  oe=e->op;
  run=e->next;
  orun=run->op;
  vertex=e->start;
  dummy=start+15;
  odummy=dummy->op;
  (start+10)->end=dummy->start=vertex; degree[vertex]++;
  (ostart+10)->end=odummy->start=vertex;
  e->next=run->prev=dummy; dummy->next=run; dummy->prev=e;
  oe->prev=orun->next=odummy; odummy->prev=orun; odummy->next=oe;
  adj[vertex][nv+3] = adj[nv+3][vertex] = 1;
  nnei[nv+3][e->end]=nnei[e->end][nv+3]=2;
  nnei[nv+3][run->end]=nnei[run->end][nv+3]=2;
  e1=run->next;
  while (e1 != e) {
    nnei[nv+3][e1->end]=nnei[e1->end][nv+3]=1;
    e1=e1->next;
  }

  e=run->invers;
  oe=e->op;
  run=e->next;
  orun=run->op;
  vertex=e->start;
  dummy--;
  odummy=dummy->op;
  (start+7)->end=dummy->start=vertex; degree[vertex]++;
  (ostart+7)->end=odummy->start=vertex;
  e->next=run->prev=dummy; dummy->next=run; dummy->prev=e;
  oe->prev=orun->next=odummy; odummy->prev=orun; odummy->next=oe;
  adj[vertex][nv+2] = adj[nv+2][vertex] = 1;
  nnei[nv+2][e->end]=nnei[e->end][nv+2]=2;
  nnei[nv+2][run->end]=nnei[run->end][nv+2]=2;
  e1=run->next;
  while (e1 != e) {
    nnei[nv+2][e1->end]=nnei[e1->end][nv+2]=1;
    e1=e1->next;
  }

  e=run->invers;
  oe=e->op;
  run=e->next;
  orun=run->op;
  vertex=e->start;
  dummy--;
  odummy=dummy->op;
  (start+4)->end=dummy->start=vertex; degree[vertex]++;
  (ostart+4)->end=odummy->start=vertex;
  e->next=run->prev=dummy; dummy->next=run; dummy->prev=e;
  oe->prev=orun->next=odummy; odummy->prev=orun; odummy->next=oe;
  adj[vertex][nv+1] = adj[nv+1][vertex] = 1;
  nnei[nv+1][e->end]=nnei[e->end][nv+1]=2;
  nnei[nv+1][run->end]=nnei[run->end][nv+1]=2;
  e1=run->next;
  while (e1 != e) {
    nnei[nv+1][e1->end]=nnei[e1->end][nv+1]=1;
    e1=e1->next;
  }

  firstedge[nv]=start;     degree[nv]=3;
  firstedge[nv+1]=start+3; degree[nv+1]=3;
  firstedge[nv+2]=start+6; degree[nv+2]=3;
  firstedge[nv+3]=start+9; degree[nv+3]=3;

  adj[nv][nv+1] = adj[nv+1][nv] = 1;
  adj[nv+1][nv+2] = adj[nv+2][nv+1] = 1;
  adj[nv+2][nv+3] = adj[nv+3][nv+2] = 1;
  adj[nv+3][nv] = adj[nv][nv+3] = 1;
  nnei[nv][nv+2]=nnei[nv+1][nv+3]=nnei[nv+2][nv]=nnei[nv+3][nv+1]=2;

  nv+=4; ne+=16;

  update_ncontract_quad();

  check_it(451);

  return (run->invers);
}

/*******************************************************************/

static void
reduce_quadr_P3(EDGE *e)

/* Reduces the graph formerly extended by the type P3 extension for
   3-connected quadrangulations. That is: It removes a square on the
   right hand side of e.
*/

{
  register EDGE *dummy,*odummy,*oe;
  int start,i;
  
  check_it(458);

  nv-=4; ne-=16;

  oe=e->op;
  dummy=e->next->next;
  odummy=dummy->op;
  start=e->start;
  degree[start]--; firstedge[start]=e;
  e->next=dummy; dummy->prev=e;
  oe->prev=odummy; odummy->next=oe;
  adj[nv][start] = adj[start][nv] = 0;

  e=dummy->invers;
  oe=e->op;
  dummy=e->next->next;
  odummy=dummy->op;
  start=e->start;
  degree[start]--; firstedge[start]=e;
  e->next=dummy; dummy->prev=e;
  oe->prev=odummy; odummy->next=oe;
  adj[nv+3][start] = adj[start][nv+3] = 0;

  e=dummy->invers;
  oe=e->op;
  dummy=e->next->next;
  odummy=dummy->op;
  start=e->start;
  degree[start]--; firstedge[start]=e;
  e->next=dummy; dummy->prev=e;
  oe->prev=odummy; odummy->next=oe;
  adj[nv+2][start] = adj[start][nv+2] = 0;

  e=dummy->invers;
  oe=e->op;
  dummy=e->next->next;
  odummy=dummy->op;
  start=e->start;
  degree[start]--; firstedge[start]=e;
  e->next=dummy; dummy->prev=e;
  oe->prev=odummy; odummy->next=oe;
  adj[nv+1][start] = adj[start][nv+1] = 0;

  for (i=0;i<nv;i++) {
    nnei[i][nv]=nnei[nv][i]=0;
    nnei[i][nv+1]=nnei[nv+1][i]=0;
    nnei[i][nv+2]=nnei[nv+2][i]=0;
    nnei[i][nv+3]=nnei[nv+3][i]=0;
  }
  degree[nv]=degree[nv+1]=degree[nv+2]=degree[nv+3]=0;
  adj[nv][nv+1] = adj[nv+1][nv] = 0;
  adj[nv+1][nv+2] = adj[nv+2][nv+1] = 0;
  adj[nv+2][nv+3] = adj[nv+3][nv+2] = 0;
  adj[nv+3][nv] = adj[nv][nv+3] = 0;
  nnei[nv][nv+2]=nnei[nv+1][nv+3]=nnei[nv+2][nv]=nnei[nv+3][nv+1]=0;

  update_ncontract_quad();

  check_it(459);
}

/*******************************************************************/

static int
legal_P4_reduction_all(EDGE *e)

/* checks whether the edge e characterizes a legal P4-reduction.
   General quadrangulations version.
   Returns TRUE if legal, FALSE otherwise.

   It is assumed that e->start has degree >= 3, and that there are
   no vertices of degree 2.
*/

{
  int v, w; 
  
  v = e->start;
  w = e->prev->invers->next->end;
  if (adj[v][w]) return FALSE;
  if (nnei[v][w] != 2) return FALSE;
  
  return TRUE;
}

/***********************************************************************/

static int
legal_P4_reduction_min3(EDGE *e)

/* checks whether the edge e characterizes a legal P4-reduction.
   mindeg 3 version
   Returns TRUE if yes, FALSE otherwise.

   It is assumed that edge->start has degree at least 3, and that both the side
   vertices edge->prev->end and edge->prev->prev->end have degree at least 4.
*/

{
  int v, w; 

  v = e->start;
  w = e->prev->invers->next->end;
  if (adj[v][w]) return FALSE;
  if (nnei[v][w] != 2) return FALSE;

 return TRUE;
}

/**********************************************************************/

static int
legal_P3_reduction_min3(EDGE *e)

/* checks whether the edge corresponds to a legal P3 reduction.
   Returns 1 if it is a legal reduction, 0 else. 
   mindeg 3 version

   Again it is assumed (necessary or not) that P1 has higher priority
   -- the nonexistence of a P1 reduction forces some subconfiguration.

   Furthermore it is assumed that the graph inspected is not the cube.
*/

{
  EDGE *run;

  if (degree[e->next->end]!=3) return 0;
  run=e->invers->prev;
  if (degree[run->end]!=3) return 0;
  run=run->prev->invers->prev;  
  if (degree[run->end]!=3) return 0;
  run=run->prev->invers->prev;  
  if (degree[run->end]!=3) return 0;
 
  return 1;
}

/****************************************************************************/

static void
find_extensions_quad_all(int nbtot, int nbop,
			 EDGE *ext[], int degext[], int *next)
  
/* Determine the inequivalent places to make extensions, for general
   quadrangulations.  The results are put in the arrays
   ext[0..*next-1] and degext[0..*next-1].  degext[] = 2 for P0, 
   degext[] = 3 for P1, degext[] > 3 for P4.
   nbtot and nbop are the usual group parameters.
*/
  
{
  EDGE *e,*e1,*elast;
  EDGE **nb,**nb0,**nb1,**nblim,**nboplim;
  int i,j,k,l,x,y;
  int mindeg;        /* minimum degree */
  int nmindeg;       /* number of vertices with degree of mindeg */
  int ideg;          /* degree index */
#define VCOLP0(i,j) (degree[i]<degree[j] ?		\
		     (degree[j]<<10)+degree[i] : (degree[i]<<10)+degree[j])
  
  /* first condition is to make compiler happy */
  if (nv > 0 && degree[nv-1] == 2)
    {
      x = firstedge[nv-1]->end;
      y = firstedge[nv-1]->next->end;
    }
  else
    x = -1;
  
  mindeg = nv;
  nmindeg = 0;
  for (i = 0; i < nv; ++i)
    if (ncontract[i]>0) {
      if (degree[i] == mindeg)
	++nmindeg;
      else if (degree[i] < mindeg)
	{
	  mindeg = degree[i];
	  nmindeg = 1;
	}
    }

  k = 0;
  if (nbtot == 1)
    {
      /* P0 extension for trivial group */
      
      RESETMARKS;
      for (l = 0; l < nv; ++l)
	{
	  e = elast = firstedge[l];
	  do
	    {
	      if (!ISMARKEDLO(e))
		{
		  i = e->end;
		  j = e->next->end;
		  if (x < 0 || i == nv-1 || j == nv-1) {
		    degext[k] = 2;
		    ext[k++] = e;
		  }
		  else
		    {
		      ++degree[i]; ++degree[j];
		      if (VCOLP0(i,j) >= VCOLP0(x,y)) 
			{degext[k] = 2; ext[k++] = e;}
		      --degree[i]; --degree[j];
		    }
		  MARKLO(e->next->op);
		  MARKLO(e->next->invers->next->invers);
		  MARKLO(e->next->invers->next->invers->next->op);
		}
	      e = e->next;
	    } while (e != elast);
	}
      
      /* P1/P4 extension for trivial group with degree 3,4,..,mindeg */
      
      for (ideg=3;ideg<=mindeg;ideg++)
	for (i = 0; i < nv; ++i)
	  {
	    if (degree[i] < 2*ideg-2) continue;
	    e = elast = firstedge[i];
	    do
	      {
		degext[k] = ideg;
		ext[k++] = e;
		e = e->next;
	      } while (e != elast);
	  }
      
      /* P1/P4 extension for trivial group with degree mindeg+1 */
      
      if (nmindeg <= 2)
	for (i = 0; i < nv; ++i)
	  {
	    if (degree[i] < 2*(mindeg+1)-2) continue;
	    e = elast = firstedge[i];
	    do
	      {
		e1 = e;
		for (j=0;j<mindeg-1;j++)
		  e1 = e1->next;
		if ((degree[e->prev->end] == mindeg && 
		     ncontract[e->prev->end]>0) +
		    (degree[e1->end] == mindeg && 
		     ncontract[e1->end]>0) ==
		    nmindeg)
		  {
		    degext[k] = (mindeg+1);
		    ext[k++] = e;
		  }
		e = e->next;
	      } while (e != elast);
	  }
    }
  
  /* code for nontrivial group */

  else
    {
      nboplim = (EDGE**)numbering[nbop==0?nbtot:nbop];
      nblim = (EDGE**)numbering[nbtot];
      nb0 = (EDGE**)numbering[0];
      
      for (i = 0; i < ne; ++i) {
	nb0[i]->index = i;
	nb0[i]->op->index = -1;
      }
      
      /* P0 extensions for non-trivial group */
      
      RESETMARKS;
      
      for (l = 0; l < ne; ++l)
	if (!ISMARKED(nb0[l]))
	  {
	    e = nb0[l];
	    i = e->end;
	    j = e->next->end;
	    if (x < 0 || i == nv-1 || j == nv-1) {
	      degext[k] = 2;
	      ext[k++] = e;
	    }
	    else
	      {
		++degree[i]; ++degree[j];
		if (VCOLP0(i,j) >= VCOLP0(x,y)) 
		  {degext[k] = 2; ext[k++] = e;}
		--degree[i]; --degree[j];
	      }
	    for (nb = nb0+l; nb < nboplim; nb += MAXE) {
	      MARKLO(*nb);
	      MARKLO((*nb)->next->op);
	    }
	    for ( ; nb < nblim; nb += MAXE) {
	      MARKLO((*nb)->invers->next->invers);	
	      MARKLO((*nb)->invers->next->invers->next->op);	
	    }
	    e1 = e->next->invers->next->invers;
	    if (e1->index >= 0) {
	      for (nb = nb0+(e1->index); nb < nboplim; nb += MAXE) {
		MARKLO(*nb);
		MARKLO((*nb)->next->op);
	      }
	      for ( ; nb < nblim; nb += MAXE) {
		MARKLO((*nb)->invers->next->invers);
		MARKLO((*nb)->invers->next->invers->next->op);	
	      }
	    }
	    else {
	      for (nb = nb0+(e1->op->index); nb < nboplim; nb += MAXE) {
		MARKLO((*nb)->op);
		MARKLO((*nb)->prev);
	      }
	      for ( ; nb < nblim; nb += MAXE) {
		MARKLO((*nb)->op->invers->next->invers);
		MARKLO((*nb)->invers->prev->invers->prev);	
	      }
	    }
	  }
      
      /* P1/P4 extensions for non-trivial group with degree n=3,4,...mindeg*/
      
      for (ideg=3;ideg<=mindeg;ideg++)
	{
	  RESETMARKS;
	  
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e) || degree[e->start] < 2*ideg-2) continue;
	
	      degext[k] = ideg; ext[k++] = e;
	
	      for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
		MARKLO(*nb1);
		e1 = (*nb1)->op;
		for (j=0;j<ideg-3;j++)
		  e1 = e1->prev;
		MARKLO(e1);
	      }

	      for (; nb1 < nblim; nb1 += MAXE) {
		e1 = *nb1;
		for (j=0;j<ideg-3;j++)
		  e1 = e1->prev;
		MARKLO(e1);
		MARKLO((*nb1)->op);
	      }
	    }
        }

      /* P1/P4 extensions with degree mindeg+1 */

      if (nmindeg <= 2)
        {
	  RESETMARKS;
	
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e) || degree[e->start] < 2*(mindeg+1)-2) continue;
	
	      e1 = e;
	      for (j=0;j<mindeg-1;j++)
		e1 = e1->next;
	      if ((degree[e->prev->end] == mindeg && 
		   ncontract[e->prev->end]>0) +
		  (degree[e1->end] == mindeg && 
		   ncontract[e1->end]>0) != nmindeg)
		continue;
	      degext[k] = mindeg+1; ext[k++] = e;
	
	      for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
		MARKLO(*nb1);
		e1 = (*nb1)->op;
		for (j=0;j<(mindeg+1)-3;j++)
		  e1 = e1->prev;
		MARKLO(e1);
	      }
	      
	      for (; nb1 < nblim; nb1 += MAXE) {
		e1 = *nb1;
		for (j=0;j<(mindeg+1)-3;j++)
		  e1 = e1->prev;
		MARKLO(e1);
		MARKLO((*nb1)->op);
	      }
            }
        }
    }
  *next = k;
}

/****************************************************************************/

static void
find_extensions_quad_min3(int nbtot, int nbop, 
			  EDGE *ext[], int degext[], int *next,
			  EDGE *extP3[], int *nextP3)
  
/* Determine the inequivalent places to make extensions, for the
   quadrangulations with mindeg 3.  The results are put in the arrays
   ext[0..*next-1] and degext[0..*next-1]. 
   degext[] = 3 for P1, degext[] > 3 for P4.
   
   nbtot and nbop are the usual group parameters.
*/
  
{
  EDGE *e,*e1,*elast;
  EDGE **nb,**nb0,**nb1,**nblim,**nboplim;
  int i,j,k;
  int mindeg;        /* minimum degree */
  int nmindeg;       /* number of vertices with degree of mindeg */
  int ideg;          /* degree index */

  mindeg = nv;
  nmindeg = 0;
  for (i = 0; i < nv; ++i)
    if (ncontract[i]>0) {
      if (degree[i] == mindeg)
	++nmindeg;
      else if (degree[i] < mindeg)
	{
	  mindeg = degree[i];
	  nmindeg = 1;
	}
    }

  if (nbtot == 1)
    {
      
      /* trivial group */

      k = 0;
      
      /* P1/P4 extension for trivial group with degree 3,4,..,mindeg */
      
      for (ideg=3;ideg<=mindeg;ideg++)
	for (i = 0; i < nv; ++i)
	  {
	    if (degree[i] < 2*ideg-2) continue;
	    e = elast = firstedge[i];
	    do
	      {
		degext[k] = ideg;
		ext[k++] = e;
		e = e->next;
	      } while (e != elast);
	  }
      
      /* P1/P4 extension for trivial group with degree mindeg+1 */
      
      if (nmindeg <= 2)
	for (i = 0; i < nv; ++i)
	  {
	    if (degree[i] < 2*(mindeg+1)-2) continue;
	    e = elast = firstedge[i];
	    do
	      {
		e1 = e;
		for (j=0;j<mindeg-1;j++)
		  e1 = e1->next;
		if ((degree[e->prev->end] == mindeg && 
		     ncontract[e->prev->end]>0) +
		    (degree[e1->end] == mindeg && 
		     ncontract[e1->end]>0) ==
		    nmindeg)
		  {
		    degext[k] = (mindeg+1);
		    ext[k++] = e;
		  }
		e = e->next;
	      } while (e != elast);
	  }
  
      *next = k;

      /* P3 extension for trivial group */
      
      if (nv <= maxnv-4)
	{
	    k = 0;
	    RESETMARKS;

	    for (i = 0; i < nv; ++i)
            {
                e = elast = firstedge[i];
                do
                {
                    if (!ISMARKEDLO(e))
		    {
		        extP3[k++] = e;
		        MARKLO(e);
			MARKLO(e->op->prev);
		        e1 = e->invers->prev;
		        MARKLO(e1);
			MARKLO(e1->op->prev);
		        e1 = e1->invers->prev;
		        MARKLO(e1);
			MARKLO(e1->op->prev);
		        e1 = e1->invers->prev;
		        MARKLO(e1);
			MARKLO(e1->op->prev);
		    }
                    e = e->next;
                } while (e != elast);
            }
  	    *nextP3 = k;
	}
	else
	    *nextP3 = 0;
    }
  else
    {

      /* non-trivial group */

      nboplim = (EDGE**)numbering[nbop==0?nbtot:nbop];
      nblim = (EDGE**)numbering[nbtot];
      nb0 = (EDGE**)numbering[0];
      
      for (i = 0; i < ne; ++i) {
	nb0[i]->index = i;
	nb0[i]->op->index = -1;
      }
      
      k = 0;
      
      /* P1/P4 extensions for non-trivial group with degree n=3,4,...mindeg*/
      
      for (ideg=3;ideg<=mindeg;ideg++)
	{
	  RESETMARKS;
	  
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e) || degree[e->start] < 2*ideg-2) continue;
	      
	      degext[k] = ideg; ext[k++] = e;
	      
	      for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
		MARKLO(*nb1);
		e1 = (*nb1)->op;
		for (j=0;j<ideg-3;j++)
		  e1 = e1->prev;
		MARKLO(e1);
	      }
	      
	      for (; nb1 < nblim; nb1 += MAXE) {
		e1 = *nb1;
		for (j=0;j<ideg-3;j++)
		  e1 = e1->prev;
		MARKLO(e1);
		MARKLO((*nb1)->op);
	      }
	    }
        }
      
      /* P1/P4 extensions with degree mindeg+1 */
      
      if (nmindeg <= 2)
        {
	  RESETMARKS;
	  
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (ISMARKEDLO(e) || degree[e->start] < 2*(mindeg+1)-2) continue;
	      
	      e1 = e;
	      for (j=0;j<mindeg-1;j++)
		e1 = e1->next;
	      if ((degree[e->prev->end] == mindeg && 
		   ncontract[e->prev->end]>0) +
		  (degree[e1->end] == mindeg && 
		   ncontract[e1->end]>0) != nmindeg)
		continue;
	      degext[k] = mindeg+1; ext[k++] = e;
	      
	      for (nb1 = nb0 + i; nb1 < nboplim; nb1 += MAXE) {
		MARKLO(*nb1);
		e1 = (*nb1)->op;
		for (j=0;j<(mindeg+1)-3;j++)
		  e1 = e1->prev;
		MARKLO(e1);
	      }
	      
	      for (; nb1 < nblim; nb1 += MAXE) {
		e1 = *nb1;
		for (j=0;j<(mindeg+1)-3;j++)
		  e1 = e1->prev;
		MARKLO(e1);
		MARKLO((*nb1)->op);
	      }
            }
        }
      *next = k;
      
      /* P3 extensions for non-trivial group */
      
      if (nv <= maxnv-4)
	{
	  RESETMARKS;
	  k = 0;
	  for (i = 0; i < ne; ++i)
            {
	      e = nb0[i];
	      if (!ISMARKEDLO(e))
		{
		  extP3[k++] = e;
		  for (nb = nb0 + i; nb < nboplim; nb += MAXE) {
		    e1 = elast = *nb;
		    do 
		      {
			MARKLO(e1);
			MARKLO(e1->op->invers);
			e1 = e1->invers->prev;
		      } while (e1 != elast);
		  }
		  for ( ; nb < nblim; nb += MAXE) {
		    e1 = elast = (*nb)->invers;
		    do 
		      {
			MARKLO(e1);
			MARKLO(e1->op->invers);
			e1 = e1->invers->prev;
		      } while (e1 != elast);
		  }
		}
            }
	  *nextP3 = k;
	}
      else
	*nextP3 = 0;
    }
}

/****************************************************************************/

static int
has_quadr_P0()

/* Test whether there is a legal P1 reduction */

{
    int i;

    for (i = nv; --i >= 0; )
        if (degree[i] == 2) return TRUE;
    
    return FALSE;
}

/****************************************************************************/

static int
has_quadr_P4_min3()

/* Test whether there is a legal P4 reduction, mindeg3 version.
   It is enough that there is a vertex with at least
   two consecutive neighbours of degree at least 4.
*/

{
  int i,prev4,v,w;
    EDGE *e,*run;

    for (i = 0; i < nv; ++i)
      {
        run = e = firstedge[i];
	prev4 = (degree[e->prev->end] > 3);
	do {
	  if (degree[run->end] > 3) {
	    if (prev4) {
	      v = run->start;
	      w = run->invers->next->end;
	      if (!adj[v][w] && nnei[v][w] == 2)
		return TRUE;
	    }
	    prev4 = TRUE;
	  }
	  else
	    prev4 = FALSE;
	  run = run->next;
	}
	while (run != e);
      }

    return FALSE;
}

/****************************************************************************/

static void
quadr_P4_legal_min3(EDGE *ref, EDGE *good_or[], int *ngood_or, int *ngood_ref,
               EDGE *good_mir[], int *ngood_mir, int *ngood_mir_ref,
               EDGE *prevP4[], int nprevP4, EDGE **hint)

/* The P4-operation with reference edge *ref has just been performed.
   prevP4[0..nprevP4-1] are all earlier P4s since the last P3.
   hint (unless it is NULL) is a suggestion for a P4-reduction that
   might be better than ref.
   Make a list in good_or[0..*ngood_or-1] of the reference edges of
   legal P4-reductions (oriented editions) that might be canonical,
   with the first *ngood_ref of those being ref.
   Make a list in good_mir[0..*ngood_mir-1] of the
   reference edges of legal P4-reductions (mirror-image editions)
   that might be canonical, with the first *ngood_mir_ref of those being
   ref->prev->prev->prev.
   *ngood_ref and *ngood_mir_ref might each be 0-1.  If they are
   both 0, nothing else need be correct.
   All the edges in good_or[] and good_mir[] must start with the same
   vertex degree and end with the same vertex degree (actually, colour
   as passed to canon_edge_oriented).
   P4-reductions have a priority (colour) based on the degrees of the
   end vertex and two side vertices.  It cannot be changed without
   changing find_extensions_quad too.

   mindeg 3 version
*/

{
    EDGE *e,*elast;
    int mindeg;
    int col,maxcol;
    int i,nor,nmir;

#define QORCOL(e) (((long)degree[e->end]<<15)+((long)degree[e->prev->prev->prev->end]<<10)+((long)degree[e->prev->end]<<5)+(long)degree[e->prev->prev->end])
#define QMIRCOL(e) (((long)degree[e->prev->prev->prev->end]<<15)+((long)degree[e->end]<<10)+((long)degree[e->prev->prev->end]<<5)+(long)degree[e->prev->end])

    RESETMARKS;
    nor = nmir = 0;

    mindeg = degree[ref->start];
    maxcol = QORCOL(ref);
    col = QMIRCOL(ref);
    if (col < maxcol)
        good_or[nor++] = ref;
    else if (col == maxcol) {
	good_or[nor++] = ref;
	good_mir[nmir++] = ref->prev->prev->prev;
    }
    else {
      maxcol = col;
      good_mir[nmir++] = ref->prev->prev->prev;
    }
 
    *ngood_ref = nor;
    *ngood_mir_ref = nmir;

    MARKLO(ref);
    MARKLO(ref->prev->prev->prev->op);

    e = *hint;
    if (e && !ISMARKEDLO(e)
             && degree[e->start] <= mindeg 
	     && degree[e->prev->end] >= 4 && degree[e->prev->prev->end] >= 4
             && legal_P4_reduction_min3(e))
    {
        if (degree[e->start] < mindeg)
        {
            *ngood_ref = *ngood_mir_ref = 0;
            return;
        }   
        else
	  {   
            col = QORCOL(e); 
            if (col > maxcol)
	      {
                *ngood_ref = *ngood_mir_ref = 0;
                return;
	      }
            else if (col == maxcol)
	      good_or[nor++] = e;
	    
            col = QMIRCOL(e);
            if (col > maxcol)
	      {
                *ngood_ref = *ngood_mir_ref = 0;
                return;
	      }
            else if (col == maxcol)
	      good_mir[nmir++] = e->prev->prev->prev;
	    
	  }
        MARKLO(e);
	MARKLO(e->prev->prev->prev->op);
    }

    for (i = nprevP4; --i >= 0; )
    {
	e = prevP4[i];
	if (!ISMARKEDLO(e) 
	    && degree[e->start] <= mindeg
	    && degree[e->prev->end] >= 4 && degree[e->prev->prev->end] >= 4
	    && legal_P4_reduction_min3(e)) {
	  if (degree[e->start] < mindeg)
	    {
	      *ngood_ref = *ngood_mir_ref = 0;
	      return;
	    }
	  else
	    {
	      col = QORCOL(e);
	      if (col > maxcol)
                {
		  *ngood_ref = *ngood_mir_ref = 0;
		  return;
                }
	      else if (col == maxcol)
		good_or[nor++] = e;
	      
	      col = QMIRCOL(e);
	      if (col > maxcol)
                {
		  *ngood_ref = *ngood_mir_ref = 0;
		  return;
                }
	      else if (col == maxcol)
		good_mir[nmir++] = e->prev->prev->prev;
	    }
	}
	MARKLO(e);
	MARKLO(e->prev->prev->prev->op);
    }

    if (nor > *ngood_ref || nmir > *ngood_mir_ref)
        prune_oriented_lists(good_or,&nor,ngood_ref,
                             good_mir,&nmir,ngood_mir_ref);

    if (*ngood_ref == 0 && *ngood_mir_ref == 0) return;

    for (i = 0; i < nv; ++i)
      if (degree[i] < mindeg)
        {
	  e = elast = firstedge[i];
	  do
            {
	      if (degree[e->prev->end] >= 4
		  && degree[e->prev->prev->end] >= 4
		  && legal_P4_reduction_min3(e))
		{
		  *ngood_ref = *ngood_mir_ref = 0;
		  *hint = e;
		  return;
		}
	      e = e->next;
	    } while (e != elast);
        }
      else if (degree[i] == mindeg)
        {
	  e = elast = firstedge[i];
	  do
            {
	      if (!ISMARKEDLO(e) 
		  && degree[e->prev->end] >= 4
		  && degree[e->prev->prev->end] >= 4
		  && legal_P4_reduction_min3(e))
                {
		  col = QORCOL(e);
		  if (col > maxcol)
                    {
		      *ngood_ref = *ngood_mir_ref = 0;
		      *hint = e;
		      return;
                    }
		  else if (col == maxcol) {
		    good_or[nor++] = e;
		    MARKLO(e->prev->prev->prev->op);
		  }

		  col = QMIRCOL(e);
		  if (col > maxcol)
                    {
		      *ngood_ref = *ngood_mir_ref = 0;
		      *hint = e;
		      return;
                    }
		  else if (col == maxcol) {
		    good_mir[nmir++] = e->prev->prev->prev;
		    MARKLO(e->prev->prev->prev->op);
		  }
                }
	      e = e->next;
            } while (e != elast);
        }

    if (nor > *ngood_ref || nmir > *ngood_mir_ref)
        prune_oriented_lists(good_or,&nor,ngood_ref,
                             good_mir,&nmir,ngood_mir_ref);

    *ngood_or = nor;
    *ngood_mir = nmir;    
}

/****************************************************************************/

static void
quadr_P0_legal_all(EDGE *ref, EDGE *good_or[], int *ngood_or, int *ngood_ref,
           EDGE *good_mir[], int *ngood_mir, int *ngood_mir_ref)

/* The P0-operation with reference edge *ref has just been performed.
   Make a list in good_or[0..*ngood_or-1] of the reference edges of
   legal P0-reductions (oriented editions) that might be canonical,
   with the first *ngood_ref of those being ref.
   Make a list in good_mir[0..*ngood_mir-1] of the
   reference edges of legal P0-reductions (mirror-image editions)
   that might be canonical, with the first *ngood_mir_ref of those being
   ref->next.
   *ngood_ref and *ngood_mir_ref might each be 0-1.  If they are
   both 0, nothing else need be correct.
   All the edges in good_or[] and good_mir[] must start with the same
   vertex degree and end with the same vertex degree (actually, colour
   as passed to canon_edge_oriented).
   P0-reductions have a priority (colour) based on the degrees of the
   end vertex and two side vertices.  It cannot be changed without
   changing find_extensions_quad too.

   version for general quadrangulations
*/

{
    EDGE *e;
    int col,col2,maxcol,maxcol2;
    int i,nor,nmir;
    int d1,d2,d3,d4;

#define VCOLPD(di,dj) (di<dj ? (dj<<10)+di : (di<<10)+dj)

    nor = nmir = 0;

    d1 = degree[ref->end];
    d2 = degree[ref->next->end];
    d3 = degree[ref->invers->next->end];
    d4 = degree[ref->invers->prev->end];

    maxcol = VCOLPD(d1,d2);
    maxcol2 = VCOLPD(d3,d4);

    if (d1 >= d2)
    {
        if (d3 >= d4) good_or[nor++] = ref;
        if (d4 >= d3) good_mir[nmir++] = ref;
    }
    if (d2 >= d1)
    {
        if (d4 >= d3) good_or[nor++] = ref->next;
        if (d3 >= d4) good_mir[nmir++] = ref->next;
    }
    
    *ngood_ref = nor;
    *ngood_mir_ref = nmir;

    for (i = nv-1; --i >= 0; )
        if (degree[i] == 2)
        {
	    e = firstedge[i];
    	    d1 = degree[e->end];
    	    d2 = degree[e->next->end];
	    col = VCOLPD(d1,d2);
	    if (col > maxcol)
	    {
		*ngood_ref = *ngood_mir_ref = 0;
		return;
	    }
	    else if (col == maxcol)
	    {
    		d3 = degree[e->invers->next->end];
    		d4 = degree[e->invers->prev->end];
		col2 = VCOLPD(d3,d4);
    
		if (col2 > maxcol2)
		{
                    *ngood_ref = *ngood_mir_ref = 0;
                    return;
                }
                else if (col2 == maxcol2)
                {
    		    if (d1 >= d2)
    		    {
        	        if (d3 >= d4) good_or[nor++] = e;
        	        if (d4 >= d3) good_mir[nmir++] = e;
    		    }
    		    if (d2 >= d1)
    		    {
        	        if (d4 >= d3) good_or[nor++] = e->next;
        	        if (d3 >= d4) good_mir[nmir++] = e->next;
    		    }
		}
	    }
        }

    if (nor > *ngood_ref || nmir > *ngood_mir_ref)
        prune_oriented_lists(good_or,&nor,ngood_ref,
                             good_mir,&nmir,ngood_mir_ref);

    *ngood_or = nor;
    *ngood_mir = nmir;    
}

/****************************************************************************/

static void
quadr_P4_legal_all(EDGE *ref, EDGE *good_or[], int *ngood_or, int *ngood_ref,
               EDGE *good_mir[], int *ngood_mir, int *ngood_mir_ref)

/* The P4-operation with reference edge *ref has just been performed.
   Make a list in good_or[0..*ngood_or-1] of the reference edges of
   legal P4-reductions (oriented editions) that might be canonical,
   with the first *ngood_ref of those being ref.
   Make a list in good_mir[0..*ngood_mir-1] of the
   reference edges of legal P4-reductions (mirror-image editions)
   that might be canonical, with the first *ngood_mir_ref of those being
   ref->prev->prev->prev.
   *ngood_ref and *ngood_mir_ref might each be 0-1.  If they are
   both 0, nothing else need be correct.
   All the edges in good_or[] and good_mir[] must start with the same
   vertex degree and end with the same vertex degree (actually, colour
   as passed to canon_edge_oriented).
   P4-reductions have a priority (colour) based on degrees of the
   next-to-side vertices and two side vertices.  It cannot be changed without
   changing find_extensions_quad too.

   version for general quadrangulations
   assumes no vertices of degree 2
*/

{
    EDGE *e,*elast;
    int mindeg;
    int col,maxcol;
    int i,nor,nmir;

#define QORCOL(e) (((long)degree[e->end]<<15)+((long)degree[e->prev->prev->prev->end]<<10)+((long)degree[e->prev->end]<<5)+(long)degree[e->prev->prev->end])
#define QMIRCOL(e) (((long)degree[e->prev->prev->prev->end]<<15)+((long)degree[e->end]<<10)+((long)degree[e->prev->prev->end]<<5)+(long)degree[e->prev->end])

    RESETMARKS;
    nor = nmir = 0;

    mindeg = degree[ref->start];
    maxcol = QORCOL(ref);
    col = QMIRCOL(ref);
    if (col < maxcol)
      good_or[nor++] = ref;
    else if (col == maxcol) {
      good_or[nor++] = ref;
      good_mir[nmir++] = ref->prev->prev->prev;
    }
    else {
      maxcol = col;
      good_mir[nmir++] = ref->prev->prev->prev;
    }

    *ngood_ref = nor;
    *ngood_mir_ref = nmir;

    MARKLO(ref);
    MARKLO(ref->prev->prev->prev->op);

    /* how could this be true??????  see three line above */

    if (nor > *ngood_ref || nmir > *ngood_mir_ref)
        prune_oriented_lists(good_or,&nor,ngood_ref,
                             good_mir,&nmir,ngood_mir_ref);

    if (*ngood_ref == 0 && *ngood_mir_ref == 0) return;

    for (i = 0; i < nv; ++i)
      if (degree[i] < mindeg)
        {
	  e = elast = firstedge[i];
	  do
            {
	      if (legal_P4_reduction_all(e))
		{
		  *ngood_ref = *ngood_mir_ref = 0;
		  return;
		}
	      e = e->next;
	    } while (e != elast);
	}
      else if (degree[i] == mindeg)
        {
	  e = elast = firstedge[i];
	  do
            {
	      if (!ISMARKEDLO(e) 
		  && legal_P4_reduction_all(e))
                {
		  col = QORCOL(e);
		  if (col > maxcol)
                    {
		      *ngood_ref = *ngood_mir_ref = 0;
		      return;
                    }
		  else if (col == maxcol) {
		    good_or[nor++] = e;
		    MARKLO(e->prev->prev->prev->op);
		  }
		  col = QMIRCOL(e);
		  if (col > maxcol)
                    {
		      *ngood_ref = *ngood_mir_ref = 0;
		      return;
                    }
		  else if (col == maxcol) {
		    good_mir[nmir++] = e->prev->prev->prev;
		    MARKLO(e->prev->prev->prev->op);
		  }
                }
	      e = e->next;
            } while (e != elast);
        }
    
    if (nor > *ngood_ref || nmir > *ngood_mir_ref)
      prune_oriented_lists(good_or,&nor,ngood_ref,
			   good_mir,&nmir,ngood_mir_ref);
    
    *ngood_or = nor;
    *ngood_mir = nmir;    
}

/****************************************************************************/

static void
quadr_P3_legal_min3(EDGE *ref, EDGE *good_or[], int *ngood_or, int *ngood_ref,
               EDGE *good_mir[], int *ngood_mir, int *ngood_mir_ref)

/* The P3-operation with reference edge *ref has just been performed.
   Make a list in good_or[0..*ngood_or-1] of the reference edges of
   legal P3-reductions (oriented editions) that might be canonical,
   with the first *ngood_ref of those being ref.
   Make a list in good_mir[0..*ngood_mir-1] of the
   reference edges of legal P3-reductions (mirror-image editions)
   that might be canonical, with the first *ngood_mir_ref of those being
   ref->next.
   *ngood_ref and *ngood_mir_ref might each be 0-4.  If they are
   both 0, nothing else need be correct.
   All the edges in good_or[] and good_mir[] must start with the same
   vertex degree and end with the same vertex degree (actually, colour
   as passed to canon_edge_oriented).
   P3-reductions have a priority (colour) based on the degrees of the
   end vertex and two side vertices.  It cannot be changed without
   changing find_extensions_quad too.

   mindeg 3 version
*/

{
    EDGE *e,*elast,*e1;
    int maxdeg;
    int col,col1,col2,maxcol;
    int i,j,nor,nmir;

#define QORCOL3(e) \
      (((long)degree[e->next->end]<<10)+(long)degree[e->prev->end])
#define QMIRCOL3(e) \
      (((long)degree[e->prev->end]<<10)+(long)degree[e->next->end])

    /* make compiler happy */
    maxcol = 0;

    RESETMARKS;

    maxdeg = 0;
    e1 = ref->next;

    for (j = 0; j < 4; ++j)
    {
	if (degree[e1->start] > maxdeg)
	{
	    maxdeg = degree[e1->start];
	    maxcol = QORCOL3(e1);
            nor = nmir = 0;
	    good_or[nor++] = e1;

	    col = QMIRCOL3(e1);
	    if (col > maxcol)
	    {
		nor = nmir = 0;
		good_mir[nmir++] = e1;
		maxcol = col;
	    }
	    else if (col == maxcol)
		good_mir[nmir++] = e1;
	}
	else if (degree[e1->start] == maxdeg)
	{
	    col = QORCOL3(e1);
	    if (col > maxcol)
            {
                nor = nmir = 0; 
                good_or[nor++] = e1;
                maxcol = col;
            }
            else if (col == maxcol)
                good_or[nor++] = e1;

            col = QMIRCOL3(e1);
            if (col > maxcol)
            {
                nor = nmir = 0; 
                good_mir[nmir++] = e1;
                maxcol = col;
            }
            else if (col == maxcol)
                good_mir[nmir++] = e1;
        }

	MARKLO(e1);
	e1 = e1->prev->invers->prev;
    }

    *ngood_ref = nor;
    *ngood_mir_ref = nmir;

    for (i = 0; i < nv; ++i)
        if (degree[i] > maxdeg)
        {
	    e = elast = firstedge[i];
	    do
	    {
	        if (!ISMARKEDLO(e) && legal_P3_reduction_min3(e->prev))
	        {
		    *ngood_ref = *ngood_mir_ref = 0;
		    return;
	        }
		e = e->next;
            } while (e != elast);
        }
	else if (degree[i] == maxdeg)
	{
	    e = elast = firstedge[i];
	    do
	    {
		if (!ISMARKEDLO(e))
		{
		    col1 = QORCOL3(e);
		    col2 = QMIRCOL3(e);
		    if (col1 > maxcol || col2 > maxcol)
		    {
			if (legal_P3_reduction_min3(e->prev))
			{
                            *ngood_ref = *ngood_mir_ref = 0;
                            return;
			}
		    }
		    else if ((col1 == maxcol || col2 == maxcol)
		                        && legal_P3_reduction_min3(e->prev))
		    {
			if (col1 == maxcol) good_or[nor++] = e;
			if (col2 == maxcol) good_mir[nmir++] = e;
		    }
                }
	        e = e->next;
	    } while (e != elast);
        }
    
    if (nor > *ngood_ref || nmir > *ngood_mir_ref)
        prune_oriented_lists(good_or,&nor,ngood_ref,
                             good_mir,&nmir,ngood_mir_ref);

    *ngood_or = nor;
    *ngood_mir = nmir;    
}

/****************************************************************************/

static void
scanquadrangulations_min3(int nbtot, int nbop, int dosplit,
                          EDGE *P4edge[], int nP4edges)

/* The main node of the recursion for quadrangulations of mindeg 3.
   As this procedure is entered, nv,ne,degree etc are set for some graph,
   and nbtot/nbop are the values returned by canon() for that graph.
   If dosplit==TRUE, this is the place to do splitting (if any).
   Splitting is a bit more complicated because the P operation adds
   two vertices.
   P4edge[0..nP4edges-1] are edges of the graph that were reference
   edges for P4-operations in ancestors of this graph.  They may not
   be reference edges for P4-reductions now, but at least they are edges.
   However, if nP4edges>0, we can say that certainly P4edge[nP4edges-1]
   is a P4-operation that was just done.
*/

{
#ifndef SPLITTEST
    EDGE *firstedge_save[MAXN];
#endif
    EDGE *extP4[MAXE],*extP3[MAXE];
    EDGE *good_or[MAXE],*good_mir[MAXE];
    int ngood_or,ngood_mir,ngood_ref,ngood_mir_ref;
    int degextP4[MAXE];
    int nextP4,nextP3,i;
    int xnbtot,xnbop;
    EDGE *newP4edge[MAXN],*rededge;
    EDGE *hint;

    /* make compiler happy */
    ngood_mir = ngood_or = 0;

    if (nv == maxnv)
    {
	got_one(nbtot,nbop,3);   /* Note connectivity is unknown */
	return;
    }

    if (dosplit && nv != nvroot)
    {
#ifdef SPLITTEST
        ADDBIG(splitcases,1);
        return;
#else
        if (splitcount-- != 0) return;
        splitcount = mod - 1;

        for (i = 0; i < nv; ++i) firstedge_save[i] = firstedge[i];
#endif
    }

#ifdef PRE_FILTER_QUAD_MIN3
    if (!(PRE_FILTER_QUAD_MIN3)) return;
#endif

#ifndef FIND_EXTENSIONS_QUAD_MIN3
#define FIND_EXTENSIONS_QUAD_MIN3 find_extensions_quad_min3
#endif

    FIND_EXTENSIONS_QUAD_MIN3(nbtot,nbop,extP4,degextP4,&nextP4,extP3,&nextP3);

    hint = NULL;
    for (i = 0; i < nextP4; ++i)
	{
	  if (degextP4[i] == 3)
	    rededge = extend_quadr_P1(extP4[i]);
	  else
	    rededge = extend_quadr_P4(extP4[i],degextP4[i]);
#ifdef FAST_FILTER_QUAD_MIN3
	  if (FAST_FILTER_QUAD_MIN3)
#endif  
	    {

	      /* ?????????????????? */
	      //	      if (nv == maxnv)
	      //got_one(1,1,3);

	      quadr_P4_legal_min3(rededge,good_or,&ngood_or,&ngood_ref,good_mir,
				  &ngood_mir,&ngood_mir_ref,P4edge,nP4edges,&hint);
	      if (ngood_ref+ngood_mir_ref > 0)
		{
		  if (nv == maxnv && !needgroup && ngood_or == ngood_ref
		      && ngood_mir == ngood_mir_ref)
		    got_one(1,1,3);
		  else if (ngood_or+ngood_mir==1)
		    {
		      P4edge[nP4edges] = rededge;
		      scanquadrangulations_min3(1,1,nv==splitlevel,
						P4edge,nP4edges+1);
		    }
		  else if (canon_edge_oriented(good_or,ngood_or,ngood_ref,
					       good_mir,ngood_mir,ngood_mir_ref,
					       degree,numbering,&xnbtot,&xnbop))
		    {
		      P4edge[nP4edges] = rededge;
		      scanquadrangulations_min3(xnbtot,xnbop,nv==splitlevel,
						P4edge,nP4edges+1);
		    }
		}
	    }
	  if (degextP4[i] == 3)
	    reduce_quadr_P1(rededge);
	  else
	    reduce_quadr_P4(rededge);
	}
    
    
    for (i = 0; i < nextP3; ++i)
      {   
        rededge = extend_quadr_P3(extP3[i]);
#ifdef FAST_FILTER_QUAD_MIN3
        if (FAST_FILTER_QUAD_MIN3)
#endif  
	  {
            if (!has_quadr_P4_min3())
	      {   
                quadr_P3_legal_min3(rededge,good_or,&ngood_or,&ngood_ref,
				    good_mir,&ngood_mir,&ngood_mir_ref);
                if (ngood_ref+ngood_mir_ref > 0
                    && canon_edge_oriented(good_or,ngood_or,ngood_ref,
                                           good_mir,ngood_mir,ngood_mir_ref,
                                           degree,numbering,&xnbtot,&xnbop))
		  scanquadrangulations_min3(xnbtot,xnbop,
					    nv>=splitlevel&&nv<=splitlevel+3,
					    newP4edge,0);
	      }
	  }
        reduce_quadr_P3(rededge);
      }
    
#ifndef SPLITTEST
    if (dosplit && nv != nvroot)
        for (i = 0; i < nv; ++i) firstedge[i] = firstedge_save[i];
#endif
}

/****************************************************************************/

static void
scanquadrangulations_all(int nbtot, int nbop) 

/* The main node of the recursion for general quadrangulations.
   As this procedure is entered, nv,ne,degree etc are set for some graph,
   and nbtot/nbop are the values returned by canon() for that graph.
*/

{
#ifndef SPLITTEST
    EDGE *firstedge_save[MAXN];
#endif
    EDGE *ext[MAXE];
    EDGE *good_or[MAXE],*good_mir[MAXE];
    int ngood_or,ngood_mir,ngood_ref,ngood_mir_ref;
    int degext[MAXE];
    int next,i;
    int xnbtot,xnbop;
    EDGE *rededge;

    if (nv == maxnv)
    {
	got_one(nbtot,nbop,2);   /* Note connectivity is really 2 or 3 */
	return;
    }

    if (nv == splitlevel)
      {
#ifdef SPLITTEST
        ADDBIG(splitcases,1);
        return;
#else
        if (splitcount-- != 0) return;
        splitcount = mod - 1;
	
        for (i = 0; i < nv; ++i) firstedge_save[i] = firstedge[i];
#endif
      }
    
#ifdef PRE_FILTER_QUAD_ALL
    if (!(PRE_FILTER_QUAD_ALL)) return;
#endif
    
#ifndef FIND_EXTENSIONS_QUAD_ALL
#define FIND_EXTENSIONS_QUAD_ALL find_extensions_quad_all
#endif
    
    FIND_EXTENSIONS_QUAD_ALL(nbtot,nbop,ext,degext,&next);
    
    for (i = 0; i < next; ++i)
      if (degext[i] == 2)
	{
	  rededge = extend_quadr_P0(ext[i]);
#ifdef FAST_FILTER_QUAD_ALL
	  if (FAST_FILTER_QUAD_ALL)
#endif  
	    {
	      quadr_P0_legal_all(rededge,good_or,&ngood_or,&ngood_ref,good_mir,
				 &ngood_mir,&ngood_mir_ref);
	      if (ngood_ref+ngood_mir_ref > 0)
		{
		  if (nv == maxnv && !needgroup && ngood_or == ngood_ref
		      && ngood_mir == ngood_mir_ref)
		    got_one(1,1,2);
		  else if (ngood_or+ngood_mir==1)
    		    scanquadrangulations_all(1,1);
		  else if (canon_edge_oriented(good_or,ngood_or,ngood_ref,
					       good_mir,ngood_mir,ngood_mir_ref,
					       degree,numbering,&xnbtot,&xnbop))
	            scanquadrangulations_all(xnbtot,xnbop);
		}
	    }
	  reduce_quadr_P0(rededge);
	}
      else if (degext[i] == 3)
	{   
	  rededge = extend_quadr_P1(ext[i]);
#ifdef FAST_FILTER_QUAD_ALL
	  if (FAST_FILTER_QUAD_ALL)
#endif  
	    {
	      if (!has_quadr_P0())
		{   
		  quadr_P4_legal_all(rededge,good_or,&ngood_or,&ngood_ref,
				     good_mir,&ngood_mir,&ngood_mir_ref);
		  if (ngood_ref+ngood_mir_ref > 0
		      && canon_edge_oriented(good_or,ngood_or,ngood_ref,
					     good_mir,ngood_mir,ngood_mir_ref,
					     degree,numbering,&xnbtot,&xnbop))
                    scanquadrangulations_all(xnbtot,xnbop);
		}
	    }
	  reduce_quadr_P1(rededge);
	}
      else
	{   
	  rededge = extend_quadr_P4(ext[i],degext[i]);
#ifdef FAST_FILTER_QUAD_ALL
	  if (FAST_FILTER_QUAD_ALL)
#endif  
	    {
	      if (!has_quadr_P0())
		{   
		  quadr_P4_legal_all(rededge,good_or,&ngood_or,&ngood_ref,
				     good_mir,&ngood_mir,&ngood_mir_ref);
		  if (ngood_ref+ngood_mir_ref > 0
		      && canon_edge_oriented(good_or,ngood_or,ngood_ref,
					     good_mir,ngood_mir,ngood_mir_ref,
					     degree,numbering,&xnbtot,&xnbop))
                    scanquadrangulations_all(xnbtot,xnbop);
		}
	    }
	  reduce_quadr_P4(rededge);
	}
#ifndef SPLITTEST
    if (nv == splitlevel)
        for (i = 0; i < nv; ++i) firstedge[i] = firstedge_save[i];
#endif
}

/***********************************************************************/

static int
getswitchvalue(char *arg, int *pj)

/* Find integer value for switch.
   arg is a pointer to a command-line argument.
   pj is an index into arg, which is updated.
   The value of the switch is the function return value.
   For example, if arg="-xyz1432q" and *pj=3 (pointing at "z"),
       the value 1432 is returned and *pj=7 (pointing at "2").
   An absent value is equivalent to 0.
*/

{
  int j,ans;

  ans = 0;
  for (j = *pj; arg[j+1] >= '0' && arg[j+1] <= '9'; ++j)
    ans = ans * 10 + (arg[j+1] - '0');

  *pj = j;
  return ans;
}

/****************************************************************************/

static void
getswitchvaluelist(char *arg, int *pj, int list[], char delim[],
		   int limit, int *count, char *sep)

/* Decode an integer-list argument for a switch.
   arg is a pointer to a command-line argument.
   *pj is the index of the switch name
     Following the switch name is a list of up to limit non-negative
     integers separated by characters from the string *sep.  They are
     found and put into list[0...] and the number of them into *count.
     The character before each number is put into delim[0..].
     The final value of *pj is the last digit of the last value.
     Empty values are taken as 0.
   For example, if arg = "az12:45-98q" and *pj=2 (at the 'z'), then
   the result will be list[0]=12, list[1]=45, list[2]=98,
                      delim[0]='z', delim[1]=':', delim[2]='-',
   *count=3, and *pj=10 (at the 'q').
*/
{
    int j,val;
    int go;

    *count = 0;
    j = *pj + 1;

    go = TRUE;
    delim[0] = arg[*pj];

    while (go)
    {
	if (*count >= limit)
	{
	    fprintf(stderr,
		    ">E %s: too many values for -%c\n",cmdname,arg[*pj]);
	    exit(1);
	}

	val = 0;
	while (arg[j] >= '0' && arg[j] <= '9')
	{
	    val = val * 10 + (arg[j] - '0');
	    ++j;
	}
        list[(*count)++] = val;	

	if (arg[j] != '\0' && strchr(sep,arg[j]))
	    delim[*count] = arg[j++];
        else
            go = FALSE;
    }

    *pj = j - 1;
}

/****************************************************************************/

static void
check_switch(char sw, char *ok_switches)

/* If ok_switches[sw] is zero, write an error message and exit. */

{
  if (!ok_switches[(int) sw])
    {
	fprintf(stderr,">E %s:  -%c is not permitted\n",cmdname,sw);
	exit(1);
    }
}

/****************************************************************************/

static void
decode_command_line(int argc, char *argv[])

/* Decode the command line, setting the global variables which
   give the switch values.  Some basic checking is done too, but
   the most detailed checking is done later.  If an error is
   found, this procedure never returns.

   The values for numerical parameters (minimumdeg) are set as -1 if the
   parameter is not mentioned, 0 if it appears without an integer
   following, and the given integer value if there is one.  Negative
   values are not currently allowed.

*/

{
  int i,j,ares,amod;
  char *arg,delims[2],*as;
  int badargs,argsgot;
  int numbounds;
  char ok_switches[256];

  for (i = 0; i < argc ; ++i) fprintf(stderr,"%s ",argv[i]);
  fprintf(stderr,"\n");

  cmdname = argv[0];

  for (i = 0; i < 256; ++i) ok_switches[i] = 0;
  for (as = SWITCHES; *as != '\0'; ++as) ok_switches[(int) *as] = 1;
  ok_switches['['] = ok_switches[']'] = ok_switches[' '] = 0;
  ok_switches[':'] = ok_switches['-'] = ok_switches['#'] = 0;
  for (as = SECRET_SWITCHES; *as != '\0'; ++as) ok_switches[(int) *as] = 1;

  argsgot = 0;
  badargs = FALSE;
  outfilename = NULL;
  aswitch = FALSE;
  upaswitch = FALSE;
  gswitch = FALSE;
  sswitch = FALSE;
  hswitch = FALSE;
  oswitch = FALSE;
  dswitch = FALSE;
  nswitch = FALSE;
  uswitch = FALSE;
  vswitch = FALSE;
  xswitch = FALSE;
  yswitch = FALSE;
  pswitch = FALSE;
  qswitch = FALSE;
  zeroswitch = FALSE;
  minconnec = -1;
  edgebound[0] = edgebound[1] = -1;
  maxfacesize = -1;
  polygonsize = -1;
  minimumdeg = -1;
  irrhole = FALSE;
  res = 0; mod = 1;

  for (i = 1; !badargs && i < argc; ++i)
    {
      arg = argv[i];
      if (arg[0] == '-' && arg[1] != '\0')
        {
	  for (j = 1; arg[j] != '\0'; ++j)
	    if (arg[j] == '\0') { }
	    BOOLSWITCH('o',oswitch)
	    BOOLSWITCH('d',dswitch)
	    BOOLSWITCH('h',hswitch)
	    BOOLSWITCH('a',aswitch)
	    BOOLSWITCH('A',upaswitch)
	    BOOLSWITCH('g',gswitch)
	    BOOLSWITCH('s',sswitch)
	    BOOLSWITCH('u',uswitch)
	    BOOLSWITCH('n',nswitch)
	    BOOLSWITCH('v',vswitch)
	    BOOLSWITCH('y',yswitch)
	    BOOLSWITCH('x',xswitch)
	    BOOLSWITCH('p',pswitch)
	    BOOLSWITCH('q',qswitch)
	    INTSWITCH('f',maxfacesize)
	    INTSWITCH('c',minconnec)
	    BOOLSWITCH('0',zeroswitch)
	    INTSWITCH('m',minimumdeg)
	    else if (arg[j] == 'e')
	    {
		CHECKSWITCH('e');
		getswitchvaluelist(arg,&j,edgebound,delims,2,&numbounds,":-");
		if (numbounds == 1) edgebound[1] = edgebound[0];
		if (edgebound[1] == 0) edgebound[1] = MAXE/2;
	    }
#ifdef PLUGIN_SWITCHES
	  PLUGIN_SWITCHES
#endif
            else
	      {
		CHECKSWITCH(arg[j]);
                badargs = TRUE;
	      }
        }
      else if (argsgot >= 4)
	badargs = TRUE;
      else if (argsgot == 0)
        {
	  j = -1;
	  maxnv = getswitchvalue(arg,&j);
	  if (arg[j+1] == 'd' && arg[j+2] == '\0')
	    if (maxnv & 1)
	      {
		fprintf(stderr,">E %s: n with 'd' must be even\n",cmdname);
		exit(1);
	      }
	    else  maxnv = maxnv / 2 + 2;
	  else if (arg[j+1] != '\0')
	    badargs = TRUE;
	  ++argsgot;
        }
      else if (argsgot == 1)
        {
	  j = -1;
	  ggenus = getswitchvalue(arg,&j);
	  if (arg[j+1] != '\0')
	    badargs = TRUE;
	  if (!nswitch && (ggenus < 0 || ggenus > MAXG/2))
	    {
	      fprintf(stderr,
		      ">E %s: g (without -n switch) must be between 0 and %d, inclusive\n",
		      cmdname,MAXG/2);
	      exit(1);
	    }
	  if (nswitch && (ggenus <= 0 || ggenus > MAXG))
	    {
	      fprintf(stderr,
		      ">E %s: g (with -n switch) must be between 1 and %d, inclusive\n",
		      cmdname,MAXG);
	      exit(1);
	    }
	  orient = !nswitch;
	  if (orient)
	    eulerc = 2 - 2*ggenus;
	  else
	    eulerc = 2 - ggenus;
	
	  ++argsgot;
        }
      else
        {
	  if (arg[0] == '-')
            {
	      if (argsgot == 0) badargs = TRUE;
            }
	  else if (sscanf(arg,"%d/%d",&ares,&amod) == 2)
            {
	      res = ares;
	      mod = amod;
            }
	  else
	    outfilename = arg;
	  ++argsgot;
        }
    }

  if (argsgot < 2) badargs = TRUE;

  if (badargs)
    {
      fprintf(stderr,
	      ">E Usage: %s %s n g [res/mod] [outfile]\n",cmdname,SWITCHES);
      exit(1);
    }

  if (res < 0 || res >= mod)
    {
      fprintf(stderr,">E %s: must have 0 <= res < mod\n",cmdname);
      exit(1);
    }

}

/****************************************************************************/

static void
initialize_splitting(int minlevel, int hint, int maxlevel)

/* Set splitlevel and splitcount.  minlevel and maxlevel are bounds
   on its value.  It must be that both minlevel and maxlevel are at
   least equal to the largest starting order (nv for external calls
   to scansimple() or similar routines), and at most equal to the
   smallest parent of a parent of an internal-output graph (call
   from scansimple() or similar to got_one() or similar).  The size
   of internal-output graphs is maxnv-1 for planar triangulations
   and maxnv for other classes.

   hint is a desirable value, which can be anything as the actual
   value used is forced between minlevel and maxlevel.  For plugins,
   the value of splithint is used instead if it is >= 0.

   In case there is no way to use splitting within those limits,
   it is turned off by setting splitlevel=0.  In that case only
   subcase 0 should produce output.

   Splitting occurs at the first level where nv >= splitlevel.
   If an operation can add k vertices, it must be that
   splitlevel <= maxnv - 5.
*/
{
  splitlevel = hint;
#ifdef PLUGIN
  if (splithint >= 0) splitlevel = splithint;
#endif

  if (splitlevel > maxlevel) splitlevel = maxlevel;

  if (splitlevel < minlevel && splitlevel > 0)
    {
      if (minlevel <= maxlevel) splitlevel = minlevel;
      else                      splitlevel = 0;
    }
  if (mod == 1) splitlevel = 0;

  splitcount = res;
}

/****************************************************************************/

static void
open_output_file(void)
  
/* Open the output file, and write a header if one is called for.
   All the needed information is in global vars.  Also check if
   maxn is too large for this format, and set the global procedure
   variables write_graph() and write_dual_graph().
*/
{
  int nvf;
  
  if (aswitch + upaswitch + gswitch + sswitch + uswitch >= 2)
    {
      fprintf(stderr,">E %s: -a, -A, -g, -s, -u are incompatible\n",cmdname);
      exit(1);
    }
  /* 
  if (upaswitch && dswitch)
    {
      fprintf(stderr,">E %s: -A and -d are incompatible (not yet implemented)\n",cmdname);
      exit(1);
    }
  */
  if (!uswitch)   /*UPDATE for -E and reconconsider */
    {
      nvf = dswitch ? 2*(maxnv-(polygonsize>=0))-4 : maxnv-(polygonsize>=0);
      if ((aswitch && nvf > 99) ||
	  (upaswitch && nvf > 99) ||
	  (gswitch && nvf > 255) ||
	  (sswitch && nvf > 255) ||
	  (!aswitch && !upaswitch && !gswitch && !sswitch && nvf > 255))
        {
	  fprintf(stderr,">E %s: n is too large for that output format\n",
		  cmdname);
	  exit(1);
        }
    }
  
  msgfile = stderr;
  if (outfilename == NULL)
    {
      outfilename = "stdout";
      outfile = stdout;
    }
  else if ((outfile = fopen(outfilename,
			    zeroswitch || aswitch || upaswitch || gswitch || sswitch ? "w" : "wb")) == NULL)
    {
      fprintf(stderr,
	      ">E %s: can't open %s for writing\n",cmdname,outfilename);
      perror(">E ");
      exit(1);
    }

  if (zeroswitch)  /* implied by oneswitch */
    write_graph = write_dual_graph = write_digits;
  else if (aswitch)
    {
      write_graph = write_alpha;
      write_dual_graph = write_dual_alpha;
    }
  else if (upaswitch)
    {
      write_graph = write_ualpha;
      write_dual_graph = write_dual_ualpha;
    }
  else if (gswitch)
    {
      write_graph = write_graph6;
      write_dual_graph = write_dual_graph6;
    }
  else if (sswitch)
    {
      write_graph = write_sparse6;
      write_dual_graph = write_dual_sparse6;
    }
  else
    {
      write_graph = write_planar_code;
      write_dual_graph = write_dual_planar_code;
    }

  if (!uswitch && !aswitch && !upaswitch)
    {
      if ((!zeroswitch && !hswitch && !gswitch && !sswitch &&
	   fwrite(PCODE,(size_t)1,PCODELEN,outfile) != PCODELEN) ||
          (hswitch && gswitch &&
	   fwrite(G6CODE,(size_t)1,G6CODELEN,outfile) != G6CODELEN) ||
	  (hswitch && sswitch &&
	   fwrite(S6CODE,(size_t)1,S6CODELEN,outfile) != S6CODELEN))
        {
	  fprintf(stderr,">E %s: error writing header\n",cmdname);
	  perror(">E ");
	  exit(1);
        }
    }

  if (oswitch && nswitch)
    {
      fprintf(stderr,">E %s: -o invalid with nonorientible surface\n",cmdname);
      perror(">E ");
      exit(1);
    }
}

/****************************************************************************/

static void
simple_dispatch()

/* All the cases not handled in the other dispatch routines. */
{
  int startingsize,nbtot,nbop,i,hint;
  char irrfilename[] = "genusx.alpha  ";  /* name of input file for
					     irreducible graphs */
  char inputline[MAXN+MAXE+4];

  if (minimumdeg <= 0) minimumdeg = 3;
  if (minconnec <= 0)  minconnec = MIN(minimumdeg,3);

  startingsize = 4;
  while ((startingsize-3)*(startingsize-4) < 6*(2-eulerc))
    startingsize++;
  if ((orient && ggenus == 2) || 
      (!orient && ggenus == 2) || 
      (!orient && ggenus == 3)) 
    startingsize++;

  CHECKRANGE(maxnv,"n",startingsize,MAXN);
  CHECKRANGE(minconnec,"-c",1,3);
  CHECKRANGE(minimumdeg,"-m",1,3);

  INCOMPAT(edgebound[0]>=0,"without -p","-e");
  INCOMPAT(maxfacesize>=0,"without -p","-f");
  INCOMPAT(minconnec == 3 && xswitch && minimumdeg == 3,"-c3","-x");

  if (minconnec < 3 && maxnv > 8*sizeof(long))
    {
      fprintf(stderr,">E %s: connectivity < 3 is restricted to n < %d\n",
	      cmdname,(int)(8*sizeof(long)));
      exit(1);
    }
  
  strcpy(intypename,"triangulations");
  if (dswitch) strcpy(outtypename,"cubic graphs");
  else         strcpy(outtypename,"triangulations");

  open_output_file();

  needgroup = oswitch || minconnec < 3;

  if (ggenus == 0)
    hint = MIN(maxnv <= 13 ? maxnv-2 : maxnv-3, 14);
  else if (ggenus == 1)
    hint = MIN(maxnv <= 13 ? maxnv-2 : maxnv-3, 12);
  else
    hint = MIN(maxnv <= 13 ? maxnv-2 : maxnv-3, 11);

  initialize_splitting(startingsize,hint,maxnv-1);
  if (splitlevel == 0 && res > 0) return;

  xconnec = minconnec;

  initialize();

  /* loop over all irreducible triangulations for this surface */

  if (orient)
    sprintf(irrfilename,"genus%d.alpha%c",ggenus,0);
  else
    sprintf(irrfilename,"genus%d.alpha%c",-ggenus,0);
  if ((irrfile = fopen(irrfilename,"r")) == NULL)
    {
      fprintf(stderr,
	      ">E %s: can't open %s for reading\n",cmdname,irrfilename);
      perror(">E ");
      exit(1);
    }

  nirr = 0;
  while (fgets(inputline,MAXN+MAXE+4,irrfile) != NULL)
    {
      if (sscanf(inputline,"%d",&nv) != 1)
	{
	  fprintf(stderr,
		  ">E %s: error read nv from %s for reading\n",
		  cmdname,irrfilename);
	  perror(">E ");
	  exit(1);
	}

      if (nv <= maxnv)
	{
	  if (splitlevel != 0 && nv >= splitlevel)
	    {
#ifdef SPLITTEST
	      ADDBIG(splitcases,1);
	      continue;
#endif
	      if (splitcount-- != 0) continue;
	      splitcount = mod - 1;
	    }
	  initialize_irr(inputline,6*(nv-eulerc));
	  check_it(999);

	  nirr++;
	  nvroot = nv;
	
	  ADDBIG(irrcount[nv],1);

	  if (!canon(degree,numbering,&nbtot,&nbop) && !yswitch)
	    {
	      fprintf(stderr,
		      ">E %s: line from %s does not have canonical vertex first\n",
		      cmdname,irrfilename);
	      fprintf(stderr,
		      ">E %s: input line: %s\n",
		      cmdname,inputline);
	      perror(">E ");
	      exit(1);
	    }
	
#ifdef FAST_FILTER_SIMPLE
	  if (FAST_FILTER_SIMPLE)
#endif 
	    scansimple(nbtot,nbop);
	}
    }

#ifdef STATS
  fprintf(msgfile,"Counts by min degree: ");
  for (i = minimumdeg; i <= 5; ++i)
    {
      PRINTBIG(msgfile,nummindeg[i]);
      if (i != 5) fprintf(msgfile," ");
      else fprintf(msgfile,"\n");
    }
#endif
  
  if (vswitch && minconnec < 3 && !xswitch)
    {
      fprintf(msgfile,"By embedding type: ");
      for (i = minconnec; i <= 3; ++i)
        {
	  if (i > minconnec) fprintf(msgfile," ");
	  PRINTBIG(msgfile,nout[i]);
	  if (i == 3) fprintf(msgfile," c3");
	  else        fprintf(msgfile," c%d",i);
	  if (i < 3) fprintf(msgfile,";");
        }
      fprintf(msgfile,"\n");
      
      if (oswitch)
	{
	  fprintf(msgfile,"By embedding type (O-P): ");
	  for (i = minconnec; i <= 3; ++i)
            {
	      if (i > minconnec) fprintf(msgfile," ");
	      PRINTBIG(msgfile,nout_op[i]);
	      if (i == 3) fprintf(msgfile," c3");
	      else        fprintf(msgfile," c%d",i);
	      if (i < 3) fprintf(msgfile,";");
            }
	  fprintf(msgfile,"\n");
	}
    }
}

/****************************************************************************/

static void
min4_dispatch()

/* Case of minimum degree >= 4, except for polygons and polytopes. */
{
  int startingsize,nbtot,nbop,hint;
  char irrfilename[] = "genusxxx.alpha  ";  /* name of input file for
						  irreducible graphs */
  char inputline[MAXN+MAXE+4];

  if (minimumdeg <= 0) minimumdeg = 4;
  if (minconnec <= 0) minconnec = 3;

  /*
    if (ggenus == 0)
      startingsize = 6;
    else if (orient && ggenus == 1)
      startingsize = 7;
    else if (orient && (ggenus == 2 || ggenus == 3))
      startingsize = 10;
    else if (orient && ggenus == 4)
      startingsize = 11;
    else if (!orient && ggenus == 1)
      startingsize = 6;
    else if (!orient && ggenus == 2)
      startingsize = 8;
    else if (!orient && (ggenus == 3 || ggenus == 4))
      startingsize = 9;
    else
      {
	fprintf(stderr,
		">E %s: don't know startingsize for genus %d\n",
		cmdname,ggenus);
	perror(">E ");
	exit(1);
      }
  */

  startingsize = 6;
  while ((startingsize-3)*(startingsize-4) < 6*(2-eulerc))
    startingsize++;
  if ((orient && ggenus == 2) || 
      (!orient && ggenus == 2) || 
      (!orient && ggenus == 3)) 
    startingsize++;
  
  CHECKRANGE(maxnv,"n",startingsize,MAXN);
  CHECKRANGE(minconnec,"-c",3,4);

  /* maximum possible value for minimumdeg depends on surface */

  CHECKRANGE(minimumdeg,"-m",4,6*(maxnv-eulerc)/maxnv);
  
  INCOMPAT(qswitch,"-m4 or -c4","-q");
  INCOMPAT(edgebound[0]>=0,"-m4 without -p","-e");
  INCOMPAT(maxfacesize>=0,"-m4 without -p","-f");
    
  strcpy(intypename,"triangulations");
  if (dswitch) strcpy(outtypename,"cubic graphs");
  else         strcpy(outtypename,"triangulations");
  
  open_output_file();
  
  if (ggenus == 0)
    hint = MIN(maxnv <= 19 ? maxnv-4 : maxnv-5, 17);
  else if (ggenus == 1)
    hint = MIN(maxnv-3, 12);
  else
    hint = MIN(maxnv-3, startingsize+3);
  
  initialize_splitting(startingsize,hint,maxnv-3);
  if (splitlevel == 0 && res > 0) return;
  
  initialize_min4();
  
  /* loop over all irreducible triangulations for this surface */
  
  if (ggenus == 0)
    sprintf(irrfilename,"genus0m4.alpha%c",0);
  else if (orient) 
    sprintf(irrfilename,"genus%d.alpha%c",ggenus,0);
  else
    sprintf(irrfilename,"genus%d.alpha%c",-ggenus,0);
  if ((irrfile = fopen(irrfilename,"r")) == NULL)
    {
      fprintf(stderr,
	      ">E %s: can't open %s for reading\n",cmdname,irrfilename);
      perror(">E ");
      exit(1);
    }
  
  nirr = 0;
  while (fgets(inputline,MAXN+MAXE+4,irrfile) != NULL)
    {
      if (sscanf(inputline,"%d",&nv) != 1)
	{
	  fprintf(stderr,
		  ">E %s: error read nv from %s for reading\n",
		  cmdname,irrfilename);
	  perror(">E ");
	  exit(1);
	}
      
      if (nv <= maxnv)
	{
	  if (splitlevel != 0 && nv >= splitlevel)
	    {
#ifdef SPLITTEST
	      ADDBIG(splitcases,1);
	      continue;
#endif
	      if (splitcount-- != 0) continue;
	      splitcount = mod - 1;
	    }
	  initialize_irr(inputline,6*(nv-eulerc));
	  check_it(888);
	  
	  nirr++;
	  nvroot = nv;
	  
	  ADDBIG(irrcount[nv],1);
	  
	  if (!canon(degree,numbering,&nbtot,&nbop) && !yswitch)
	    {
	      fprintf(stderr,
		      ">E %s: line from %s does not have canonical vertex first\n",
		      cmdname,irrfilename);
	      fprintf(stderr,
		      ">E %s: input line: %s\n",
		      cmdname,inputline);
	      perror(">E ");
	      exit(1);
	    }
	  
	  scanmin4(nbtot,nbop,splitlevel==nv,NULL);
	}
    }
}

/****************************************************************************/

static void
polytope_dispatch(void)

/* Polytopes. */
{
  int startingsize,nbtot,nbop,i,hint,maxundir;
  char irrfilename[] = "genusxxx.alpha  ";  /* name of input file for
						  irreducible graphs */
  char inputline[MAXN+MAXE+4];
  char fileextention[2][6] = {"alpha", "holes"};
  int ifiletype;
  EDGE *e,*e1,*ex;

  /* Polytopes are made by first making 3-connected triangulations,
     then deleting edges in a second phase.  minconn and minimumdeg
     are used in the first phase; minpolyconn and minpolydeg in the
     second phase.  Since simple triangulations are 3-connected and
     have minimum degree at least 3, minconn and minimumdeg are at
     least 3.  However, minpolyconn and minpolydeg can be anything
     from 1 to minconn and minimumdeg. */

    if (minconnec <= 0) minconnec = 3;
    if (minconnec < 3)
    {
	minpolyconnec = minconnec;
	minconnec = 3;
    }
    else
	minpolyconnec = minconnec;

    if (minimumdeg <= 0)
    {
	minimumdeg = minconnec;
	minpolydeg = minpolyconnec;
    }
    else
    {
	minpolydeg = minimumdeg;
	minimumdeg = MAX(minimumdeg,3);
    }

    if (minconnec > minimumdeg) minimumdeg = minconnec;
    if (minpolyconnec > minpolydeg) minpolydeg = minpolyconnec;

    /*
    startingsize = (minimumdeg == 4 ? 6 : 4);
    
    while ((startingsize-3)*(startingsize-4) < 6*(2-eulerc))
      startingsize++;
    if ((orient && ggenus == 2) || 
	(!orient && ggenus == 2) || 
	(!orient && ggenus == 3)) 
      startingsize++;
    */

    if (orient && ggenus == 0) 
      startingsize = 3;
    else if (orient && ggenus == 1)
      startingsize = 4;
    else if (orient && ggenus == 2)
      startingsize = 5;
    else if (!orient && ggenus == 1)
      startingsize = 3;
    else if (!orient && ggenus == 2)
      startingsize = 4;
    else if (!orient && ggenus == 3)
      startingsize = 4;
    else if (!orient && ggenus == 4)
      startingsize = 5;
    else {
      fprintf(stderr,
	      ">E %s: -p not available yet for orientation %d and genus %d\n",
	      cmdname,orient,ggenus); 
      exit(1);
    }

    CHECKRANGE(maxnv,"n",startingsize,MAXN);
    CHECKRANGE(maxnv,"n",startingsize,64);
    CHECKRANGE(minpolyconnec,"-c",1,3);
    //CHECKRANGE(minpolydeg,"-m",1,4);
    //CHECKRANGE(minpolydeg,"-m",1,3);

    INCOMPAT(qswitch,"-p","-q");
/* INCOMPAT(minconnec == 3 && xswitch,"-c3","-x");  changed 19/2/08 */
    INCOMPAT(minpolyconnec == 3 && xswitch,"-c3","-x");
    
    maxundir = (maxnv == 2 ? 1 : 3*(maxnv-eulerc));
    if (edgebound[0] < 0) edgebound[0] = 0;
    if (edgebound[1] < 0) edgebound[1] = maxundir;
    if (edgebound[1] > maxundir) edgebound[1] = maxundir;
    if (maxfacesize <= 0) maxfacesize = MAXE;

    CHECKRANGE(edgebound[0],"-e",0,maxundir);
    CHECKRANGE(edgebound[1],"-e",0,maxundir);
    PERROR(edgebound[0] > edgebound[1],
	   "The first value of -e can't be greater than the second");
    PERROR(maxnv*minpolydeg > 2*edgebound[1],
	   "The upper bound for -e is impossibly low");
    CHECKRANGE(maxfacesize,"-f",3,MAXE);

    xconnec = minpolyconnec;

    if (maxnv >= 4 && (edgebound[0] == maxundir || maxfacesize == 3))
    {
	pswitch = FALSE;
	edgebound[0] = edgebound[1] = maxfacesize = -1;
        if      (minimumdeg >= 4) min4_dispatch();
        else                      simple_dispatch();
	return;
    }

    edgebound[0] *= 2;
    edgebound[1] *= 2;

    strcpy(intypename,"maps");
    strcpy(outtypename,"maps");

    open_output_file();

    for (i = 0; i <= maxundir; ++i)
    {
	ZEROBIG(nout_e[i]);
	ZEROBIG(nout_e_op[i]);
#ifdef STATS
	ZEROBIG(numrooted_e[i]);
#endif
    }

    needgroup = TRUE;
    nirr = 0;

    /* once for regular triangulations and once for triangulations with holes */

    for (ifiletype=0; ifiletype<2; ifiletype++) {

      irrhole = (ifiletype == 1);
      
      if ((ggenus != 0 || (maxnv >= 4 && !irrhole) || (maxnv < 4 && irrhole)) &&
	  (!irrhole || minpolyconnec < 3))
	{
	  
	  if (ggenus == 0)
	    hint = MIN(maxnv <= 13 ? maxnv-2 : maxnv-3, 14);
	  else if (ggenus == 1)
	    hint = MIN(maxnv <= 13 ? maxnv-2 : maxnv-3, 12);
	  else
	    hint = MIN(maxnv <= 13 ? maxnv-2 : maxnv-3, 11);
	  
	  initialize_splitting(startingsize,hint,maxnv-1);
	  if (splitlevel == 0 && res > 0) return;
	  
	  initialize();
	  
	  /* loop over all irreducible triangulations for this surface */
	  
	  if (orient)
	    sprintf(irrfilename,"genus%d.%s%c",ggenus,fileextention[ifiletype],0);
	  else
	    sprintf(irrfilename,"genus%d.%s%c",-ggenus,fileextention[ifiletype],0);
	  if ((irrfile = fopen(irrfilename,"r")) == NULL)
	    {
	      fprintf(stderr,
		      ">E %s: can't open %s for reading\n",cmdname,irrfilename);
	      perror(">E ");
	      exit(1);
	    }
	  
	  while (fgets(inputline,MAXN+MAXE+4,irrfile) != NULL)
	    {
	      if (sscanf(inputline,"%d",&nv) != 1)
		{
		  fprintf(stderr,
			  ">E %s: error read nv from %s for reading\n",
			  cmdname,irrfilename);
		  perror(">E ");
		  exit(1);
		}
	      
	      if (nv <= maxnv)
		{
		  if (splitlevel != 0 && nv >= splitlevel)
		    {
#ifdef SPLITTEST
		      ADDBIG(splitcases,1);
		      continue;
#endif
		      if (splitcount-- != 0) continue;
		      splitcount = mod - 1;
		    }
		  initialize_irr(inputline,6*(nv-eulerc));
		  
		  if (ifiletype == 0) {
		    check_it(999);
		  }
		  else {
		    check_it(389);
		  }

		  /* determine "connectivity" of irreducible maps.
		     not true: this will be the same for all face irreducible maps 
		     generated from it by vertex splitting. */
		     
		  if (!irrhole)
		    polyconnec = 3;
		  else {
		    polyconnec = 2;
		    for (i = 0; i < nv && polyconnec == 2; ++i) {
		      e = ex = firstedge[i];
		      do
			{
			  e1 = e;
			  do
			    {
			      e1 = e1->invers->next;
			      if (e1 != e) {
				if (e1->start == e->start)
				  polyconnec = 1;
			      }
			    } while (polyconnec == 2 && e1 != e);
			  e = e->next;
			} while (polyconnec == 2 && e != ex);
		    }
		  }

		  /* continue with this irreducible map if connectivity is ok
		     and if number of edges in resulting maps will be in range
		  */

		  if (polyconnec >= minpolyconnec &&
		      (maxundir-edgebound[0]/2)>=(3*(nv-eulerc)-ne/2)) {

		    nirr++;
		    nvroot = nv;
		    
		    ADDBIG(irrcount[nv],1);
		    
		    if (!canon(degree,numbering,&nbtot,&nbop) && !yswitch)
		      {
			fprintf(stderr,
				">E %s: line from %s does not have canonical vertex first\n",
				cmdname,irrfilename);
			fprintf(stderr,
				">E %s: input line: %s\n",
				cmdname,inputline);
			perror(">E ");
			exit(1);
		      }
		    
#ifdef FAST_FILTER_SIMPLE
		    if (FAST_FILTER_SIMPLE)
#endif 
		      scansimple(nbtot,nbop);
		  }
		}
	    }
	  fclose(irrfile);
	}
    }

    if (vswitch)
    {
        for (i = 0; i <= maxundir; ++i)
        if (!ISZEROBIG(nout_e[i]))
        {
	    fprintf(msgfile," With %2d edges and %2d faces: ",i,eulerc+i-maxnv);
	    if (oswitch)
	    {
		PRINTBIG(msgfile,nout_e_op[i]);
		fprintf(msgfile," (");
                PRINTBIG(msgfile,nout_e[i]);
		fprintf(msgfile," classes)");
#ifdef STATS
		fprintf(msgfile," ");
		PRINTBIG(msgfile,numrooted_e[i]);
		fprintf(msgfile," rooted");
#endif
		fprintf(msgfile,"\n");
	    }
	    else
	    {
                PRINTBIG(msgfile,nout_e[i]);
	        fprintf(msgfile,"\n");
	    }
        }

	if (!xswitch && minpolyconnec < 3)
        {
            fprintf(msgfile,"By embedding type: ");
            for (i = minpolyconnec; i <= 3; ++i)
            {
                if (i > minpolyconnec) fprintf(msgfile," ");
                PRINTBIG(msgfile,nout[i]);
                if (i == 3) fprintf(msgfile," c3");
	        else        fprintf(msgfile," c%d",i);
                if (i < 3) fprintf(msgfile,";");
            }
            fprintf(msgfile,"\n");
    
	    if (oswitch)
	    {
                fprintf(msgfile,"By embedding type (O-P): ");
                for (i = minpolyconnec; i <= 3; ++i)
                {
                    if (i > minpolyconnec) fprintf(msgfile," ");
                    PRINTBIG(msgfile,nout_op[i]);
                    if (i == 3) fprintf(msgfile," c3");
	            else        fprintf(msgfile," c%d",i);
                    if (i < 3) fprintf(msgfile,";");
                }
                fprintf(msgfile,"\n");
	    }
	}
    }
}

/****************************************************************************/

static void
quadrangulation_dispatch()

/* Simple quadrangulations. */
{
  int startingsize,nbtot,nbop,hint;
  EDGE *P1edges[MAXN];
  char irrfilename[] = "genusx.quad   ";  /* name of input file for
					     irreducible graphs */

  char inputline[MAXN+MAXE+4];

  if (minimumdeg <= 0) minimumdeg = 2;
  if (orient && ggenus == 0)
    startingsize = (minimumdeg == 2 ? 4 : 8);
  else if (orient && ggenus == 1)
    startingsize = 5;
  else if (orient && ggenus == 2)
    startingsize = 7;
  else if (orient && ggenus == 3)
    startingsize = 8;
  else if (!orient && ggenus == 1)
    startingsize = 4;
  else if (!orient && ggenus == 2)
    startingsize = 6;
  else if (!orient && ggenus == 3)
    startingsize = 6;
  else if (!orient && ggenus == 4)
    startingsize = 7;
  else if (!orient && ggenus == 5)
    startingsize = 7;
  else {
    fprintf(stderr,
	    "unimplemented values for orientability (%d) and genus (%d).\n",
	    orient,ggenus);
    exit(0);
  }

  CHECKRANGE(maxnv,"n",startingsize,MAXN);
  CHECKRANGE(minimumdeg,"-m",2,3);

  strcpy(intypename,"quadrangulations");
  if (dswitch) strcpy(outtypename,"quartic graphs");
  else         strcpy(outtypename,"quadrangulations");
  
  open_output_file();
  
  if (minimumdeg == 2)
    {
        hint = MIN(maxnv <= 16 ? maxnv-5 : maxnv-6, maxnv < 25 ? 17 : 18);
        initialize_splitting(startingsize,hint,maxnv-1);
        if (splitlevel == 0 && res > 0) return;
        initialize_quadrangulations();

	/* loop over all irreducible quadrangulations for this surface */
	
	if (orient)
	  sprintf(irrfilename,"genus%d.quad%c",ggenus,0);
	else
	  sprintf(irrfilename,"genus%d.quad%c",-ggenus,0);
	if ((irrfile = fopen(irrfilename,"r")) == NULL)
	  {
	    fprintf(stderr,
		    ">E %s: can't open %s for reading\n",cmdname,irrfilename);
	    perror(">E ");
	    exit(1);
	  }
	
	nirr = 0;
	while (fgets(inputline,MAXN+MAXE+4,irrfile) != NULL)
	  {
	    if (sscanf(inputline,"%d",&nv) != 1)
	      {
		fprintf(stderr,
			">E %s: error read nv from %s for reading\n",
			cmdname,irrfilename);
		perror(">E ");
		exit(1);
	      }
	    
	    if (nv <= maxnv)
	      {
		if (splitlevel != 0 && nv >= splitlevel)
		  {
#ifdef SPLITTEST
		    ADDBIG(splitcases,1);
		    continue;
#endif
		    if (splitcount-- != 0) continue;
		    splitcount = mod - 1;
		  }
		initialize_irr_quad(inputline,4*(nv-eulerc));
		check_it(400);
		
		nirr++;
		nvroot = nv;
		
		ADDBIG(irrcount[nv],1);
		
		if (!canon_quad(degree,numbering,&nbtot,&nbop))
		  {
		    fprintf(stderr,
			    ">E %s: line from %s does not have canonical vertex first\n",
			    cmdname,irrfilename);
		    fprintf(stderr,
			    ">E %s: input line: %s\n",
			    cmdname,inputline);
		    perror(">E ");
		    //exit(1);
		  }
		
		scanquadrangulations_all(nbtot,nbop);
#ifdef STATS
		fprintf(msgfile,"Counts by min degree: ");
		PRINTBIG(msgfile,nummindeg[2]);
		fprintf(msgfile," ");
		PRINTBIG(msgfile,nummindeg[3]);
		fprintf(msgfile,"\n");
#endif
	      }
	  }
    }
    else
    {
        hint = MIN(maxnv <= 23 ? maxnv-5 : maxnv-6, maxnv < 30 ? 23 : 24);
        initialize_splitting(startingsize,hint,maxnv-2);
        if (splitlevel == 0 && res > 0) return;
        initialize_min3_quadrangulations();

	/* loop over all min3 irreducible quadrangulations for this surface */
	
	if (ggenus == 0)
	  sprintf(irrfilename,"genus0m3.quad%c",0);
	else if (!orient && ggenus == 1)
	  sprintf(irrfilename,"genus-1m3.quad%c",0);
	else if (orient)
	  sprintf(irrfilename,"genus%d.quad%c",ggenus,0);
	else
	  sprintf(irrfilename,"genus%d.quad%c",-ggenus,0);
	if ((irrfile = fopen(irrfilename,"r")) == NULL)
	  {
	    fprintf(stderr,
		    ">E %s: can't open %s for reading\n",cmdname,irrfilename);
	    perror(">E ");
	    exit(1);
	  }
	
	nirr = 0;
	while (fgets(inputline,MAXN+MAXE+4,irrfile) != NULL)
	  {
	    if (sscanf(inputline,"%d",&nv) != 1)
	      {
		fprintf(stderr,
			">E %s: error read nv from %s for reading\n",
			cmdname,irrfilename);
		perror(">E ");
		exit(1);
	      }
	    
	    if (nv <= maxnv)
	      {
		if (splitlevel != 0 && nv >= splitlevel)
		  {
#ifdef SPLITTEST
		    ADDBIG(splitcases,1);
		    continue;
#endif
		    if (splitcount-- != 0) continue;
		    splitcount = mod - 1;
		  }
		initialize_irr_quad(inputline,4*(nv-eulerc));
		check_it(440);
		
		nirr++;
		nvroot = nv;
		
		ADDBIG(irrcount[nv],1);
		
		if (!canon_quad(degree,numbering,&nbtot,&nbop))
		  {
		    fprintf(stderr,
			    ">E %s: line from %s does not have canonical vertex first\n",
			    cmdname,irrfilename);
		    fprintf(stderr,
			    ">E %s: input line: %s\n",
			    cmdname,inputline);
		    perror(">E ");
		    //exit(1);
		  }
		
		scanquadrangulations_min3(nbtot,nbop,
					  splitlevel==8,P1edges,0);
#ifdef STATS
		fprintf(msgfile,"Counts by min degree: ");
		PRINTBIG(msgfile,nummindeg[2]);
		fprintf(msgfile," ");
		PRINTBIG(msgfile,nummindeg[3]);
		fprintf(msgfile,"\n");
#endif
	      }
	  }
    }
}

/****************************************************************************/

int
main(int argc, char *argv[])
{
  int i;
#if CPUTIME
  struct tms timestruct0,timestruct1;

  times(&timestruct0);
#endif

  if (argc > 1
      && (strcmp(argv[1],"-help") == 0 || (strcmp(argv[1],"--help") == 0)))
    {
      fprintf(stderr,"Surftri version %s\n",VERSION);
      fprintf(stderr,
	      "Usage: %s %s n [res/mod] [outfile]\n",argv[0],SWITCHES);
#ifdef HELPMESSAGE
	HELPMESSAGE;
#endif
      fprintf(stderr,"See surftri-guide.txt for more information.\n");
      return 0;
    }

  decode_command_line(argc,argv);

#ifdef SPLITTEST
  if (mod == 1) mod = 2;
  uswitch = TRUE;
  aswitch = upaswitch = gswitch = sswitch = FALSE;
#endif

    if (minconnec < 0 && (xswitch))
    {
	fprintf(stderr, 
	 ">E %s: -x can only be used in conjunction with -c\n",cmdname);
	exit(1);
    }

    for (i = 0; i < 6; ++i)
    {
        ZEROBIG(nout[i]);
        ZEROBIG(nout_op[i]);
    }

#ifdef STATS
    ZEROBIG(ntriv);
    ZEROBIG(numrooted);
    ZEROBIG(nummindeg[1]);
    ZEROBIG(nummindeg[2]);
    ZEROBIG(nummindeg[3]);
    ZEROBIG(nummindeg[4]);
    ZEROBIG(nummindeg[5]);
#endif

  for (i=0; i<MAXN; i++)
    ZEROBIG(irrcount[i]);

#ifdef SPLITTEST
  ZEROBIG(splitcases);
#endif

#ifdef PLUGIN_INIT
  PLUGIN_INIT;
#endif
  
  minpolydeg = -1;
  minpolyconnec = -1;
  
  if (pswitch)              polytope_dispatch();
  else if (qswitch)         quadrangulation_dispatch();
  else if (minimumdeg >= 4) min4_dispatch();
  else                      simple_dispatch();

#if CPUTIME
  times(&timestruct1);
#endif

  ZEROBIG(totalout);
  ZEROBIG(totalout_op);
  for (i = 0; i < 6; ++i)
    {
      SUMBIGS(totalout,nout[i]);
      if (oswitch) SUMBIGS(totalout_op,nout_op[i]);
    }
  
  dosummary = TRUE;
#ifdef SUMMARY
  nv = maxnv;
  SUMMARY();
#endif

  if (vswitch)
    for (i=0; i<MAXN; i++)
      if (!ISZEROBIG(irrcount[i]))
	{
	  PRINTBIG(msgfile,irrcount[i]);
	  fprintf(msgfile,
		  " irreducible %s read and used with nv %d\n",intypename,i);
	}

#ifdef SPLITTEST
  PRINTBIG(msgfile,splitcases);
  fprintf(msgfile," splitting cases at level=%d",splitlevel);
#if CPUTIME
  fprintf(msgfile,"; cpu=%.2f sec\n",
	  (double)(timestruct1.tms_utime+timestruct1.tms_stime
		   -timestruct0.tms_utime-timestruct0.tms_stime) /
	  (double)CLK_TCK);
#else
  fprintf(msgfile,"\n");
#endif
  return 0;
#endif

  if (!dosummary) return 0;

  if (oswitch && vswitch)
    {
      PRINTBIG(msgfile,totalout);
      fprintf(msgfile," isomorphism classes\n");
    }

  PRINTBIG(msgfile,(oswitch ? totalout_op : totalout));
  fprintf(msgfile," %s",outtypename);
  if (uswitch) fprintf(msgfile," generated");
  else         fprintf(msgfile," written to %s",outfilename);
#if CPUTIME
  fprintf(msgfile,"; cpu=%.2f sec\n",
	  (double)(timestruct1.tms_utime+timestruct1.tms_stime
		   -timestruct0.tms_utime+timestruct0.tms_stime) /
	  (double)CLK_TCK);
#else
  fprintf(msgfile,"\n");
#endif

  return 0;
}
